"""
.. currentmodule:: metawards.analysis

Functions
=========

.. autosummary::
    :toctree: generated/

    animate_plots
    create_average_plot
    create_overview_plot
    import_animate_modules
    import_graphics_modules
    save_summary_plots

"""

from ._animate_plots import *
from ._summary_plot import *
