"""
.. currentmodule:: metawards.app

Functions
=========

.. autosummary::
    :toctree: generated/

    get_parallel_scheme
    parse_args
    get_hostfile
    get_cores_per_node
    get_threads_per_task
    scoop_supervisor
    mpi_supervisor
    cli

"""

from .run import *
from .plot import *
