"""
.. currentmodule:: metawards.movers

Functions
=========

.. autosummary::
    :toctree: generated/

    go_isolate
    go_to

    move_custom
    move_default

"""

from ._go_isolate import *
from ._go_to import *

from ._move_custom import *
from ._move_default import *

# from ._go_on_holiday import *
# from ._go_back_home import *
# from ._go_weekend import *
# from ._go_weekday import *
# from ._go_to_hospital import *
# from ._go_on_lockdown import *
# from ._go_back_to_normal import *
