/* Generated by Cython 0.29.19 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-fopenmp"
        ],
        "libraries": [
            "metawards_random"
        ],
        "name": "metawards.extractors._output_core",
        "sources": [
            "src/metawards/extractors/_output_core.pyx"
        ]
    },
    "module_name": "metawards.extractors._output_core"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_19"
#define CYTHON_HEX_VERSION 0x001D13F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__metawards__extractors___output_core
#define __PYX_HAVE_API__metawards__extractors___output_core
/* Early includes */
#include <string.h>
#include <stdlib.h>
#include <omp.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/metawards/extractors/_output_core.pyx",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/*--- Type declarations ---*/
struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer;
struct __pyx_opt_args_9metawards_10extractors_12_output_core__allocate_inf_buffers;
struct __pyx_opt_args_9metawards_10extractors_12_output_core__add_to_buffer;
struct __pyx_t_9metawards_10extractors_12_output_core__red_variables;

/* "metawards/extractors/_output_core.pyx":26
 * 
 * 
 * cdef struct _inf_buffer:             # <<<<<<<<<<<<<<
 *     int count
 *     int *index
 */
struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer {
  int count;
  int *index;
  int *infected;
};

/* "metawards/extractors/_output_core.pyx":32
 * 
 * 
 * cdef _inf_buffer* _allocate_inf_buffers(int nthreads,             # <<<<<<<<<<<<<<
 *                                         int buffer_size=4096) nogil:
 *     cdef int size = buffer_size
 */
struct __pyx_opt_args_9metawards_10extractors_12_output_core__allocate_inf_buffers {
  int __pyx_n;
  int buffer_size;
};

/* "metawards/extractors/_output_core.pyx":75
 * 
 * 
 * cdef inline void _add_to_buffer(_inf_buffer *buffer, int index, int value,             # <<<<<<<<<<<<<<
 *                                 int *wards_infected,
 *                                 openmp.omp_lock_t *lock,
 */
struct __pyx_opt_args_9metawards_10extractors_12_output_core__add_to_buffer {
  int __pyx_n;
  int buffer_size;
};

/* "metawards/extractors/_output_core.pyx":93
 * 
 * # struct to hold all of the variables that will be reduced across threads
 * cdef struct _red_variables:             # <<<<<<<<<<<<<<
 *     int n_inf_wards
 *     int total_new
 */
struct __pyx_t_9metawards_10extractors_12_output_core__red_variables {
  int n_inf_wards;
  int total_new;
  int inf_tot;
  int pinf_tot;
  int susceptibles;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* MergeKeywords.proto */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'openmp' */

/* Module declarations from 'metawards.utils._get_array_ptr' */
static double *(*__pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr)(PyObject *); /*proto*/
static int *(*__pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr)(PyObject *); /*proto*/

/* Module declarations from 'metawards.extractors._output_core' */
static int __pyx_v_9metawards_10extractors_12_output_core__buffer_nthreads;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__S_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__E_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__I_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_9metawards_10extractors_12_output_core__R_buffers;
static struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_9metawards_10extractors_12_output_core__redvars;
static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(int, struct __pyx_opt_args_9metawards_10extractors_12_output_core__allocate_inf_buffers *__pyx_optional_args); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *, int); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__reset_inf_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *, int); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *, int *); /*proto*/
static CYTHON_INLINE void __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *, int, int, int *, omp_lock_t *, struct __pyx_opt_args_9metawards_10extractors_12_output_core__add_to_buffer *__pyx_optional_args); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__reset_reduce_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *, int); /*proto*/
static struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_f_9metawards_10extractors_12_output_core__allocate_red_variables(int); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__free_red_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *); /*proto*/
static void __pyx_f_9metawards_10extractors_12_output_core__reduce_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *, int); /*proto*/
#define __Pyx_MODULE_NAME "metawards.extractors._output_core"
extern int __pyx_module_is_main_metawards__extractors___output_core;
int __pyx_module_is_main_metawards__extractors___output_core = 0;

/* Implementation of 'metawards.extractors._output_core' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_enumerate;
static const char __pyx_k_[] = ", ";
static const char __pyx_k_E[] = "E";
static const char __pyx_k_I[] = "I";
static const char __pyx_k_R[] = "R";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_vs[] = " vs ";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_lock[] = "lock";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_play[] = "play";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_work[] = "work";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_files[] = "_files";
static const char __pyx_k_ifrom[] = "ifrom";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_links[] = "links";
static const char __pyx_k_nodes[] = "nodes";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_total[] = "total";
static const char __pyx_k_wards[] = "wards";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_latent[] = "latent";
static const char __pyx_k_nlinks[] = "nlinks";
static const char __pyx_k_nnodes[] = "nnodes";
static const char __pyx_k_output[] = "output-";
static const char __pyx_k_redvar[] = "redvar";
static const char __pyx_k_subnet[] = "subnet";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_Console[] = "Console";
static const char __pyx_k_Network[] = "Network";
static const char __pyx_k_Union_2[] = "_Union";
static const char __pyx_k_inf_tot[] = "inf_tot";
static const char __pyx_k_network[] = "network";
static const char __pyx_k_overall[] = "overall";
static const char __pyx_k_subinfs[] = "subinfs";
static const char __pyx_k_subnets[] = "subnets";
static const char __pyx_k_subpops[] = "subpops";
static const char __pyx_k_suscept[] = "suscept";
static const char __pyx_k_E_buffer[] = "E_buffer";
static const char __pyx_k_I_buffer[] = "I_buffer";
static const char __pyx_k_Networks[] = "Networks";
static const char __pyx_k_Profiler[] = "Profiler";
static const char __pyx_k_R_buffer[] = "R_buffer";
static const char __pyx_k_S_buffer[] = "S_buffer";
static const char __pyx_k_networks[] = "_networks";
static const char __pyx_k_nthreads[] = "nthreads";
static const char __pyx_k_pinf_tot[] = "pinf_tot";
static const char __pyx_k_profiler[] = "profiler";
static const char __pyx_k_safe_run[] = "_safe_run";
static const char __pyx_k_zero_all[] = "zero_all";
static const char __pyx_k_Workspace[] = "Workspace";
static const char __pyx_k_aggregate[] = "aggregate";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_incidence[] = "incidence";
static const char __pyx_k_inf_tot_i[] = "inf_tot_i";
static const char __pyx_k_network_2[] = "_network";
static const char __pyx_k_subspaces[] = "subspaces";
static const char __pyx_k_thread_id[] = "thread_id";
static const char __pyx_k_total_new[] = "total_new";
static const char __pyx_k_workspace[] = "workspace";
static const char __pyx_k_E_in_wards[] = "E_in_wards";
static const char __pyx_k_I_in_wards[] = "I_in_wards";
static const char __pyx_k_Infections[] = "Infections";
static const char __pyx_k_Population[] = "Population";
static const char __pyx_k_R_in_wards[] = "R_in_wards";
static const char __pyx_k_S_in_wards[] = "S_in_wards";
static const char __pyx_k_infections[] = "infections";
static const char __pyx_k_pinf_tot_i[] = "pinf_tot_i";
static const char __pyx_k_population[] = "population";
static const char __pyx_k_recovereds[] = "recovereds";
static const char __pyx_k_setup_core[] = "setup_core";
static const char __pyx_k_specialise[] = "specialise";
static const char __pyx_k_assert_sane[] = "assert_sane";
static const char __pyx_k_links_ifrom[] = "links_ifrom";
static const char __pyx_k_n_inf_wards[] = "n_inf_wards";
static const char __pyx_k_num_threads[] = "num_threads";
static const char __pyx_k_output_core[] = "output_core";
static const char __pyx_k_output_func[] = "output_func";
static const char __pyx_k_total_new_i[] = "total_new_i";
static const char __pyx_k_workspace_2[] = "_workspace";
static const char __pyx_k_NullProfiler[] = "NullProfiler";
static const char __pyx_k_demographics[] = "demographics";
static const char __pyx_k_infections_2[] = "_infections";
static const char __pyx_k_infections_i[] = "infections_i";
static const char __pyx_k_play_suscept[] = "play_suscept";
static const char __pyx_k_population_2[] = "_population";
static const char __pyx_k_susceptibles[] = "susceptibles";
static const char __pyx_k_N_INF_CLASSES[] = "N_INF_CLASSES";
static const char __pyx_k_links_suscept[] = "links_suscept";
static const char __pyx_k_multi_network[] = "multi-network";
static const char __pyx_k_n_inf_wards_i[] = "n_inf_wards_i";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_overall_output[] = "overall_output";
static const char __pyx_k_susceptibles_i[] = "susceptibles_i";
static const char __pyx_k_total_inf_ward[] = "total_inf_ward";
static const char __pyx_k_utils__console[] = "utils._console";
static const char __pyx_k_zero_subspaces[] = "zero_subspaces";
static const char __pyx_k_metawards_utils[] = "metawards.utils";
static const char __pyx_k_nlinks_plus_one[] = "nlinks_plus_one";
static const char __pyx_k_nnodes_plus_one[] = "nnodes_plus_one";
static const char __pyx_k_output_core_omp[] = "output_core_omp";
static const char __pyx_k_play_infections[] = "play_infections";
static const char __pyx_k_utils__profiler[] = "utils._profiler";
static const char __pyx_k_print_population[] = "print_population";
static const char __pyx_k_play_infections_i[] = "play_infections_i";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_output_core_serial[] = "output_core_serial";
static const char __pyx_k_total_new_inf_ward[] = "total_new_inf_ward";
static const char __pyx_k_total_inf_ward_buffer[] = "total_inf_ward_buffer";
static const char __pyx_k_N_INF_CLASSES_MINUS_ONE[] = "N_INF_CLASSES_MINUS_ONE";
static const char __pyx_k_total_new_inf_ward_buffer[] = "total_new_inf_ward_buffer";
static const char __pyx_k_Calculation_successful_the_bug_h[] = "Calculation successful, the bug has been mitigated";
static const char __pyx_k_Disagreement_in_accumulated_tota[] = "Disagreement in accumulated totals - indicates a program bug! ";
static const char __pyx_k_Repeating_this_part_of_the_calcu[] = "Repeating this part of the calculation in serial to correct this problem";
static const char __pyx_k_metawards_extractors__output_cor[] = "metawards.extractors._output_core";
static const char __pyx_k_src_metawards_extractors__output[] = "src/metawards/extractors/_output_core.pyx";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_kp_u_Calculation_successful_the_bug_h;
static PyObject *__pyx_n_s_Console;
static PyObject *__pyx_kp_u_Disagreement_in_accumulated_tota;
static PyObject *__pyx_n_s_E;
static PyObject *__pyx_n_s_E_buffer;
static PyObject *__pyx_n_s_E_in_wards;
static PyObject *__pyx_n_s_I;
static PyObject *__pyx_n_s_I_buffer;
static PyObject *__pyx_n_s_I_in_wards;
static PyObject *__pyx_n_s_Infections;
static PyObject *__pyx_n_s_N_INF_CLASSES;
static PyObject *__pyx_n_s_N_INF_CLASSES_MINUS_ONE;
static PyObject *__pyx_n_s_Network;
static PyObject *__pyx_n_s_Networks;
static PyObject *__pyx_n_s_NullProfiler;
static PyObject *__pyx_n_s_Population;
static PyObject *__pyx_n_s_Profiler;
static PyObject *__pyx_n_s_R;
static PyObject *__pyx_n_s_R_buffer;
static PyObject *__pyx_n_s_R_in_wards;
static PyObject *__pyx_kp_u_Repeating_this_part_of_the_calcu;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_n_s_S_buffer;
static PyObject *__pyx_n_s_S_in_wards;
static PyObject *__pyx_n_s_Union;
static PyObject *__pyx_n_s_Union_2;
static PyObject *__pyx_n_s_Workspace;
static PyObject *__pyx_n_s_aggregate;
static PyObject *__pyx_n_u_aggregate;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_assert_sane;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_demographics;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_files;
static PyObject *__pyx_n_s_func;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_ifrom;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_incidence;
static PyObject *__pyx_n_s_inf_tot;
static PyObject *__pyx_n_s_inf_tot_i;
static PyObject *__pyx_n_s_infections;
static PyObject *__pyx_n_s_infections_2;
static PyObject *__pyx_n_s_infections_i;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_latent;
static PyObject *__pyx_n_s_links;
static PyObject *__pyx_n_s_links_ifrom;
static PyObject *__pyx_n_s_links_suscept;
static PyObject *__pyx_n_s_lock;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metawards_extractors__output_cor;
static PyObject *__pyx_n_s_metawards_utils;
static PyObject *__pyx_kp_u_multi_network;
static PyObject *__pyx_n_s_n_inf_wards;
static PyObject *__pyx_n_s_n_inf_wards_i;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_network;
static PyObject *__pyx_n_s_network_2;
static PyObject *__pyx_n_s_networks;
static PyObject *__pyx_n_s_nlinks;
static PyObject *__pyx_n_s_nlinks_plus_one;
static PyObject *__pyx_n_s_nnodes;
static PyObject *__pyx_n_s_nnodes_plus_one;
static PyObject *__pyx_n_s_nodes;
static PyObject *__pyx_n_s_nthreads;
static PyObject *__pyx_n_u_nthreads;
static PyObject *__pyx_n_s_num_threads;
static PyObject *__pyx_kp_u_output;
static PyObject *__pyx_n_s_output_core;
static PyObject *__pyx_n_u_output_core;
static PyObject *__pyx_n_s_output_core_omp;
static PyObject *__pyx_n_u_output_core_omp;
static PyObject *__pyx_n_s_output_core_serial;
static PyObject *__pyx_n_u_output_core_serial;
static PyObject *__pyx_n_s_output_func;
static PyObject *__pyx_n_s_overall;
static PyObject *__pyx_n_u_overall_output;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_pinf_tot;
static PyObject *__pyx_n_s_pinf_tot_i;
static PyObject *__pyx_n_s_play;
static PyObject *__pyx_n_s_play_infections;
static PyObject *__pyx_n_s_play_infections_i;
static PyObject *__pyx_n_s_play_suscept;
static PyObject *__pyx_n_s_population;
static PyObject *__pyx_n_s_population_2;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_print_population;
static PyObject *__pyx_n_s_profiler;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_recovereds;
static PyObject *__pyx_n_s_redvar;
static PyObject *__pyx_n_s_safe_run;
static PyObject *__pyx_n_s_setup_core;
static PyObject *__pyx_n_u_setup_core;
static PyObject *__pyx_n_s_specialise;
static PyObject *__pyx_kp_s_src_metawards_extractors__output;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_subinfs;
static PyObject *__pyx_n_s_subnet;
static PyObject *__pyx_n_s_subnets;
static PyObject *__pyx_n_s_subpops;
static PyObject *__pyx_n_s_subspaces;
static PyObject *__pyx_n_s_suscept;
static PyObject *__pyx_n_s_susceptibles;
static PyObject *__pyx_n_s_susceptibles_i;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_thread_id;
static PyObject *__pyx_n_s_total;
static PyObject *__pyx_n_s_total_inf_ward;
static PyObject *__pyx_n_s_total_inf_ward_buffer;
static PyObject *__pyx_n_s_total_new;
static PyObject *__pyx_n_s_total_new_i;
static PyObject *__pyx_n_s_total_new_inf_ward;
static PyObject *__pyx_n_s_total_new_inf_ward_buffer;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_n_s_utils__console;
static PyObject *__pyx_n_s_utils__profiler;
static PyObject *__pyx_kp_u_vs;
static PyObject *__pyx_n_s_wards;
static PyObject *__pyx_n_s_work;
static PyObject *__pyx_n_s_workspace;
static PyObject *__pyx_n_s_workspace_2;
static PyObject *__pyx_n_s_zero_all;
static PyObject *__pyx_n_s_zero_subspaces;
static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_setup_core(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nthreads, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_2output_core_omp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, PyObject *__pyx_v_nthreads, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_4output_core_serial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_6_safe_run(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_8output_core(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, PyObject *__pyx_v_nthreads, PyObject *__pyx_v_profiler, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_4;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
static PyObject *__pyx_codeobj__11;
/* Late includes */

/* "metawards/extractors/_output_core.pyx":32
 * 
 * 
 * cdef _inf_buffer* _allocate_inf_buffers(int nthreads,             # <<<<<<<<<<<<<<
 *                                         int buffer_size=4096) nogil:
 *     cdef int size = buffer_size
 */

static struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(int __pyx_v_nthreads, struct __pyx_opt_args_9metawards_10extractors_12_output_core__allocate_inf_buffers *__pyx_optional_args) {
  int __pyx_v_buffer_size = ((int)0x1000);
  int __pyx_v_size;
  int __pyx_v_n;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_buffers;
  long __pyx_v_i;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_buffer_size = __pyx_optional_args->buffer_size;
    }
  }

  /* "metawards/extractors/_output_core.pyx":34
 * cdef _inf_buffer* _allocate_inf_buffers(int nthreads,
 *                                         int buffer_size=4096) nogil:
 *     cdef int size = buffer_size             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 * 
 */
  __pyx_v_size = __pyx_v_buffer_size;

  /* "metawards/extractors/_output_core.pyx":35
 *                                         int buffer_size=4096) nogil:
 *     cdef int size = buffer_size
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 * 
 *     cdef _inf_buffer *buffers = <_inf_buffer *> calloc(n, sizeof(_inf_buffer))
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":37
 *     cdef int n = nthreads
 * 
 *     cdef _inf_buffer *buffers = <_inf_buffer *> calloc(n, sizeof(_inf_buffer))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, n):
 */
  __pyx_v_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)calloc(__pyx_v_n, (sizeof(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer))));

  /* "metawards/extractors/_output_core.pyx":39
 *     cdef _inf_buffer *buffers = <_inf_buffer *> calloc(n, sizeof(_inf_buffer))
 * 
 *     for i in range(0, n):             # <<<<<<<<<<<<<<
 *         buffers[i].count = 0
 *         buffers[i].index = <int *> calloc(size, sizeof(int))
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "metawards/extractors/_output_core.pyx":40
 * 
 *     for i in range(0, n):
 *         buffers[i].count = 0             # <<<<<<<<<<<<<<
 *         buffers[i].index = <int *> calloc(size, sizeof(int))
 *         buffers[i].infected = <int *> calloc(size, sizeof(int))
 */
    (__pyx_v_buffers[__pyx_v_i]).count = 0;

    /* "metawards/extractors/_output_core.pyx":41
 *     for i in range(0, n):
 *         buffers[i].count = 0
 *         buffers[i].index = <int *> calloc(size, sizeof(int))             # <<<<<<<<<<<<<<
 *         buffers[i].infected = <int *> calloc(size, sizeof(int))
 * 
 */
    (__pyx_v_buffers[__pyx_v_i]).index = ((int *)calloc(__pyx_v_size, (sizeof(int))));

    /* "metawards/extractors/_output_core.pyx":42
 *         buffers[i].count = 0
 *         buffers[i].index = <int *> calloc(size, sizeof(int))
 *         buffers[i].infected = <int *> calloc(size, sizeof(int))             # <<<<<<<<<<<<<<
 * 
 *     return buffers
 */
    (__pyx_v_buffers[__pyx_v_i]).infected = ((int *)calloc(__pyx_v_size, (sizeof(int))));
  }

  /* "metawards/extractors/_output_core.pyx":44
 *         buffers[i].infected = <int *> calloc(size, sizeof(int))
 * 
 *     return buffers             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_buffers;
  goto __pyx_L0;

  /* "metawards/extractors/_output_core.pyx":32
 * 
 * 
 * cdef _inf_buffer* _allocate_inf_buffers(int nthreads,             # <<<<<<<<<<<<<<
 *                                         int buffer_size=4096) nogil:
 *     cdef int size = buffer_size
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":47
 * 
 * 
 * cdef void _free_inf_buffers(_inf_buffer *buffers, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 * 
 */

static void __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_buffers, int __pyx_v_nthreads) {
  int __pyx_v_n;
  long __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;

  /* "metawards/extractors/_output_core.pyx":48
 * 
 * cdef void _free_inf_buffers(_inf_buffer *buffers, int nthreads) nogil:
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, n):
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":50
 *     cdef int n = nthreads
 * 
 *     for i in range(0, n):             # <<<<<<<<<<<<<<
 *         free(buffers[i].index)
 *         free(buffers[i].infected)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "metawards/extractors/_output_core.pyx":51
 * 
 *     for i in range(0, n):
 *         free(buffers[i].index)             # <<<<<<<<<<<<<<
 *         free(buffers[i].infected)
 * 
 */
    free((__pyx_v_buffers[__pyx_v_i]).index);

    /* "metawards/extractors/_output_core.pyx":52
 *     for i in range(0, n):
 *         free(buffers[i].index)
 *         free(buffers[i].infected)             # <<<<<<<<<<<<<<
 * 
 *     free(buffers)
 */
    free((__pyx_v_buffers[__pyx_v_i]).infected);
  }

  /* "metawards/extractors/_output_core.pyx":54
 *         free(buffers[i].infected)
 * 
 *     free(buffers)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_buffers);

  /* "metawards/extractors/_output_core.pyx":47
 * 
 * 
 * cdef void _free_inf_buffers(_inf_buffer *buffers, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 * 
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":57
 * 
 * 
 * cdef void _reset_inf_buffer(_inf_buffer *buffers, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 *     cdef int i = 0
 */

static void __pyx_f_9metawards_10extractors_12_output_core__reset_inf_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_buffers, int __pyx_v_nthreads) {
  int __pyx_v_n;
  int __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "metawards/extractors/_output_core.pyx":58
 * 
 * cdef void _reset_inf_buffer(_inf_buffer *buffers, int nthreads) nogil:
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 *     cdef int i = 0
 * 
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":59
 * cdef void _reset_inf_buffer(_inf_buffer *buffers, int nthreads) nogil:
 *     cdef int n = nthreads
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, n):
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":61
 *     cdef int i = 0
 * 
 *     for i in range(0, n):             # <<<<<<<<<<<<<<
 *         buffers[i].count = 0
 * 
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "metawards/extractors/_output_core.pyx":62
 * 
 *     for i in range(0, n):
 *         buffers[i].count = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_buffers[__pyx_v_i]).count = 0;
  }

  /* "metawards/extractors/_output_core.pyx":57
 * 
 * 
 * cdef void _reset_inf_buffer(_inf_buffer *buffers, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 *     cdef int i = 0
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":65
 * 
 * 
 * cdef void _add_from_buffer(_inf_buffer *buffer, int *wards_infected) nogil:             # <<<<<<<<<<<<<<
 *     cdef int i = 0
 *     cdef int count = buffer[0].count
 */

static void __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_buffer, int *__pyx_v_wards_infected) {
  int __pyx_v_i;
  int __pyx_v_count;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "metawards/extractors/_output_core.pyx":66
 * 
 * cdef void _add_from_buffer(_inf_buffer *buffer, int *wards_infected) nogil:
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 *     cdef int count = buffer[0].count
 * 
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":67
 * cdef void _add_from_buffer(_inf_buffer *buffer, int *wards_infected) nogil:
 *     cdef int i = 0
 *     cdef int count = buffer[0].count             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, count):
 */
  __pyx_t_1 = (__pyx_v_buffer[0]).count;
  __pyx_v_count = __pyx_t_1;

  /* "metawards/extractors/_output_core.pyx":69
 *     cdef int count = buffer[0].count
 * 
 *     for i in range(0, count):             # <<<<<<<<<<<<<<
 *         wards_infected[buffer[0].index[i]] += buffer[0].infected[i]
 * 
 */
  __pyx_t_1 = __pyx_v_count;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "metawards/extractors/_output_core.pyx":70
 * 
 *     for i in range(0, count):
 *         wards_infected[buffer[0].index[i]] += buffer[0].infected[i]             # <<<<<<<<<<<<<<
 * 
 *     buffer[0].count = 0
 */
    __pyx_t_4 = ((__pyx_v_buffer[0]).index[__pyx_v_i]);
    (__pyx_v_wards_infected[__pyx_t_4]) = ((__pyx_v_wards_infected[__pyx_t_4]) + ((__pyx_v_buffer[0]).infected[__pyx_v_i]));
  }

  /* "metawards/extractors/_output_core.pyx":72
 *         wards_infected[buffer[0].index[i]] += buffer[0].infected[i]
 * 
 *     buffer[0].count = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_buffer[0]).count = 0;

  /* "metawards/extractors/_output_core.pyx":65
 * 
 * 
 * cdef void _add_from_buffer(_inf_buffer *buffer, int *wards_infected) nogil:             # <<<<<<<<<<<<<<
 *     cdef int i = 0
 *     cdef int count = buffer[0].count
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":75
 * 
 * 
 * cdef inline void _add_to_buffer(_inf_buffer *buffer, int index, int value,             # <<<<<<<<<<<<<<
 *                                 int *wards_infected,
 *                                 openmp.omp_lock_t *lock,
 */

static CYTHON_INLINE void __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_buffer, int __pyx_v_index, int __pyx_v_value, int *__pyx_v_wards_infected, omp_lock_t *__pyx_v_lock, struct __pyx_opt_args_9metawards_10extractors_12_output_core__add_to_buffer *__pyx_optional_args) {
  int __pyx_v_buffer_size = ((int)0x1000);
  int __pyx_v_count;
  int __pyx_t_1;
  int __pyx_t_2;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_buffer_size = __pyx_optional_args->buffer_size;
    }
  }

  /* "metawards/extractors/_output_core.pyx":79
 *                                 openmp.omp_lock_t *lock,
 *                                 int buffer_size=4096) nogil:
 *     cdef int count = buffer[0].count             # <<<<<<<<<<<<<<
 * 
 *     buffer[0].index[count] = index
 */
  __pyx_t_1 = (__pyx_v_buffer[0]).count;
  __pyx_v_count = __pyx_t_1;

  /* "metawards/extractors/_output_core.pyx":81
 *     cdef int count = buffer[0].count
 * 
 *     buffer[0].index[count] = index             # <<<<<<<<<<<<<<
 *     buffer[0].infected[count] = value
 *     buffer[0].count = count + 1
 */
  ((__pyx_v_buffer[0]).index[__pyx_v_count]) = __pyx_v_index;

  /* "metawards/extractors/_output_core.pyx":82
 * 
 *     buffer[0].index[count] = index
 *     buffer[0].infected[count] = value             # <<<<<<<<<<<<<<
 *     buffer[0].count = count + 1
 * 
 */
  ((__pyx_v_buffer[0]).infected[__pyx_v_count]) = __pyx_v_value;

  /* "metawards/extractors/_output_core.pyx":83
 *     buffer[0].index[count] = index
 *     buffer[0].infected[count] = value
 *     buffer[0].count = count + 1             # <<<<<<<<<<<<<<
 * 
 *     if buffer[0].count >= buffer_size:
 */
  (__pyx_v_buffer[0]).count = (__pyx_v_count + 1);

  /* "metawards/extractors/_output_core.pyx":85
 *     buffer[0].count = count + 1
 * 
 *     if buffer[0].count >= buffer_size:             # <<<<<<<<<<<<<<
 *         openmp.omp_set_lock(lock)
 *         _add_from_buffer(buffer, wards_infected)
 */
  __pyx_t_2 = (((__pyx_v_buffer[0]).count >= __pyx_v_buffer_size) != 0);
  if (__pyx_t_2) {

    /* "metawards/extractors/_output_core.pyx":86
 * 
 *     if buffer[0].count >= buffer_size:
 *         openmp.omp_set_lock(lock)             # <<<<<<<<<<<<<<
 *         _add_from_buffer(buffer, wards_infected)
 *         openmp.omp_unset_lock(lock)
 */
    omp_set_lock(__pyx_v_lock);

    /* "metawards/extractors/_output_core.pyx":87
 *     if buffer[0].count >= buffer_size:
 *         openmp.omp_set_lock(lock)
 *         _add_from_buffer(buffer, wards_infected)             # <<<<<<<<<<<<<<
 *         openmp.omp_unset_lock(lock)
 *         buffer[0].count = 0
 */
    __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_buffer, __pyx_v_wards_infected);

    /* "metawards/extractors/_output_core.pyx":88
 *         openmp.omp_set_lock(lock)
 *         _add_from_buffer(buffer, wards_infected)
 *         openmp.omp_unset_lock(lock)             # <<<<<<<<<<<<<<
 *         buffer[0].count = 0
 * 
 */
    omp_unset_lock(__pyx_v_lock);

    /* "metawards/extractors/_output_core.pyx":89
 *         _add_from_buffer(buffer, wards_infected)
 *         openmp.omp_unset_lock(lock)
 *         buffer[0].count = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_buffer[0]).count = 0;

    /* "metawards/extractors/_output_core.pyx":85
 *     buffer[0].count = count + 1
 * 
 *     if buffer[0].count >= buffer_size:             # <<<<<<<<<<<<<<
 *         openmp.omp_set_lock(lock)
 *         _add_from_buffer(buffer, wards_infected)
 */
  }

  /* "metawards/extractors/_output_core.pyx":75
 * 
 * 
 * cdef inline void _add_to_buffer(_inf_buffer *buffer, int index, int value,             # <<<<<<<<<<<<<<
 *                                 int *wards_infected,
 *                                 openmp.omp_lock_t *lock,
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":101
 * 
 * 
 * cdef void _reset_reduce_variables(_red_variables * variables,             # <<<<<<<<<<<<<<
 *                                   int nthreads) nogil:
 *     cdef int i = 0
 */

static void __pyx_f_9metawards_10extractors_12_output_core__reset_reduce_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_variables, int __pyx_v_nthreads) {
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "metawards/extractors/_output_core.pyx":103
 * cdef void _reset_reduce_variables(_red_variables * variables,
 *                                   int nthreads) nogil:
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 * 
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":104
 *                                   int nthreads) nogil:
 *     cdef int i = 0
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 * 
 *     for i in range(0, n):
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":106
 *     cdef int n = nthreads
 * 
 *     for i in range(0, n):             # <<<<<<<<<<<<<<
 *         variables[i].n_inf_wards = 0
 *         variables[i].total_new = 0
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "metawards/extractors/_output_core.pyx":107
 * 
 *     for i in range(0, n):
 *         variables[i].n_inf_wards = 0             # <<<<<<<<<<<<<<
 *         variables[i].total_new = 0
 *         variables[i].inf_tot = 0
 */
    (__pyx_v_variables[__pyx_v_i]).n_inf_wards = 0;

    /* "metawards/extractors/_output_core.pyx":108
 *     for i in range(0, n):
 *         variables[i].n_inf_wards = 0
 *         variables[i].total_new = 0             # <<<<<<<<<<<<<<
 *         variables[i].inf_tot = 0
 *         variables[i].pinf_tot = 0
 */
    (__pyx_v_variables[__pyx_v_i]).total_new = 0;

    /* "metawards/extractors/_output_core.pyx":109
 *         variables[i].n_inf_wards = 0
 *         variables[i].total_new = 0
 *         variables[i].inf_tot = 0             # <<<<<<<<<<<<<<
 *         variables[i].pinf_tot = 0
 *         variables[i].susceptibles = 0
 */
    (__pyx_v_variables[__pyx_v_i]).inf_tot = 0;

    /* "metawards/extractors/_output_core.pyx":110
 *         variables[i].total_new = 0
 *         variables[i].inf_tot = 0
 *         variables[i].pinf_tot = 0             # <<<<<<<<<<<<<<
 *         variables[i].susceptibles = 0
 * 
 */
    (__pyx_v_variables[__pyx_v_i]).pinf_tot = 0;

    /* "metawards/extractors/_output_core.pyx":111
 *         variables[i].inf_tot = 0
 *         variables[i].pinf_tot = 0
 *         variables[i].susceptibles = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_variables[__pyx_v_i]).susceptibles = 0;
  }

  /* "metawards/extractors/_output_core.pyx":101
 * 
 * 
 * cdef void _reset_reduce_variables(_red_variables * variables,             # <<<<<<<<<<<<<<
 *                                   int nthreads) nogil:
 *     cdef int i = 0
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":114
 * 
 * 
 * cdef _red_variables* _allocate_red_variables(int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     """Allocate space for thread-local reduction variables for each
 *        of the 'nthreads' threads. Note that we need this as
 */

static struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_f_9metawards_10extractors_12_output_core__allocate_red_variables(int __pyx_v_nthreads) {
  int __pyx_v_n;
  struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_variables;
  struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_r;

  /* "metawards/extractors/_output_core.pyx":119
 *        cython won't allow reduction in parallel blocks (only loops)
 *     """
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 * 
 *     cdef _red_variables *variables = <_red_variables *> calloc(
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":121
 *     cdef int n = nthreads
 * 
 *     cdef _red_variables *variables = <_red_variables *> calloc(             # <<<<<<<<<<<<<<
 *                                                 n, sizeof(_red_variables))
 * 
 */
  __pyx_v_variables = ((struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *)calloc(__pyx_v_n, (sizeof(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables))));

  /* "metawards/extractors/_output_core.pyx":125
 * 
 * 
 *     _reset_reduce_variables(variables, n)             # <<<<<<<<<<<<<<
 * 
 *     return variables
 */
  __pyx_f_9metawards_10extractors_12_output_core__reset_reduce_variables(__pyx_v_variables, __pyx_v_n);

  /* "metawards/extractors/_output_core.pyx":127
 *     _reset_reduce_variables(variables, n)
 * 
 *     return variables             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_variables;
  goto __pyx_L0;

  /* "metawards/extractors/_output_core.pyx":114
 * 
 * 
 * cdef _red_variables* _allocate_red_variables(int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     """Allocate space for thread-local reduction variables for each
 *        of the 'nthreads' threads. Note that we need this as
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":130
 * 
 * 
 * cdef void _free_red_variables(_red_variables *variables) nogil:             # <<<<<<<<<<<<<<
 *     """Free the memory held by the reduction variables"""
 *     free(variables)
 */

static void __pyx_f_9metawards_10extractors_12_output_core__free_red_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_variables) {

  /* "metawards/extractors/_output_core.pyx":132
 * cdef void _free_red_variables(_red_variables *variables) nogil:
 *     """Free the memory held by the reduction variables"""
 *     free(variables)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_variables);

  /* "metawards/extractors/_output_core.pyx":130
 * 
 * 
 * cdef void _free_red_variables(_red_variables *variables) nogil:             # <<<<<<<<<<<<<<
 *     """Free the memory held by the reduction variables"""
 *     free(variables)
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":135
 * 
 * 
 * cdef void _reduce_variables(_red_variables *variables, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     """Reduce all of the thread-local totals in 'variables' for
 *        the 'nthreads' threads into the values in variables[0]
 */

static void __pyx_f_9metawards_10extractors_12_output_core__reduce_variables(struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_variables, int __pyx_v_nthreads) {
  int __pyx_v_i;
  int __pyx_v_n;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":139
 *        the 'nthreads' threads into the values in variables[0]
 *     """
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 *     cdef int n = nthreads
 * 
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":140
 *     """
 *     cdef int i = 0
 *     cdef int n = nthreads             # <<<<<<<<<<<<<<
 * 
 *     if n > 1:
 */
  __pyx_v_n = __pyx_v_nthreads;

  /* "metawards/extractors/_output_core.pyx":142
 *     cdef int n = nthreads
 * 
 *     if n > 1:             # <<<<<<<<<<<<<<
 *         for i in range(1, n):
 *             variables[0].n_inf_wards += variables[i].n_inf_wards
 */
  __pyx_t_1 = ((__pyx_v_n > 1) != 0);
  if (__pyx_t_1) {

    /* "metawards/extractors/_output_core.pyx":143
 * 
 *     if n > 1:
 *         for i in range(1, n):             # <<<<<<<<<<<<<<
 *             variables[0].n_inf_wards += variables[i].n_inf_wards
 *             variables[0].total_new += variables[i].total_new
 */
    __pyx_t_2 = __pyx_v_n;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 1; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "metawards/extractors/_output_core.pyx":144
 *     if n > 1:
 *         for i in range(1, n):
 *             variables[0].n_inf_wards += variables[i].n_inf_wards             # <<<<<<<<<<<<<<
 *             variables[0].total_new += variables[i].total_new
 *             variables[0].inf_tot += variables[i].inf_tot
 */
      __pyx_t_5 = 0;
      (__pyx_v_variables[__pyx_t_5]).n_inf_wards = ((__pyx_v_variables[__pyx_t_5]).n_inf_wards + (__pyx_v_variables[__pyx_v_i]).n_inf_wards);

      /* "metawards/extractors/_output_core.pyx":145
 *         for i in range(1, n):
 *             variables[0].n_inf_wards += variables[i].n_inf_wards
 *             variables[0].total_new += variables[i].total_new             # <<<<<<<<<<<<<<
 *             variables[0].inf_tot += variables[i].inf_tot
 *             variables[0].pinf_tot += variables[i].pinf_tot
 */
      __pyx_t_5 = 0;
      (__pyx_v_variables[__pyx_t_5]).total_new = ((__pyx_v_variables[__pyx_t_5]).total_new + (__pyx_v_variables[__pyx_v_i]).total_new);

      /* "metawards/extractors/_output_core.pyx":146
 *             variables[0].n_inf_wards += variables[i].n_inf_wards
 *             variables[0].total_new += variables[i].total_new
 *             variables[0].inf_tot += variables[i].inf_tot             # <<<<<<<<<<<<<<
 *             variables[0].pinf_tot += variables[i].pinf_tot
 *             variables[0].susceptibles += variables[i].susceptibles
 */
      __pyx_t_5 = 0;
      (__pyx_v_variables[__pyx_t_5]).inf_tot = ((__pyx_v_variables[__pyx_t_5]).inf_tot + (__pyx_v_variables[__pyx_v_i]).inf_tot);

      /* "metawards/extractors/_output_core.pyx":147
 *             variables[0].total_new += variables[i].total_new
 *             variables[0].inf_tot += variables[i].inf_tot
 *             variables[0].pinf_tot += variables[i].pinf_tot             # <<<<<<<<<<<<<<
 *             variables[0].susceptibles += variables[i].susceptibles
 * 
 */
      __pyx_t_5 = 0;
      (__pyx_v_variables[__pyx_t_5]).pinf_tot = ((__pyx_v_variables[__pyx_t_5]).pinf_tot + (__pyx_v_variables[__pyx_v_i]).pinf_tot);

      /* "metawards/extractors/_output_core.pyx":148
 *             variables[0].inf_tot += variables[i].inf_tot
 *             variables[0].pinf_tot += variables[i].pinf_tot
 *             variables[0].susceptibles += variables[i].susceptibles             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = 0;
      (__pyx_v_variables[__pyx_t_5]).susceptibles = ((__pyx_v_variables[__pyx_t_5]).susceptibles + (__pyx_v_variables[__pyx_v_i]).susceptibles);
    }

    /* "metawards/extractors/_output_core.pyx":142
 *     cdef int n = nthreads
 * 
 *     if n > 1:             # <<<<<<<<<<<<<<
 *         for i in range(1, n):
 *             variables[0].n_inf_wards += variables[i].n_inf_wards
 */
  }

  /* "metawards/extractors/_output_core.pyx":135
 * 
 * 
 * cdef void _reduce_variables(_red_variables *variables, int nthreads) nogil:             # <<<<<<<<<<<<<<
 *     """Reduce all of the thread-local totals in 'variables' for
 *        the 'nthreads' threads into the values in variables[0]
 */

  /* function exit code */
}

/* "metawards/extractors/_output_core.pyx":167
 * 
 * 
 * def setup_core(nthreads: int = 1, **kwargs):             # <<<<<<<<<<<<<<
 *     """This is the setup function that corresponds with
 *        :meth:`~metawards.extractors.output_core`.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_1setup_core(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9metawards_10extractors_12_output_core_setup_core[] = "setup_core(nthreads: int = 1, **kwargs)\nThis is the setup function that corresponds with\n       :meth:`~metawards.extractors.output_core`.\n    ";
static PyMethodDef __pyx_mdef_9metawards_10extractors_12_output_core_1setup_core = {"setup_core", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9metawards_10extractors_12_output_core_1setup_core, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9metawards_10extractors_12_output_core_setup_core};
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_1setup_core(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nthreads = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_core (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nthreads,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_1);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nthreads);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "setup_core") < 0)) __PYX_ERR(0, 167, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nthreads = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup_core", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 167, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("metawards.extractors._output_core.setup_core", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9metawards_10extractors_12_output_core_setup_core(__pyx_self, __pyx_v_nthreads, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_setup_core(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_nthreads, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup_core", 0);

  /* "metawards/extractors/_output_core.pyx":173
 *     global _buffer_nthreads
 * 
 *     if _buffer_nthreads != nthreads:             # <<<<<<<<<<<<<<
 *         global _total_new_inf_ward_buffers
 *         global _total_inf_ward_buffers
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_9metawards_10extractors_12_output_core__buffer_nthreads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_v_nthreads, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "metawards/extractors/_output_core.pyx":179
 *         global _redvars
 * 
 *         if _buffer_nthreads != 0:             # <<<<<<<<<<<<<<
 *             _free_red_variables(_redvars)
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)
 */
    __pyx_t_3 = ((__pyx_v_9metawards_10extractors_12_output_core__buffer_nthreads != 0) != 0);
    if (__pyx_t_3) {

      /* "metawards/extractors/_output_core.pyx":180
 * 
 *         if _buffer_nthreads != 0:
 *             _free_red_variables(_redvars)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_total_inf_ward_buffers, nthreads)
 */
      __pyx_f_9metawards_10extractors_12_output_core__free_red_variables(__pyx_v_9metawards_10extractors_12_output_core__redvars);

      /* "metawards/extractors/_output_core.pyx":181
 *         if _buffer_nthreads != 0:
 *             _free_red_variables(_redvars)
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_total_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_S_buffers, nthreads)
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":182
 *             _free_red_variables(_redvars)
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_total_inf_ward_buffers, nthreads)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_S_buffers, nthreads)
 *             _free_inf_buffers(_E_buffers, nthreads)
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":183
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_total_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_S_buffers, nthreads)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_E_buffers, nthreads)
 *             _free_inf_buffers(_I_buffers, nthreads)
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__S_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":184
 *             _free_inf_buffers(_total_inf_ward_buffers, nthreads)
 *             _free_inf_buffers(_S_buffers, nthreads)
 *             _free_inf_buffers(_E_buffers, nthreads)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_I_buffers, nthreads)
 *             _free_inf_buffers(_R_buffers, nthreads)
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__E_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":185
 *             _free_inf_buffers(_S_buffers, nthreads)
 *             _free_inf_buffers(_E_buffers, nthreads)
 *             _free_inf_buffers(_I_buffers, nthreads)             # <<<<<<<<<<<<<<
 *             _free_inf_buffers(_R_buffers, nthreads)
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__I_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":186
 *             _free_inf_buffers(_E_buffers, nthreads)
 *             _free_inf_buffers(_I_buffers, nthreads)
 *             _free_inf_buffers(_R_buffers, nthreads)             # <<<<<<<<<<<<<<
 * 
 *         _redvars = _allocate_red_variables(nthreads)
 */
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 186, __pyx_L1_error)
      __pyx_f_9metawards_10extractors_12_output_core__free_inf_buffers(__pyx_v_9metawards_10extractors_12_output_core__R_buffers, __pyx_t_4);

      /* "metawards/extractors/_output_core.pyx":179
 *         global _redvars
 * 
 *         if _buffer_nthreads != 0:             # <<<<<<<<<<<<<<
 *             _free_red_variables(_redvars)
 *             _free_inf_buffers(_total_new_inf_ward_buffers, nthreads)
 */
    }

    /* "metawards/extractors/_output_core.pyx":188
 *             _free_inf_buffers(_R_buffers, nthreads)
 * 
 *         _redvars = _allocate_red_variables(nthreads)             # <<<<<<<<<<<<<<
 *         _total_new_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _total_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__redvars = __pyx_f_9metawards_10extractors_12_output_core__allocate_red_variables(__pyx_t_4);

    /* "metawards/extractors/_output_core.pyx":189
 * 
 *         _redvars = _allocate_red_variables(nthreads)
 *         _total_new_inf_ward_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _total_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _S_buffers = _allocate_inf_buffers(nthreads)
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":190
 *         _redvars = _allocate_red_variables(nthreads)
 *         _total_new_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _total_inf_ward_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _S_buffers = _allocate_inf_buffers(nthreads)
 *         _E_buffers = _allocate_inf_buffers(nthreads)
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 190, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":191
 *         _total_new_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _total_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _S_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _E_buffers = _allocate_inf_buffers(nthreads)
 *         _I_buffers = _allocate_inf_buffers(nthreads)
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__S_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":192
 *         _total_inf_ward_buffers = _allocate_inf_buffers(nthreads)
 *         _S_buffers = _allocate_inf_buffers(nthreads)
 *         _E_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _I_buffers = _allocate_inf_buffers(nthreads)
 *         _R_buffers = _allocate_inf_buffers(nthreads)
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 192, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__E_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":193
 *         _S_buffers = _allocate_inf_buffers(nthreads)
 *         _E_buffers = _allocate_inf_buffers(nthreads)
 *         _I_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _R_buffers = _allocate_inf_buffers(nthreads)
 *         _buffer_nthreads = nthreads
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 193, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__I_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":194
 *         _E_buffers = _allocate_inf_buffers(nthreads)
 *         _I_buffers = _allocate_inf_buffers(nthreads)
 *         _R_buffers = _allocate_inf_buffers(nthreads)             # <<<<<<<<<<<<<<
 *         _buffer_nthreads = nthreads
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__R_buffers = __pyx_f_9metawards_10extractors_12_output_core__allocate_inf_buffers(__pyx_t_4, NULL);

    /* "metawards/extractors/_output_core.pyx":195
 *         _I_buffers = _allocate_inf_buffers(nthreads)
 *         _R_buffers = _allocate_inf_buffers(nthreads)
 *         _buffer_nthreads = nthreads             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
    __pyx_v_9metawards_10extractors_12_output_core__buffer_nthreads = __pyx_t_4;

    /* "metawards/extractors/_output_core.pyx":173
 *     global _buffer_nthreads
 * 
 *     if _buffer_nthreads != nthreads:             # <<<<<<<<<<<<<<
 *         global _total_new_inf_ward_buffers
 *         global _total_inf_ward_buffers
 */
  }

  /* "metawards/extractors/_output_core.pyx":167
 * 
 * 
 * def setup_core(nthreads: int = 1, **kwargs):             # <<<<<<<<<<<<<<
 *     """This is the setup function that corresponds with
 *        :meth:`~metawards.extractors.output_core`.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("metawards.extractors._output_core.setup_core", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":198
 * 
 * 
 * def output_core_omp(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                     workspace: Workspace,
 *                     infections: Infections,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_3output_core_omp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9metawards_10extractors_12_output_core_2output_core_omp[] = "output_core_omp(network: Network, population: Population, workspace: Workspace, infections: Infections, nthreads: int, **kwargs)\nThis is the core output function that must be called\n       every iteration as it is responsible for accumulating\n       the core data each day, which is used to report a summary\n       to the main output file. This is the parallel version\n       of this function.\n\n       Parameters\n       ----------\n       network: Network\n         The network over which the outbreak is being modelled\n       population: Population\n         The population experiencing the outbreak\n       workspace: Workspace\n         A workspace that can be used to extract data\n       infections: Infections\n         All of the infections that have been recorded\n       nthreads: int\n         The number of threads to use to help extract the data\n       kwargs\n         Extra argumentst that are ignored by this function\n    ";
static PyMethodDef __pyx_mdef_9metawards_10extractors_12_output_core_3output_core_omp = {"output_core_omp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9metawards_10extractors_12_output_core_3output_core_omp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9metawards_10extractors_12_output_core_2output_core_omp};
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_3output_core_omp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_network = 0;
  PyObject *__pyx_v_population = 0;
  PyObject *__pyx_v_workspace = 0;
  PyObject *__pyx_v_infections = 0;
  PyObject *__pyx_v_nthreads = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output_core_omp (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_network,&__pyx_n_s_population,&__pyx_n_s_workspace,&__pyx_n_s_infections,&__pyx_n_s_nthreads,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_network)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_population)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_omp", 1, 5, 5, 1); __PYX_ERR(0, 198, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_omp", 1, 5, 5, 2); __PYX_ERR(0, 198, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_infections)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_omp", 1, 5, 5, 3); __PYX_ERR(0, 198, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nthreads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_omp", 1, 5, 5, 4); __PYX_ERR(0, 198, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "output_core_omp") < 0)) __PYX_ERR(0, 198, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_network = values[0];
    __pyx_v_population = values[1];
    __pyx_v_workspace = values[2];
    __pyx_v_infections = values[3];
    __pyx_v_nthreads = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("output_core_omp", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 198, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core_omp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9metawards_10extractors_12_output_core_2output_core_omp(__pyx_self, __pyx_v_network, __pyx_v_population, __pyx_v_workspace, __pyx_v_infections, __pyx_v_nthreads, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_2output_core_omp(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, PyObject *__pyx_v_nthreads, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_links = NULL;
  PyObject *__pyx_v_wards = NULL;
  PyObject *__pyx_v_play_infections = NULL;
  int __pyx_v_N_INF_CLASSES;
  int *__pyx_v_inf_tot;
  int *__pyx_v_pinf_tot;
  int *__pyx_v_total_inf_ward;
  int *__pyx_v_total_new_inf_ward;
  int *__pyx_v_n_inf_wards;
  int *__pyx_v_incidence;
  int *__pyx_v_S_in_wards;
  int *__pyx_v_E_in_wards;
  int *__pyx_v_I_in_wards;
  int *__pyx_v_R_in_wards;
  int *__pyx_v_links_ifrom;
  double *__pyx_v_links_suscept;
  double *__pyx_v_play_suscept;
  CYTHON_UNUSED int __pyx_v_nlinks_plus_one;
  int __pyx_v_nnodes_plus_one;
  int *__pyx_v_infections_i;
  int *__pyx_v_play_infections_i;
  int __pyx_v_total_new;
  int __pyx_v_susceptibles;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_num_threads;
  int __pyx_v_thread_id;
  int __pyx_v_ifrom;
  int __pyx_v_N_INF_CLASSES_MINUS_ONE;
  omp_lock_t __pyx_v_lock;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_total_new_inf_ward_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_total_inf_ward_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_S_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_E_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_I_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *__pyx_v_R_buffer;
  struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *__pyx_v_redvar;
  PyObject *__pyx_v_latent = NULL;
  PyObject *__pyx_v_total = NULL;
  int __pyx_v_recovereds;
  int __pyx_v_S;
  int __pyx_v_E;
  int __pyx_v_I;
  int __pyx_v_R;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_Console = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  long __pyx_t_11;
  int __pyx_t_12;
  long __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("output_core_omp", 0);
  __Pyx_INCREF(__pyx_v_infections);

  /* "metawards/extractors/_output_core.pyx":225
 *     from cython.parallel import parallel, prange
 * 
 *     links = network.links             # <<<<<<<<<<<<<<
 *     wards = network.nodes
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_links); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":226
 * 
 *     links = network.links
 *     wards = network.nodes             # <<<<<<<<<<<<<<
 * 
 *     # set up main variables
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wards = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":229
 * 
 *     # set up main variables
 *     play_infections = infections.play             # <<<<<<<<<<<<<<
 *     infections = infections.work
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_play); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_play_infections = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":230
 *     # set up main variables
 *     play_infections = infections.play
 *     infections = infections.work             # <<<<<<<<<<<<<<
 * 
 *     cdef int N_INF_CLASSES = len(infections)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_work); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_infections, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":232
 *     infections = infections.work
 * 
 *     cdef int N_INF_CLASSES = len(infections)             # <<<<<<<<<<<<<<
 *     assert len(infections) == len(play_infections)
 * 
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_infections); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_v_N_INF_CLASSES = __pyx_t_2;

  /* "metawards/extractors/_output_core.pyx":233
 * 
 *     cdef int N_INF_CLASSES = len(infections)
 *     assert len(infections) == len(play_infections)             # <<<<<<<<<<<<<<
 * 
 *     global _total_inf_ward_buffers
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = PyObject_Length(__pyx_v_infections); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 233, __pyx_L1_error)
    __pyx_t_3 = PyObject_Length(__pyx_v_play_infections); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 233, __pyx_L1_error)
    if (unlikely(!((__pyx_t_2 == __pyx_t_3) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 233, __pyx_L1_error)
    }
  }
  #endif

  /* "metawards/extractors/_output_core.pyx":240
 * 
 *     # get pointers to arrays in workspace to write data
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)             # <<<<<<<<<<<<<<
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_inf_tot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inf_tot = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":241
 *     # get pointers to arrays in workspace to write data
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)             # <<<<<<<<<<<<<<
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_pinf_tot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pinf_tot = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":242
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)             # <<<<<<<<<<<<<<
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_total_inf_ward); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_total_inf_ward = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":244
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)             # <<<<<<<<<<<<<<
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_total_new_inf_ward); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "metawards/extractors/_output_core.pyx":243
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(             # <<<<<<<<<<<<<<
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 */
  __pyx_v_total_new_inf_ward = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":245
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)             # <<<<<<<<<<<<<<
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_n_inf_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_n_inf_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":246
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)             # <<<<<<<<<<<<<<
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_incidence); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_incidence = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":248
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_S_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_S_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":249
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_E_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_E_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":250
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_I_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_I_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":251
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)             # <<<<<<<<<<<<<<
 * 
 *     # get pointers to arrays from links and play to read data
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_R_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_R_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":254
 * 
 *     # get pointers to arrays from links and play to read data
 *     cdef int * links_ifrom = get_int_array_ptr(links.ifrom)             # <<<<<<<<<<<<<<
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_links, __pyx_n_s_ifrom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links_ifrom = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":256
 *     cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)             # <<<<<<<<<<<<<<
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_links, __pyx_n_s_suscept); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links_suscept = __pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":257
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)             # <<<<<<<<<<<<<<
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wards, __pyx_n_s_play_suscept); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_play_suscept = __pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":259
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1             # <<<<<<<<<<<<<<
 *     cdef int nnodes_plus_one = network.nnodes + 1
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nlinks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_nlinks_plus_one = __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":260
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1
 *     cdef int nnodes_plus_one = network.nnodes + 1             # <<<<<<<<<<<<<<
 * 
 *     # Now pointers into the infections and play_infections arrays
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nnodes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nnodes_plus_one = __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":268
 * 
 *     # Now some loop variables used for accumulation
 *     cdef int total_new = 0             # <<<<<<<<<<<<<<
 *     cdef int susceptibles = 0
 * 
 */
  __pyx_v_total_new = 0;

  /* "metawards/extractors/_output_core.pyx":269
 *     # Now some loop variables used for accumulation
 *     cdef int total_new = 0
 *     cdef int susceptibles = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int i = 0
 */
  __pyx_v_susceptibles = 0;

  /* "metawards/extractors/_output_core.pyx":271
 *     cdef int susceptibles = 0
 * 
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 *     cdef int j = 0
 * 
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":272
 * 
 *     cdef int i = 0
 *     cdef int j = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int num_threads = nthreads
 */
  __pyx_v_j = 0;

  /* "metawards/extractors/_output_core.pyx":274
 *     cdef int j = 0
 * 
 *     cdef int num_threads = nthreads             # <<<<<<<<<<<<<<
 *     cdef int thread_id = 0
 *     cdef int ifrom = 0
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nthreads); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)
  __pyx_v_num_threads = __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":275
 * 
 *     cdef int num_threads = nthreads
 *     cdef int thread_id = 0             # <<<<<<<<<<<<<<
 *     cdef int ifrom = 0
 * 
 */
  __pyx_v_thread_id = 0;

  /* "metawards/extractors/_output_core.pyx":276
 *     cdef int num_threads = nthreads
 *     cdef int thread_id = 0
 *     cdef int ifrom = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int N_INF_CLASSES_MINUS_ONE = N_INF_CLASSES - 1
 */
  __pyx_v_ifrom = 0;

  /* "metawards/extractors/_output_core.pyx":278
 *     cdef int ifrom = 0
 * 
 *     cdef int N_INF_CLASSES_MINUS_ONE = N_INF_CLASSES - 1             # <<<<<<<<<<<<<<
 * 
 *     # Finally some variables used to control parallelisation and
 */
  __pyx_v_N_INF_CLASSES_MINUS_ONE = (__pyx_v_N_INF_CLASSES - 1);

  /* "metawards/extractors/_output_core.pyx":283
 *     # some reduction buffers
 *     cdef openmp.omp_lock_t lock
 *     openmp.omp_init_lock(&lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef _inf_buffer * total_new_inf_ward_buffer
 */
  omp_init_lock((&__pyx_v_lock));

  /* "metawards/extractors/_output_core.pyx":299
 * 
 *     # reset the workspace so that we can accumulate new data for a new day
 *     workspace.zero_all()             # <<<<<<<<<<<<<<
 * 
 *     # loop over each of the disease stages
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_zero_all); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":302
 * 
 *     # loop over each of the disease stages
 *     for i in range(0, N_INF_CLASSES):             # <<<<<<<<<<<<<<
 *         # zero the number of infected wards, and the total number
 *         # of infections for this stage
 */
  __pyx_t_5 = __pyx_v_N_INF_CLASSES;
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "metawards/extractors/_output_core.pyx":305
 *         # zero the number of infected wards, and the total number
 *         # of infections for this stage
 *         n_inf_wards[i] = 0             # <<<<<<<<<<<<<<
 *         inf_tot[i] = 0
 *         pinf_tot[i] = 0
 */
    (__pyx_v_n_inf_wards[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":306
 *         # of infections for this stage
 *         n_inf_wards[i] = 0
 *         inf_tot[i] = 0             # <<<<<<<<<<<<<<
 *         pinf_tot[i] = 0
 * 
 */
    (__pyx_v_inf_tot[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":307
 *         n_inf_wards[i] = 0
 *         inf_tot[i] = 0
 *         pinf_tot[i] = 0             # <<<<<<<<<<<<<<
 * 
 *         # similarly zero the number of infections per ward and
 */
    (__pyx_v_pinf_tot[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":311
 *         # similarly zero the number of infections per ward and
 *         # number of new infections per ward
 *         _reset_reduce_variables(_redvars, num_threads)             # <<<<<<<<<<<<<<
 *         _reset_inf_buffer(_total_inf_ward_buffers, num_threads)
 *         _reset_inf_buffer(_total_new_inf_ward_buffers, num_threads)
 */
    __pyx_f_9metawards_10extractors_12_output_core__reset_reduce_variables(__pyx_v_9metawards_10extractors_12_output_core__redvars, __pyx_v_num_threads);

    /* "metawards/extractors/_output_core.pyx":312
 *         # number of new infections per ward
 *         _reset_reduce_variables(_redvars, num_threads)
 *         _reset_inf_buffer(_total_inf_ward_buffers, num_threads)             # <<<<<<<<<<<<<<
 *         _reset_inf_buffer(_total_new_inf_ward_buffers, num_threads)
 * 
 */
    __pyx_f_9metawards_10extractors_12_output_core__reset_inf_buffer(__pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers, __pyx_v_num_threads);

    /* "metawards/extractors/_output_core.pyx":313
 *         _reset_reduce_variables(_redvars, num_threads)
 *         _reset_inf_buffer(_total_inf_ward_buffers, num_threads)
 *         _reset_inf_buffer(_total_new_inf_ward_buffers, num_threads)             # <<<<<<<<<<<<<<
 * 
 *         # get pointers to the work and play infections for this disease
 */
    __pyx_f_9metawards_10extractors_12_output_core__reset_inf_buffer(__pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers, __pyx_v_num_threads);

    /* "metawards/extractors/_output_core.pyx":317
 *         # get pointers to the work and play infections for this disease
 *         # stage
 *         infections_i = get_int_array_ptr(infections[i])             # <<<<<<<<<<<<<<
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_infections, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_infections_i = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":318
 *         # stage
 *         infections_i = get_int_array_ptr(infections[i])
 *         play_infections_i = get_int_array_ptr(play_infections[i])             # <<<<<<<<<<<<<<
 * 
 *         with nogil, parallel(num_threads=num_threads):
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_play_infections, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_play_infections_i = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":320
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 *         with nogil, parallel(num_threads=num_threads):             # <<<<<<<<<<<<<<
 *             # get the space for this thread to accumulate data.
 *             # All of this data will be reduced in critical sections
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {
          {
              #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
                  #undef likely
                  #undef unlikely
                  #define likely(x)   (x)
                  #define unlikely(x) (x)
              #endif
              #ifdef _OPENMP
              #pragma omp parallel private(__pyx_v_E_buffer, __pyx_v_I_buffer, __pyx_v_R_buffer, __pyx_v_S_buffer, __pyx_v_redvar, __pyx_v_thread_id, __pyx_v_total_inf_ward_buffer, __pyx_v_total_new_inf_ward_buffer) private(__pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_15, __pyx_t_9) num_threads(__pyx_v_num_threads)
              #endif /* _OPENMP */
              {
                  /* Initialize private variables to invalid values */
                  __pyx_v_E_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);
                  __pyx_v_I_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);
                  __pyx_v_R_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);
                  __pyx_v_S_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);
                  __pyx_v_redvar = ((struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *)1);
                  __pyx_v_thread_id = ((int)0xbad0bad0);
                  __pyx_v_total_inf_ward_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);
                  __pyx_v_total_new_inf_ward_buffer = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)1);

                  /* "metawards/extractors/_output_core.pyx":323
 *             # get the space for this thread to accumulate data.
 *             # All of this data will be reduced in critical sections
 *             thread_id = cython.parallel.threadid()             # <<<<<<<<<<<<<<
 *             total_inf_ward_buffer = &(_total_inf_ward_buffers[thread_id])
 *             total_new_inf_ward_buffer = \
 */
                  #ifdef _OPENMP
                  __pyx_t_9 = omp_get_thread_num();
                  #else
                  __pyx_t_9 = 0;
                  #endif
                  __pyx_v_thread_id = __pyx_t_9;

                  /* "metawards/extractors/_output_core.pyx":324
 *             # All of this data will be reduced in critical sections
 *             thread_id = cython.parallel.threadid()
 *             total_inf_ward_buffer = &(_total_inf_ward_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             total_new_inf_ward_buffer = \
 *                                 &(_total_new_inf_ward_buffers[thread_id])
 */
                  __pyx_v_total_inf_ward_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":326
 *             total_inf_ward_buffer = &(_total_inf_ward_buffers[thread_id])
 *             total_new_inf_ward_buffer = \
 *                                 &(_total_new_inf_ward_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             S_buffer = &(_S_buffers[thread_id])
 *             E_buffer = &(_E_buffers[thread_id])
 */
                  __pyx_v_total_new_inf_ward_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":327
 *             total_new_inf_ward_buffer = \
 *                                 &(_total_new_inf_ward_buffers[thread_id])
 *             S_buffer = &(_S_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             E_buffer = &(_E_buffers[thread_id])
 *             I_buffer = &(_I_buffers[thread_id])
 */
                  __pyx_v_S_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__S_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":328
 *                                 &(_total_new_inf_ward_buffers[thread_id])
 *             S_buffer = &(_S_buffers[thread_id])
 *             E_buffer = &(_E_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             I_buffer = &(_I_buffers[thread_id])
 *             R_buffer = &(_R_buffers[thread_id])
 */
                  __pyx_v_E_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__E_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":329
 *             S_buffer = &(_S_buffers[thread_id])
 *             E_buffer = &(_E_buffers[thread_id])
 *             I_buffer = &(_I_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             R_buffer = &(_R_buffers[thread_id])
 *             redvar = &(_redvars[thread_id])
 */
                  __pyx_v_I_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__I_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":330
 *             E_buffer = &(_E_buffers[thread_id])
 *             I_buffer = &(_I_buffers[thread_id])
 *             R_buffer = &(_R_buffers[thread_id])             # <<<<<<<<<<<<<<
 *             redvar = &(_redvars[thread_id])
 * 
 */
                  __pyx_v_R_buffer = (&(__pyx_v_9metawards_10extractors_12_output_core__R_buffers[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":331
 *             I_buffer = &(_I_buffers[thread_id])
 *             R_buffer = &(_R_buffers[thread_id])
 *             redvar = &(_redvars[thread_id])             # <<<<<<<<<<<<<<
 * 
 *             #loop over all links and accumulate infections associated
 */
                  __pyx_v_redvar = (&(__pyx_v_9metawards_10extractors_12_output_core__redvars[__pyx_v_thread_id]));

                  /* "metawards/extractors/_output_core.pyx":335
 *             #loop over all links and accumulate infections associated
 *             # with this link
 *             for j in prange(1, nlinks_plus_one, schedule="static"):             # <<<<<<<<<<<<<<
 *                 ifrom = links_ifrom[j]
 * 
 */
                  __pyx_t_9 = __pyx_v_nlinks_plus_one;
                  if ((1 == 0)) abort();
                  {
                      __pyx_t_11 = (__pyx_t_9 - 1 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_11 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp for lastprivate(__pyx_v_ifrom) firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(static)
                          #endif /* _OPENMP */
                          for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_11; __pyx_t_10++){
                              {
                                  __pyx_v_j = (int)(1 + 1 * __pyx_t_10);
                                  /* Initialize private variables to invalid values */
                                  __pyx_v_ifrom = ((int)0xbad0bad0);

                                  /* "metawards/extractors/_output_core.pyx":336
 *             # with this link
 *             for j in prange(1, nlinks_plus_one, schedule="static"):
 *                 ifrom = links_ifrom[j]             # <<<<<<<<<<<<<<
 * 
 *                 if i == 0:
 */
                                  __pyx_v_ifrom = (__pyx_v_links_ifrom[__pyx_v_j]);

                                  /* "metawards/extractors/_output_core.pyx":338
 *                 ifrom = links_ifrom[j]
 * 
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += links[j].suscept
 *                     redvar[0].susceptibles += <int>(links_suscept[j])
 */
                                  __pyx_t_12 = ((__pyx_v_i == 0) != 0);
                                  if (__pyx_t_12) {

                                    /* "metawards/extractors/_output_core.pyx":340
 *                 if i == 0:
 *                     # susceptibles += links[j].suscept
 *                     redvar[0].susceptibles += <int>(links_suscept[j])             # <<<<<<<<<<<<<<
 *                     _add_to_buffer(S_buffer, ifrom, <int>(links_suscept[j]),
 *                                    &(S_in_wards[0]), &lock)
 */
                                    __pyx_t_13 = 0;
                                    (__pyx_v_redvar[__pyx_t_13]).susceptibles = ((__pyx_v_redvar[__pyx_t_13]).susceptibles + ((int)(__pyx_v_links_suscept[__pyx_v_j])));

                                    /* "metawards/extractors/_output_core.pyx":341
 *                     # susceptibles += links[j].suscept
 *                     redvar[0].susceptibles += <int>(links_suscept[j])
 *                     _add_to_buffer(S_buffer, ifrom, <int>(links_suscept[j]),             # <<<<<<<<<<<<<<
 *                                    &(S_in_wards[0]), &lock)
 * 
 */
                                    __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_S_buffer, __pyx_v_ifrom, ((int)(__pyx_v_links_suscept[__pyx_v_j])), (&(__pyx_v_S_in_wards[0])), (&__pyx_v_lock), NULL);

                                    /* "metawards/extractors/_output_core.pyx":344
 *                                    &(S_in_wards[0]), &lock)
 * 
 *                     if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         _add_to_buffer(total_new_inf_ward_buffer,
 */
                                    __pyx_t_12 = (((__pyx_v_infections_i[__pyx_v_j]) != 0) != 0);
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":346
 *                     if infections_i[j] != 0:
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         _add_to_buffer(total_new_inf_ward_buffer,             # <<<<<<<<<<<<<<
 *                                        ifrom, infections_i[j],
 *                                        &(total_new_inf_ward[0]), &lock)
 */
                                      __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_total_new_inf_ward_buffer, __pyx_v_ifrom, (__pyx_v_infections_i[__pyx_v_j]), (&(__pyx_v_total_new_inf_ward[0])), (&__pyx_v_lock), NULL);

                                      /* "metawards/extractors/_output_core.pyx":344
 *                                    &(S_in_wards[0]), &lock)
 * 
 *                     if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         _add_to_buffer(total_new_inf_ward_buffer,
 */
                                    }

                                    /* "metawards/extractors/_output_core.pyx":338
 *                 ifrom = links_ifrom[j]
 * 
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += links[j].suscept
 *                     redvar[0].susceptibles += <int>(links_suscept[j])
 */
                                  }

                                  /* "metawards/extractors/_output_core.pyx":350
 *                                        &(total_new_inf_ward[0]), &lock)
 * 
 *                 if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                     # inf_tot[i] += infections[i][j]
 *                     redvar[0].inf_tot += infections_i[j]
 */
                                  __pyx_t_12 = (((__pyx_v_infections_i[__pyx_v_j]) != 0) != 0);
                                  if (__pyx_t_12) {

                                    /* "metawards/extractors/_output_core.pyx":352
 *                 if infections_i[j] != 0:
 *                     # inf_tot[i] += infections[i][j]
 *                     redvar[0].inf_tot += infections_i[j]             # <<<<<<<<<<<<<<
 *                     #total_inf_ward[ifrom] += infections[i][j]
 *                     _add_to_buffer(total_inf_ward_buffer,
 */
                                    __pyx_t_13 = 0;
                                    (__pyx_v_redvar[__pyx_t_13]).inf_tot = ((__pyx_v_redvar[__pyx_t_13]).inf_tot + (__pyx_v_infections_i[__pyx_v_j]));

                                    /* "metawards/extractors/_output_core.pyx":354
 *                     redvar[0].inf_tot += infections_i[j]
 *                     #total_inf_ward[ifrom] += infections[i][j]
 *                     _add_to_buffer(total_inf_ward_buffer,             # <<<<<<<<<<<<<<
 *                                    ifrom, infections_i[j],
 *                                    &(total_inf_ward[0]), &lock)
 */
                                    __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_total_inf_ward_buffer, __pyx_v_ifrom, (__pyx_v_infections_i[__pyx_v_j]), (&(__pyx_v_total_inf_ward[0])), (&__pyx_v_lock), NULL);

                                    /* "metawards/extractors/_output_core.pyx":358
 *                                    &(total_inf_ward[0]), &lock)
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         _add_to_buffer(R_buffer, ifrom, infections_i[j],
 *                                        &(R_in_wards[0]), &lock)
 */
                                    __pyx_t_14 = ((__pyx_v_i == 0) != 0);
                                    if (!__pyx_t_14) {
                                    } else {
                                      __pyx_t_12 = __pyx_t_14;
                                      goto __pyx_L22_bool_binop_done;
                                    }
                                    __pyx_t_14 = ((__pyx_v_i == __pyx_v_N_INF_CLASSES_MINUS_ONE) != 0);
                                    __pyx_t_12 = __pyx_t_14;
                                    __pyx_L22_bool_binop_done:;
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":359
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         _add_to_buffer(R_buffer, ifrom, infections_i[j],             # <<<<<<<<<<<<<<
 *                                        &(R_in_wards[0]), &lock)
 *                     elif i == 1:
 */
                                      __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_R_buffer, __pyx_v_ifrom, (__pyx_v_infections_i[__pyx_v_j]), (&(__pyx_v_R_in_wards[0])), (&__pyx_v_lock), NULL);

                                      /* "metawards/extractors/_output_core.pyx":358
 *                                    &(total_inf_ward[0]), &lock)
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         _add_to_buffer(R_buffer, ifrom, infections_i[j],
 *                                        &(R_in_wards[0]), &lock)
 */
                                      goto __pyx_L21;
                                    }

                                    /* "metawards/extractors/_output_core.pyx":361
 *                         _add_to_buffer(R_buffer, ifrom, infections_i[j],
 *                                        &(R_in_wards[0]), &lock)
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         _add_to_buffer(E_buffer, ifrom, infections_i[j],
 *                                        &(E_in_wards[0]), &lock)
 */
                                    __pyx_t_12 = ((__pyx_v_i == 1) != 0);
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":362
 *                                        &(R_in_wards[0]), &lock)
 *                     elif i == 1:
 *                         _add_to_buffer(E_buffer, ifrom, infections_i[j],             # <<<<<<<<<<<<<<
 *                                        &(E_in_wards[0]), &lock)
 *                     else:
 */
                                      __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_E_buffer, __pyx_v_ifrom, (__pyx_v_infections_i[__pyx_v_j]), (&(__pyx_v_E_in_wards[0])), (&__pyx_v_lock), NULL);

                                      /* "metawards/extractors/_output_core.pyx":361
 *                         _add_to_buffer(R_buffer, ifrom, infections_i[j],
 *                                        &(R_in_wards[0]), &lock)
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         _add_to_buffer(E_buffer, ifrom, infections_i[j],
 *                                        &(E_in_wards[0]), &lock)
 */
                                      goto __pyx_L21;
                                    }

                                    /* "metawards/extractors/_output_core.pyx":365
 *                                        &(E_in_wards[0]), &lock)
 *                     else:
 *                         _add_to_buffer(I_buffer, ifrom, infections_i[j],             # <<<<<<<<<<<<<<
 *                                        &(I_in_wards[0]), &lock)
 *             # end of loop over links
 */
                                    /*else*/ {

                                      /* "metawards/extractors/_output_core.pyx":366
 *                     else:
 *                         _add_to_buffer(I_buffer, ifrom, infections_i[j],
 *                                        &(I_in_wards[0]), &lock)             # <<<<<<<<<<<<<<
 *             # end of loop over links
 * 
 */
                                      __pyx_f_9metawards_10extractors_12_output_core__add_to_buffer(__pyx_v_I_buffer, __pyx_v_ifrom, (__pyx_v_infections_i[__pyx_v_j]), (&(__pyx_v_I_in_wards[0])), (&__pyx_v_lock), NULL);
                                    }
                                    __pyx_L21:;

                                    /* "metawards/extractors/_output_core.pyx":350
 *                                        &(total_new_inf_ward[0]), &lock)
 * 
 *                 if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                     # inf_tot[i] += infections[i][j]
 *                     redvar[0].inf_tot += infections_i[j]
 */
                                  }
                              }
                          }
                      }
                  }

                  /* "metawards/extractors/_output_core.pyx":370
 * 
 *             # cannot reduce in parallel, so manual reduction
 *             openmp.omp_set_lock(&lock)             # <<<<<<<<<<<<<<
 *             _add_from_buffer(total_new_inf_ward_buffer,
 *                              &(total_new_inf_ward[0]))
 */
                  omp_set_lock((&__pyx_v_lock));

                  /* "metawards/extractors/_output_core.pyx":371
 *             # cannot reduce in parallel, so manual reduction
 *             openmp.omp_set_lock(&lock)
 *             _add_from_buffer(total_new_inf_ward_buffer,             # <<<<<<<<<<<<<<
 *                              &(total_new_inf_ward[0]))
 *             _add_from_buffer(total_inf_ward_buffer,
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_total_new_inf_ward_buffer, (&(__pyx_v_total_new_inf_ward[0])));

                  /* "metawards/extractors/_output_core.pyx":373
 *             _add_from_buffer(total_new_inf_ward_buffer,
 *                              &(total_new_inf_ward[0]))
 *             _add_from_buffer(total_inf_ward_buffer,             # <<<<<<<<<<<<<<
 *                              &(total_inf_ward[0]))
 *             _add_from_buffer(S_buffer, &(S_in_wards[0]))
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_total_inf_ward_buffer, (&(__pyx_v_total_inf_ward[0])));

                  /* "metawards/extractors/_output_core.pyx":375
 *             _add_from_buffer(total_inf_ward_buffer,
 *                              &(total_inf_ward[0]))
 *             _add_from_buffer(S_buffer, &(S_in_wards[0]))             # <<<<<<<<<<<<<<
 *             _add_from_buffer(E_buffer, &(E_in_wards[0]))
 *             _add_from_buffer(I_buffer, &(I_in_wards[0]))
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_S_buffer, (&(__pyx_v_S_in_wards[0])));

                  /* "metawards/extractors/_output_core.pyx":376
 *                              &(total_inf_ward[0]))
 *             _add_from_buffer(S_buffer, &(S_in_wards[0]))
 *             _add_from_buffer(E_buffer, &(E_in_wards[0]))             # <<<<<<<<<<<<<<
 *             _add_from_buffer(I_buffer, &(I_in_wards[0]))
 *             _add_from_buffer(R_buffer, &(R_in_wards[0]))
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_E_buffer, (&(__pyx_v_E_in_wards[0])));

                  /* "metawards/extractors/_output_core.pyx":377
 *             _add_from_buffer(S_buffer, &(S_in_wards[0]))
 *             _add_from_buffer(E_buffer, &(E_in_wards[0]))
 *             _add_from_buffer(I_buffer, &(I_in_wards[0]))             # <<<<<<<<<<<<<<
 *             _add_from_buffer(R_buffer, &(R_in_wards[0]))
 *             openmp.omp_unset_lock(&lock)
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_I_buffer, (&(__pyx_v_I_in_wards[0])));

                  /* "metawards/extractors/_output_core.pyx":378
 *             _add_from_buffer(E_buffer, &(E_in_wards[0]))
 *             _add_from_buffer(I_buffer, &(I_in_wards[0]))
 *             _add_from_buffer(R_buffer, &(R_in_wards[0]))             # <<<<<<<<<<<<<<
 *             openmp.omp_unset_lock(&lock)
 * 
 */
                  __pyx_f_9metawards_10extractors_12_output_core__add_from_buffer(__pyx_v_R_buffer, (&(__pyx_v_R_in_wards[0])));

                  /* "metawards/extractors/_output_core.pyx":379
 *             _add_from_buffer(I_buffer, &(I_in_wards[0]))
 *             _add_from_buffer(R_buffer, &(R_in_wards[0]))
 *             openmp.omp_unset_lock(&lock)             # <<<<<<<<<<<<<<
 * 
 *             # loop over all wards (nodes) and accumulate infections
 */
                  omp_unset_lock((&__pyx_v_lock));

                  /* "metawards/extractors/_output_core.pyx":383
 *             # loop over all wards (nodes) and accumulate infections
 *             # from each ward
 *             for j in prange(1, nnodes_plus_one, schedule="static"):             # <<<<<<<<<<<<<<
 *                 if i == 0:
 *                     # susceptibles += wards[j].suscept
 */
                  __pyx_t_9 = __pyx_v_nnodes_plus_one;
                  if ((1 == 0)) abort();
                  {
                      __pyx_t_10 = (__pyx_t_9 - 1 + 1 - 1/abs(1)) / 1;
                      if (__pyx_t_10 > 0)
                      {
                          #ifdef _OPENMP
                          #pragma omp for firstprivate(__pyx_v_j) lastprivate(__pyx_v_j) schedule(static)
                          #endif /* _OPENMP */
                          for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11++){
                              {
                                  __pyx_v_j = (int)(1 + 1 * __pyx_t_11);

                                  /* "metawards/extractors/_output_core.pyx":384
 *             # from each ward
 *             for j in prange(1, nnodes_plus_one, schedule="static"):
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += wards[j].suscept
 *                     redvar[0].susceptibles += <int>(play_suscept[j])
 */
                                  __pyx_t_12 = ((__pyx_v_i == 0) != 0);
                                  if (__pyx_t_12) {

                                    /* "metawards/extractors/_output_core.pyx":386
 *                 if i == 0:
 *                     # susceptibles += wards[j].suscept
 *                     redvar[0].susceptibles += <int>(play_suscept[j])             # <<<<<<<<<<<<<<
 *                     S_in_wards[j] += <int>(play_suscept[j])
 * 
 */
                                    __pyx_t_13 = 0;
                                    (__pyx_v_redvar[__pyx_t_13]).susceptibles = ((__pyx_v_redvar[__pyx_t_13]).susceptibles + ((int)(__pyx_v_play_suscept[__pyx_v_j])));

                                    /* "metawards/extractors/_output_core.pyx":387
 *                     # susceptibles += wards[j].suscept
 *                     redvar[0].susceptibles += <int>(play_suscept[j])
 *                     S_in_wards[j] += <int>(play_suscept[j])             # <<<<<<<<<<<<<<
 * 
 *                     if play_infections_i[j] > 0:
 */
                                    __pyx_t_15 = __pyx_v_j;
                                    (__pyx_v_S_in_wards[__pyx_t_15]) = ((__pyx_v_S_in_wards[__pyx_t_15]) + ((int)(__pyx_v_play_suscept[__pyx_v_j])));

                                    /* "metawards/extractors/_output_core.pyx":389
 *                     S_in_wards[j] += <int>(play_suscept[j])
 * 
 *                     if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]
 */
                                    __pyx_t_12 = (((__pyx_v_play_infections_i[__pyx_v_j]) > 0) != 0);
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":391
 *                     if play_infections_i[j] > 0:
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                     if total_new_inf_ward[j] != 0:
 */
                                      __pyx_t_15 = __pyx_v_j;
                                      (__pyx_v_total_new_inf_ward[__pyx_t_15]) = ((__pyx_v_total_new_inf_ward[__pyx_t_15]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                                      /* "metawards/extractors/_output_core.pyx":389
 *                     S_in_wards[j] += <int>(play_suscept[j])
 * 
 *                     if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]
 */
                                    }

                                    /* "metawards/extractors/_output_core.pyx":393
 *                         total_new_inf_ward[j] += play_infections_i[j]
 * 
 *                     if total_new_inf_ward[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new += total_new_info_ward[j]
 *                         redvar[0].total_new += total_new_inf_ward[j]
 */
                                    __pyx_t_12 = (((__pyx_v_total_new_inf_ward[__pyx_v_j]) != 0) != 0);
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":395
 *                     if total_new_inf_ward[j] != 0:
 *                         # total_new += total_new_info_ward[j]
 *                         redvar[0].total_new += total_new_inf_ward[j]             # <<<<<<<<<<<<<<
 * 
 *                 if play_infections_i[j] > 0:
 */
                                      __pyx_t_13 = 0;
                                      (__pyx_v_redvar[__pyx_t_13]).total_new = ((__pyx_v_redvar[__pyx_t_13]).total_new + (__pyx_v_total_new_inf_ward[__pyx_v_j]));

                                      /* "metawards/extractors/_output_core.pyx":393
 *                         total_new_inf_ward[j] += play_infections_i[j]
 * 
 *                     if total_new_inf_ward[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new += total_new_info_ward[j]
 *                         redvar[0].total_new += total_new_inf_ward[j]
 */
                                    }

                                    /* "metawards/extractors/_output_core.pyx":384
 *             # from each ward
 *             for j in prange(1, nnodes_plus_one, schedule="static"):
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += wards[j].suscept
 *                     redvar[0].susceptibles += <int>(play_suscept[j])
 */
                                  }

                                  /* "metawards/extractors/_output_core.pyx":397
 *                         redvar[0].total_new += total_new_inf_ward[j]
 * 
 *                 if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     redvar[0].pinf_tot += play_infections_i[j]
 */
                                  __pyx_t_12 = (((__pyx_v_play_infections_i[__pyx_v_j]) > 0) != 0);
                                  if (__pyx_t_12) {

                                    /* "metawards/extractors/_output_core.pyx":399
 *                 if play_infections_i[j] > 0:
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     redvar[0].pinf_tot += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     # total_inf_ward[j] += play_infections[i][j]
 *                     total_inf_ward[j] += play_infections_i[j]
 */
                                    __pyx_t_13 = 0;
                                    (__pyx_v_redvar[__pyx_t_13]).pinf_tot = ((__pyx_v_redvar[__pyx_t_13]).pinf_tot + (__pyx_v_play_infections_i[__pyx_v_j]));

                                    /* "metawards/extractors/_output_core.pyx":401
 *                     redvar[0].pinf_tot += play_infections_i[j]
 *                     # total_inf_ward[j] += play_infections[i][j]
 *                     total_inf_ward[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 */
                                    __pyx_t_15 = __pyx_v_j;
                                    (__pyx_v_total_inf_ward[__pyx_t_15]) = ((__pyx_v_total_inf_ward[__pyx_t_15]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                                    /* "metawards/extractors/_output_core.pyx":403
 *                     total_inf_ward[j] += play_infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 */
                                    __pyx_t_14 = ((__pyx_v_i == 0) != 0);
                                    if (!__pyx_t_14) {
                                    } else {
                                      __pyx_t_12 = __pyx_t_14;
                                      goto __pyx_L35_bool_binop_done;
                                    }
                                    __pyx_t_14 = ((__pyx_v_i == __pyx_v_N_INF_CLASSES_MINUS_ONE) != 0);
                                    __pyx_t_12 = __pyx_t_14;
                                    __pyx_L35_bool_binop_done:;
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":404
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     elif i == 1:
 *                         E_in_wards[j] += play_infections_i[j]
 */
                                      __pyx_t_15 = __pyx_v_j;
                                      (__pyx_v_R_in_wards[__pyx_t_15]) = ((__pyx_v_R_in_wards[__pyx_t_15]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                                      /* "metawards/extractors/_output_core.pyx":403
 *                     total_inf_ward[j] += play_infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 */
                                      goto __pyx_L34;
                                    }

                                    /* "metawards/extractors/_output_core.pyx":405
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 */
                                    __pyx_t_12 = ((__pyx_v_i == 1) != 0);
                                    if (__pyx_t_12) {

                                      /* "metawards/extractors/_output_core.pyx":406
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 *                         E_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     else:
 *                         I_in_wards[j] += play_infections_i[j]
 */
                                      __pyx_t_15 = __pyx_v_j;
                                      (__pyx_v_E_in_wards[__pyx_t_15]) = ((__pyx_v_E_in_wards[__pyx_t_15]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                                      /* "metawards/extractors/_output_core.pyx":405
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 */
                                      goto __pyx_L34;
                                    }

                                    /* "metawards/extractors/_output_core.pyx":408
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 *                         I_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                 if (i < N_INF_CLASSES-1) and total_inf_ward[j] > 0:
 */
                                    /*else*/ {
                                      __pyx_t_15 = __pyx_v_j;
                                      (__pyx_v_I_in_wards[__pyx_t_15]) = ((__pyx_v_I_in_wards[__pyx_t_15]) + (__pyx_v_play_infections_i[__pyx_v_j]));
                                    }
                                    __pyx_L34:;

                                    /* "metawards/extractors/_output_core.pyx":397
 *                         redvar[0].total_new += total_new_inf_ward[j]
 * 
 *                 if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     redvar[0].pinf_tot += play_infections_i[j]
 */
                                  }

                                  /* "metawards/extractors/_output_core.pyx":410
 *                         I_in_wards[j] += play_infections_i[j]
 * 
 *                 if (i < N_INF_CLASSES-1) and total_inf_ward[j] > 0:             # <<<<<<<<<<<<<<
 *                     # n_inf_wards[i] += 1
 *                     redvar[0].n_inf_wards += 1
 */
                                  __pyx_t_14 = ((__pyx_v_i < (__pyx_v_N_INF_CLASSES - 1)) != 0);
                                  if (__pyx_t_14) {
                                  } else {
                                    __pyx_t_12 = __pyx_t_14;
                                    goto __pyx_L38_bool_binop_done;
                                  }
                                  __pyx_t_14 = (((__pyx_v_total_inf_ward[__pyx_v_j]) > 0) != 0);
                                  __pyx_t_12 = __pyx_t_14;
                                  __pyx_L38_bool_binop_done:;
                                  if (__pyx_t_12) {

                                    /* "metawards/extractors/_output_core.pyx":412
 *                 if (i < N_INF_CLASSES-1) and total_inf_ward[j] > 0:
 *                     # n_inf_wards[i] += 1
 *                     redvar[0].n_inf_wards += 1             # <<<<<<<<<<<<<<
 *             # end of loop over nodes
 *         # end of parallel
 */
                                    __pyx_t_13 = 0;
                                    (__pyx_v_redvar[__pyx_t_13]).n_inf_wards = ((__pyx_v_redvar[__pyx_t_13]).n_inf_wards + 1);

                                    /* "metawards/extractors/_output_core.pyx":410
 *                         I_in_wards[j] += play_infections_i[j]
 * 
 *                 if (i < N_INF_CLASSES-1) and total_inf_ward[j] > 0:             # <<<<<<<<<<<<<<
 *                     # n_inf_wards[i] += 1
 *                     redvar[0].n_inf_wards += 1
 */
                                  }
                              }
                          }
                      }
                  }
              }
          }
          #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
              #undef likely
              #undef unlikely
              #define likely(x)   __builtin_expect(!!(x), 1)
              #define unlikely(x) __builtin_expect(!!(x), 0)
          #endif
        }

        /* "metawards/extractors/_output_core.pyx":320
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 *         with nogil, parallel(num_threads=num_threads):             # <<<<<<<<<<<<<<
 *             # get the space for this thread to accumulate data.
 *             # All of this data will be reduced in critical sections
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "metawards/extractors/_output_core.pyx":416
 *         # end of parallel
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if i == 2:
 *                 # save the sum of infections up to i <= 2. This
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "metawards/extractors/_output_core.pyx":417
 * 
 *         with nogil:
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 */
          __pyx_t_12 = ((__pyx_v_i == 2) != 0);
          if (__pyx_t_12) {

            /* "metawards/extractors/_output_core.pyx":420
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 *                 for j in range(1, nnodes_plus_one):             # <<<<<<<<<<<<<<
 *                     incidence[j] = total_inf_ward[j]
 * 
 */
            __pyx_t_9 = __pyx_v_nnodes_plus_one;
            __pyx_t_15 = __pyx_t_9;
            for (__pyx_t_16 = 1; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
              __pyx_v_j = __pyx_t_16;

              /* "metawards/extractors/_output_core.pyx":421
 *                 # is the incidence
 *                 for j in range(1, nnodes_plus_one):
 *                     incidence[j] = total_inf_ward[j]             # <<<<<<<<<<<<<<
 * 
 *         # can now reduce across all threads (don't need to lock as each
 */
              (__pyx_v_incidence[__pyx_v_j]) = (__pyx_v_total_inf_ward[__pyx_v_j]);
            }

            /* "metawards/extractors/_output_core.pyx":417
 * 
 *         with nogil:
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 */
          }
        }

        /* "metawards/extractors/_output_core.pyx":416
 *         # end of parallel
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             if i == 2:
 *                 # save the sum of infections up to i <= 2. This
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L48;
          }
          __pyx_L48:;
        }
    }

    /* "metawards/extractors/_output_core.pyx":425
 *         # can now reduce across all threads (don't need to lock as each
 *         # thread maintained its own running total)
 *         _reduce_variables(_redvars, num_threads)             # <<<<<<<<<<<<<<
 * 
 *         inf_tot[i] = _redvars[0].inf_tot
 */
    __pyx_f_9metawards_10extractors_12_output_core__reduce_variables(__pyx_v_9metawards_10extractors_12_output_core__redvars, __pyx_v_num_threads);

    /* "metawards/extractors/_output_core.pyx":427
 *         _reduce_variables(_redvars, num_threads)
 * 
 *         inf_tot[i] = _redvars[0].inf_tot             # <<<<<<<<<<<<<<
 *         pinf_tot[i] = _redvars[0].pinf_tot
 *         n_inf_wards[i] = _redvars[0].n_inf_wards
 */
    __pyx_t_9 = (__pyx_v_9metawards_10extractors_12_output_core__redvars[0]).inf_tot;
    (__pyx_v_inf_tot[__pyx_v_i]) = __pyx_t_9;

    /* "metawards/extractors/_output_core.pyx":428
 * 
 *         inf_tot[i] = _redvars[0].inf_tot
 *         pinf_tot[i] = _redvars[0].pinf_tot             # <<<<<<<<<<<<<<
 *         n_inf_wards[i] = _redvars[0].n_inf_wards
 * 
 */
    __pyx_t_9 = (__pyx_v_9metawards_10extractors_12_output_core__redvars[0]).pinf_tot;
    (__pyx_v_pinf_tot[__pyx_v_i]) = __pyx_t_9;

    /* "metawards/extractors/_output_core.pyx":429
 *         inf_tot[i] = _redvars[0].inf_tot
 *         pinf_tot[i] = _redvars[0].pinf_tot
 *         n_inf_wards[i] = _redvars[0].n_inf_wards             # <<<<<<<<<<<<<<
 * 
 *         # total_new += total_new[i]
 */
    __pyx_t_9 = (__pyx_v_9metawards_10extractors_12_output_core__redvars[0]).n_inf_wards;
    (__pyx_v_n_inf_wards[__pyx_v_i]) = __pyx_t_9;

    /* "metawards/extractors/_output_core.pyx":432
 * 
 *         # total_new += total_new[i]
 *         total_new += _redvars[0].total_new             # <<<<<<<<<<<<<<
 *         # susceptibles += susceptibles[i]
 *         susceptibles += _redvars[0].susceptibles
 */
    __pyx_v_total_new = (__pyx_v_total_new + (__pyx_v_9metawards_10extractors_12_output_core__redvars[0]).total_new);

    /* "metawards/extractors/_output_core.pyx":434
 *         total_new += _redvars[0].total_new
 *         # susceptibles += susceptibles[i]
 *         susceptibles += _redvars[0].susceptibles             # <<<<<<<<<<<<<<
 * 
 *         _reset_reduce_variables(_redvars, num_threads)
 */
    __pyx_v_susceptibles = (__pyx_v_susceptibles + (__pyx_v_9metawards_10extractors_12_output_core__redvars[0]).susceptibles);

    /* "metawards/extractors/_output_core.pyx":436
 *         susceptibles += _redvars[0].susceptibles
 * 
 *         _reset_reduce_variables(_redvars, num_threads)             # <<<<<<<<<<<<<<
 *     # end of loop over i
 * 
 */
    __pyx_f_9metawards_10extractors_12_output_core__reset_reduce_variables(__pyx_v_9metawards_10extractors_12_output_core__redvars, __pyx_v_num_threads);
  }

  /* "metawards/extractors/_output_core.pyx":440
 * 
 *     # latent is the sum of work and play infections for stage 1
 *     latent = inf_tot[1] + pinf_tot[1]             # <<<<<<<<<<<<<<
 * 
 *     # total infections are sum of work and play for all intermediate
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(((__pyx_v_inf_tot[1]) + (__pyx_v_pinf_tot[1]))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_latent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":444
 *     # total infections are sum of work and play for all intermediate
 *     # stages
 *     total = 0             # <<<<<<<<<<<<<<
 *     for i in range(2, N_INF_CLASSES-1):
 *         total += inf_tot[i] + pinf_tot[i]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_total = __pyx_int_0;

  /* "metawards/extractors/_output_core.pyx":445
 *     # stages
 *     total = 0
 *     for i in range(2, N_INF_CLASSES-1):             # <<<<<<<<<<<<<<
 *         total += inf_tot[i] + pinf_tot[i]
 * 
 */
  __pyx_t_10 = (__pyx_v_N_INF_CLASSES - 1);
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_5 = 2; __pyx_t_5 < __pyx_t_11; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "metawards/extractors/_output_core.pyx":446
 *     total = 0
 *     for i in range(2, N_INF_CLASSES-1):
 *         total += inf_tot[i] + pinf_tot[i]             # <<<<<<<<<<<<<<
 * 
 *     # recovered is the sum of work and play for the first and last stage, e.g.
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(((__pyx_v_inf_tot[__pyx_v_i]) + (__pyx_v_pinf_tot[__pyx_v_i]))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_total, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_total, __pyx_t_4);
    __pyx_t_4 = 0;
  }

  /* "metawards/extractors/_output_core.pyx":452
 *     # (Note we don't have reverse indexingas these are plain C pointers)
 *     recovereds = inf_tot[0] + inf_tot[N_INF_CLASSES-1] + \
 *                  pinf_tot[0] + pinf_tot[N_INF_CLASSES-1]             # <<<<<<<<<<<<<<
 * 
 *     cdef int S = 0
 */
  __pyx_v_recovereds = ((((__pyx_v_inf_tot[0]) + (__pyx_v_inf_tot[(__pyx_v_N_INF_CLASSES - 1)])) + (__pyx_v_pinf_tot[0])) + (__pyx_v_pinf_tot[(__pyx_v_N_INF_CLASSES - 1)]));

  /* "metawards/extractors/_output_core.pyx":454
 *                  pinf_tot[0] + pinf_tot[N_INF_CLASSES-1]
 * 
 *     cdef int S = 0             # <<<<<<<<<<<<<<
 *     cdef int E = 0
 *     cdef int I = 0
 */
  __pyx_v_S = 0;

  /* "metawards/extractors/_output_core.pyx":455
 * 
 *     cdef int S = 0
 *     cdef int E = 0             # <<<<<<<<<<<<<<
 *     cdef int I = 0
 *     cdef int R = 0
 */
  __pyx_v_E = 0;

  /* "metawards/extractors/_output_core.pyx":456
 *     cdef int S = 0
 *     cdef int E = 0
 *     cdef int I = 0             # <<<<<<<<<<<<<<
 *     cdef int R = 0
 * 
 */
  __pyx_v_I = 0;

  /* "metawards/extractors/_output_core.pyx":457
 *     cdef int E = 0
 *     cdef int I = 0
 *     cdef int R = 0             # <<<<<<<<<<<<<<
 * 
 *     for j in range(1, nnodes_plus_one):
 */
  __pyx_v_R = 0;

  /* "metawards/extractors/_output_core.pyx":459
 *     cdef int R = 0
 * 
 *     for j in range(1, nnodes_plus_one):             # <<<<<<<<<<<<<<
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]
 */
  __pyx_t_5 = __pyx_v_nnodes_plus_one;
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_8 = 1; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_j = __pyx_t_8;

    /* "metawards/extractors/_output_core.pyx":460
 * 
 *     for j in range(1, nnodes_plus_one):
 *         S += S_in_wards[j]             # <<<<<<<<<<<<<<
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]
 */
    __pyx_v_S = (__pyx_v_S + (__pyx_v_S_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":461
 *     for j in range(1, nnodes_plus_one):
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]             # <<<<<<<<<<<<<<
 *         I += I_in_wards[j]
 *         R += R_in_wards[j]
 */
    __pyx_v_E = (__pyx_v_E + (__pyx_v_E_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":462
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]             # <<<<<<<<<<<<<<
 *         R += R_in_wards[j]
 * 
 */
    __pyx_v_I = (__pyx_v_I + (__pyx_v_I_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":463
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]
 *         R += R_in_wards[j]             # <<<<<<<<<<<<<<
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 */
    __pyx_v_R = (__pyx_v_R + (__pyx_v_R_in_wards[__pyx_v_j]));
  }

  /* "metawards/extractors/_output_core.pyx":465
 *         R += R_in_wards[j]
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:             # <<<<<<<<<<<<<<
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 */
  __pyx_t_14 = ((__pyx_v_S != __pyx_v_susceptibles) != 0);
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L57_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_E); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_v_latent, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L57_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_I); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_v_total, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L57_bool_binop_done;
  }
  __pyx_t_14 = ((__pyx_v_R != __pyx_v_recovereds) != 0);
  __pyx_t_12 = __pyx_t_14;
  __pyx_L57_bool_binop_done:;
  if (__pyx_t_12) {

    /* "metawards/extractors/_output_core.pyx":467
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \             # <<<<<<<<<<<<<<
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"
 */
    __pyx_t_4 = PyTuple_New(16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_17 = 127;
    __Pyx_INCREF(__pyx_kp_u_Disagreement_in_accumulated_tota);
    __pyx_t_3 += 62;
    __Pyx_GIVEREF(__pyx_kp_u_Disagreement_in_accumulated_tota);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Disagreement_in_accumulated_tota);

    /* "metawards/extractors/_output_core.pyx":468
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \             # <<<<<<<<<<<<<<
 *             f"{R} vs {recovereds}"
 * 
 */
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_S, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_vs);
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_susceptibles, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_);
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_E, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_kp_u_vs);
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_latent, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_17;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_kp_u_);
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_I, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 9, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 10, __pyx_kp_u_vs);
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_total, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_17;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 11, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 12, __pyx_kp_u_);

    /* "metawards/extractors/_output_core.pyx":469
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"             # <<<<<<<<<<<<<<
 * 
 *         from ..utils._console import Console
 */
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_R, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 13, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 14, __pyx_kp_u_vs);
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_recovereds, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 15, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":467
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \             # <<<<<<<<<<<<<<
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"
 */
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_4, 16, __pyx_t_3, __pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_error = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":471
 *             f"{R} vs {recovereds}"
 * 
 *         from ..utils._console import Console             # <<<<<<<<<<<<<<
 *         Console.error(error)
 *         raise AssertionError(error)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_Console);
    __Pyx_GIVEREF(__pyx_n_s_Console);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Console);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_utils__console, __pyx_t_1, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Console); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_Console = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":472
 * 
 *         from ..utils._console import Console
 *         Console.error(error)             # <<<<<<<<<<<<<<
 *         raise AssertionError(error)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_error) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_error);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":473
 *         from ..utils._console import Console
 *         Console.error(error)
 *         raise AssertionError(error)             # <<<<<<<<<<<<<<
 * 
 *     if population is not None:
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_v_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 473, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":465
 *         R += R_in_wards[j]
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:             # <<<<<<<<<<<<<<
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 */
  }

  /* "metawards/extractors/_output_core.pyx":475
 *         raise AssertionError(error)
 * 
 *     if population is not None:             # <<<<<<<<<<<<<<
 *         population.susceptibles = susceptibles
 *         population.total = total
 */
  __pyx_t_12 = (__pyx_v_population != Py_None);
  __pyx_t_14 = (__pyx_t_12 != 0);
  if (__pyx_t_14) {

    /* "metawards/extractors/_output_core.pyx":476
 * 
 *     if population is not None:
 *         population.susceptibles = susceptibles             # <<<<<<<<<<<<<<
 *         population.total = total
 *         population.recovereds = recovereds
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_susceptibles); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_susceptibles, __pyx_t_4) < 0) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":477
 *     if population is not None:
 *         population.susceptibles = susceptibles
 *         population.total = total             # <<<<<<<<<<<<<<
 *         population.recovereds = recovereds
 *         population.latent = latent
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_total, __pyx_v_total) < 0) __PYX_ERR(0, 477, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":478
 *         population.susceptibles = susceptibles
 *         population.total = total
 *         population.recovereds = recovereds             # <<<<<<<<<<<<<<
 *         population.latent = latent
 *         # save the number of wards that have at least one latent
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_recovereds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_recovereds, __pyx_t_4) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":479
 *         population.total = total
 *         population.recovereds = recovereds
 *         population.latent = latent             # <<<<<<<<<<<<<<
 *         # save the number of wards that have at least one latent
 *         # infection
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_latent, __pyx_v_latent) < 0) __PYX_ERR(0, 479, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":482
 *         # save the number of wards that have at least one latent
 *         # infection
 *         population.n_inf_wards = n_inf_wards[0]             # <<<<<<<<<<<<<<
 * 
 *     return total + latent
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_n_inf_wards[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_n_inf_wards, __pyx_t_4) < 0) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":475
 *         raise AssertionError(error)
 * 
 *     if population is not None:             # <<<<<<<<<<<<<<
 *         population.susceptibles = susceptibles
 *         population.total = total
 */
  }

  /* "metawards/extractors/_output_core.pyx":484
 *         population.n_inf_wards = n_inf_wards[0]
 * 
 *     return total + latent             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyNumber_Add(__pyx_v_total, __pyx_v_latent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "metawards/extractors/_output_core.pyx":198
 * 
 * 
 * def output_core_omp(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                     workspace: Workspace,
 *                     infections: Infections,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core_omp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_links);
  __Pyx_XDECREF(__pyx_v_wards);
  __Pyx_XDECREF(__pyx_v_play_infections);
  __Pyx_XDECREF(__pyx_v_latent);
  __Pyx_XDECREF(__pyx_v_total);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_Console);
  __Pyx_XDECREF(__pyx_v_infections);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":487
 * 
 * 
 * def output_core_serial(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                        workspace: Workspace,
 *                        infections: Infections,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_5output_core_serial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9metawards_10extractors_12_output_core_4output_core_serial[] = "output_core_serial(network: Network, population: Population, workspace: Workspace, infections: Infections, **kwargs)\nThis is the core output function that must be called\n       every iteration as it is responsible for accumulating\n       the core data each day, which is used to report a summary\n       to the main output file. This is the serial version\n       of this function.\n\n       Parameters\n       ----------\n       network: Network\n         The network over which the outbreak is being modelled\n       population: Population\n         The population experiencing the outbreak\n       workspace: Workspace\n         A workspace that can be used to extract data\n       infections: Infections\n         All of the infections that have been recorded\n       nthreads: int\n         The number of threads to use to help extract the data\n       kwargs\n         Extra argumentst that are ignored by this function\n    ";
static PyMethodDef __pyx_mdef_9metawards_10extractors_12_output_core_5output_core_serial = {"output_core_serial", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9metawards_10extractors_12_output_core_5output_core_serial, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9metawards_10extractors_12_output_core_4output_core_serial};
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_5output_core_serial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_network = 0;
  PyObject *__pyx_v_population = 0;
  PyObject *__pyx_v_workspace = 0;
  PyObject *__pyx_v_infections = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output_core_serial (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_network,&__pyx_n_s_population,&__pyx_n_s_workspace,&__pyx_n_s_infections,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_network)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_population)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_serial", 1, 4, 4, 1); __PYX_ERR(0, 487, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_serial", 1, 4, 4, 2); __PYX_ERR(0, 487, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_infections)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core_serial", 1, 4, 4, 3); __PYX_ERR(0, 487, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "output_core_serial") < 0)) __PYX_ERR(0, 487, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_network = values[0];
    __pyx_v_population = values[1];
    __pyx_v_workspace = values[2];
    __pyx_v_infections = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("output_core_serial", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 487, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core_serial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9metawards_10extractors_12_output_core_4output_core_serial(__pyx_self, __pyx_v_network, __pyx_v_population, __pyx_v_workspace, __pyx_v_infections, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_4output_core_serial(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_links = NULL;
  PyObject *__pyx_v_wards = NULL;
  PyObject *__pyx_v_play_infections = NULL;
  int __pyx_v_N_INF_CLASSES;
  int *__pyx_v_inf_tot;
  int *__pyx_v_pinf_tot;
  int *__pyx_v_total_inf_ward;
  int *__pyx_v_total_new_inf_ward;
  int *__pyx_v_n_inf_wards;
  int *__pyx_v_incidence;
  int *__pyx_v_S_in_wards;
  int *__pyx_v_E_in_wards;
  int *__pyx_v_I_in_wards;
  int *__pyx_v_R_in_wards;
  int *__pyx_v_links_ifrom;
  double *__pyx_v_links_suscept;
  double *__pyx_v_play_suscept;
  int __pyx_v_nlinks_plus_one;
  int __pyx_v_nnodes_plus_one;
  int *__pyx_v_infections_i;
  int *__pyx_v_play_infections_i;
  int __pyx_v_total_new;
  int __pyx_v_susceptibles;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_ifrom;
  int __pyx_v_n_inf_wards_i;
  int __pyx_v_total_new_i;
  int __pyx_v_inf_tot_i;
  int __pyx_v_pinf_tot_i;
  int __pyx_v_susceptibles_i;
  int __pyx_v_N_INF_CLASSES_MINUS_ONE;
  PyObject *__pyx_v_latent = NULL;
  PyObject *__pyx_v_total = NULL;
  int __pyx_v_recovereds;
  int __pyx_v_S;
  int __pyx_v_E;
  int __pyx_v_I;
  int __pyx_v_R;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_Console = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  long __pyx_t_15;
  long __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("output_core_serial", 0);
  __Pyx_INCREF(__pyx_v_infections);

  /* "metawards/extractors/_output_core.pyx":512
 *          Extra argumentst that are ignored by this function
 *     """
 *     links = network.links             # <<<<<<<<<<<<<<
 *     wards = network.nodes
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_links); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":513
 *     """
 *     links = network.links
 *     wards = network.nodes             # <<<<<<<<<<<<<<
 * 
 *     # set up main variables
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nodes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wards = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":516
 * 
 *     # set up main variables
 *     play_infections = infections.play             # <<<<<<<<<<<<<<
 *     infections = infections.work
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_play); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_play_infections = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":517
 *     # set up main variables
 *     play_infections = infections.play
 *     infections = infections.work             # <<<<<<<<<<<<<<
 * 
 *     cdef int N_INF_CLASSES = len(infections)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_work); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_infections, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":519
 *     infections = infections.work
 * 
 *     cdef int N_INF_CLASSES = len(infections)             # <<<<<<<<<<<<<<
 *     assert len(infections) == len(play_infections)
 * 
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_infections); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 519, __pyx_L1_error)
  __pyx_v_N_INF_CLASSES = __pyx_t_2;

  /* "metawards/extractors/_output_core.pyx":520
 * 
 *     cdef int N_INF_CLASSES = len(infections)
 *     assert len(infections) == len(play_infections)             # <<<<<<<<<<<<<<
 * 
 *     # get pointers to arrays in workspace to write data
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_2 = PyObject_Length(__pyx_v_infections); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 520, __pyx_L1_error)
    __pyx_t_3 = PyObject_Length(__pyx_v_play_infections); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 520, __pyx_L1_error)
    if (unlikely(!((__pyx_t_2 == __pyx_t_3) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 520, __pyx_L1_error)
    }
  }
  #endif

  /* "metawards/extractors/_output_core.pyx":523
 * 
 *     # get pointers to arrays in workspace to write data
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)             # <<<<<<<<<<<<<<
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_inf_tot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inf_tot = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":524
 *     # get pointers to arrays in workspace to write data
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)             # <<<<<<<<<<<<<<
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_pinf_tot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pinf_tot = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":525
 *     cdef int * inf_tot = get_int_array_ptr(workspace.inf_tot)
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)             # <<<<<<<<<<<<<<
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_total_inf_ward); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_total_inf_ward = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":527
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)             # <<<<<<<<<<<<<<
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_total_new_inf_ward); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "metawards/extractors/_output_core.pyx":526
 *     cdef int * pinf_tot = get_int_array_ptr(workspace.pinf_tot)
 *     cdef int * total_inf_ward = get_int_array_ptr(workspace.total_inf_ward)
 *     cdef int * total_new_inf_ward = get_int_array_ptr(             # <<<<<<<<<<<<<<
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 */
  __pyx_v_total_new_inf_ward = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":528
 *     cdef int * total_new_inf_ward = get_int_array_ptr(
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)             # <<<<<<<<<<<<<<
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_n_inf_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_n_inf_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":529
 *                                                 workspace.total_new_inf_ward)
 *     cdef int * n_inf_wards = get_int_array_ptr(workspace.n_inf_wards)
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)             # <<<<<<<<<<<<<<
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_incidence); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_incidence = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":531
 *     cdef int * incidence = get_int_array_ptr(workspace.incidence)
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_S_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_S_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":532
 * 
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_E_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_E_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":533
 *     cdef int * S_in_wards = get_int_array_ptr(workspace.S_in_wards)
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)             # <<<<<<<<<<<<<<
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_I_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_I_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":534
 *     cdef int * E_in_wards = get_int_array_ptr(workspace.E_in_wards)
 *     cdef int * I_in_wards = get_int_array_ptr(workspace.I_in_wards)
 *     cdef int * R_in_wards = get_int_array_ptr(workspace.R_in_wards)             # <<<<<<<<<<<<<<
 * 
 *     # get pointers to arrays from links and play to read data
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_R_in_wards); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_R_in_wards = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":537
 * 
 *     # get pointers to arrays from links and play to read data
 *     cdef int * links_ifrom = get_int_array_ptr(links.ifrom)             # <<<<<<<<<<<<<<
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_links, __pyx_n_s_ifrom); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links_ifrom = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":539
 *     cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)             # <<<<<<<<<<<<<<
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_links, __pyx_n_s_suscept); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_links_suscept = __pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":540
 * 
 *     cdef double * links_suscept = get_double_array_ptr(links.suscept)
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)             # <<<<<<<<<<<<<<
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_wards, __pyx_n_s_play_suscept); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_play_suscept = __pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":542
 *     cdef double * play_suscept = get_double_array_ptr(wards.play_suscept)
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1             # <<<<<<<<<<<<<<
 *     cdef int nnodes_plus_one = network.nnodes + 1
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nlinks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_nlinks_plus_one = __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":543
 * 
 *     cdef int nlinks_plus_one = network.nlinks + 1
 *     cdef int nnodes_plus_one = network.nnodes + 1             # <<<<<<<<<<<<<<
 * 
 *     # Now pointers into the infections and play_infections arrays
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_nnodes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nnodes_plus_one = __pyx_t_5;

  /* "metawards/extractors/_output_core.pyx":551
 * 
 *     # Now some loop variables used for accumulation
 *     cdef int total_new = 0             # <<<<<<<<<<<<<<
 *     cdef int susceptibles = 0
 * 
 */
  __pyx_v_total_new = 0;

  /* "metawards/extractors/_output_core.pyx":552
 *     # Now some loop variables used for accumulation
 *     cdef int total_new = 0
 *     cdef int susceptibles = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int i = 0
 */
  __pyx_v_susceptibles = 0;

  /* "metawards/extractors/_output_core.pyx":554
 *     cdef int susceptibles = 0
 * 
 *     cdef int i = 0             # <<<<<<<<<<<<<<
 *     cdef int j = 0
 *     cdef int ifrom = 0
 */
  __pyx_v_i = 0;

  /* "metawards/extractors/_output_core.pyx":555
 * 
 *     cdef int i = 0
 *     cdef int j = 0             # <<<<<<<<<<<<<<
 *     cdef int ifrom = 0
 * 
 */
  __pyx_v_j = 0;

  /* "metawards/extractors/_output_core.pyx":556
 *     cdef int i = 0
 *     cdef int j = 0
 *     cdef int ifrom = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int n_inf_wards_i = 0
 */
  __pyx_v_ifrom = 0;

  /* "metawards/extractors/_output_core.pyx":558
 *     cdef int ifrom = 0
 * 
 *     cdef int n_inf_wards_i = 0             # <<<<<<<<<<<<<<
 *     cdef int total_new_i = 0
 *     cdef int inf_tot_i = 0
 */
  __pyx_v_n_inf_wards_i = 0;

  /* "metawards/extractors/_output_core.pyx":559
 * 
 *     cdef int n_inf_wards_i = 0
 *     cdef int total_new_i = 0             # <<<<<<<<<<<<<<
 *     cdef int inf_tot_i = 0
 *     cdef int pinf_tot_i = 0
 */
  __pyx_v_total_new_i = 0;

  /* "metawards/extractors/_output_core.pyx":560
 *     cdef int n_inf_wards_i = 0
 *     cdef int total_new_i = 0
 *     cdef int inf_tot_i = 0             # <<<<<<<<<<<<<<
 *     cdef int pinf_tot_i = 0
 *     cdef int susceptibles_i = 0
 */
  __pyx_v_inf_tot_i = 0;

  /* "metawards/extractors/_output_core.pyx":561
 *     cdef int total_new_i = 0
 *     cdef int inf_tot_i = 0
 *     cdef int pinf_tot_i = 0             # <<<<<<<<<<<<<<
 *     cdef int susceptibles_i = 0
 * 
 */
  __pyx_v_pinf_tot_i = 0;

  /* "metawards/extractors/_output_core.pyx":562
 *     cdef int inf_tot_i = 0
 *     cdef int pinf_tot_i = 0
 *     cdef int susceptibles_i = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int N_INF_CLASSES_MINUS_ONE = N_INF_CLASSES - 1
 */
  __pyx_v_susceptibles_i = 0;

  /* "metawards/extractors/_output_core.pyx":564
 *     cdef int susceptibles_i = 0
 * 
 *     cdef int N_INF_CLASSES_MINUS_ONE = N_INF_CLASSES - 1             # <<<<<<<<<<<<<<
 * 
 *     ###
 */
  __pyx_v_N_INF_CLASSES_MINUS_ONE = (__pyx_v_N_INF_CLASSES - 1);

  /* "metawards/extractors/_output_core.pyx":574
 *     # (make sure not to zero the subspace networks else we lose all our
 *     #  hard work!)
 *     workspace.zero_all(zero_subspaces=False)             # <<<<<<<<<<<<<<
 * 
 *     # loop over each of the disease stages
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_zero_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_zero_subspaces, Py_False) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "metawards/extractors/_output_core.pyx":577
 * 
 *     # loop over each of the disease stages
 *     for i in range(0, N_INF_CLASSES):             # <<<<<<<<<<<<<<
 *         # zero the number of infected wards, and the total number
 *         # of infections for this stage
 */
  __pyx_t_5 = __pyx_v_N_INF_CLASSES;
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "metawards/extractors/_output_core.pyx":580
 *         # zero the number of infected wards, and the total number
 *         # of infections for this stage
 *         n_inf_wards[i] = 0             # <<<<<<<<<<<<<<
 *         inf_tot[i] = 0
 *         pinf_tot[i] = 0
 */
    (__pyx_v_n_inf_wards[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":581
 *         # of infections for this stage
 *         n_inf_wards[i] = 0
 *         inf_tot[i] = 0             # <<<<<<<<<<<<<<
 *         pinf_tot[i] = 0
 * 
 */
    (__pyx_v_inf_tot[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":582
 *         n_inf_wards[i] = 0
 *         inf_tot[i] = 0
 *         pinf_tot[i] = 0             # <<<<<<<<<<<<<<
 * 
 *         # similarly zero the number of infections per ward and
 */
    (__pyx_v_pinf_tot[__pyx_v_i]) = 0;

    /* "metawards/extractors/_output_core.pyx":586
 *         # similarly zero the number of infections per ward and
 *         # number of new infections per ward
 *         n_inf_wards_i = 0             # <<<<<<<<<<<<<<
 *         total_new_i = 0
 *         inf_tot_i = 0
 */
    __pyx_v_n_inf_wards_i = 0;

    /* "metawards/extractors/_output_core.pyx":587
 *         # number of new infections per ward
 *         n_inf_wards_i = 0
 *         total_new_i = 0             # <<<<<<<<<<<<<<
 *         inf_tot_i = 0
 *         pinf_tot_i = 0
 */
    __pyx_v_total_new_i = 0;

    /* "metawards/extractors/_output_core.pyx":588
 *         n_inf_wards_i = 0
 *         total_new_i = 0
 *         inf_tot_i = 0             # <<<<<<<<<<<<<<
 *         pinf_tot_i = 0
 *         susceptibles_i = 0
 */
    __pyx_v_inf_tot_i = 0;

    /* "metawards/extractors/_output_core.pyx":589
 *         total_new_i = 0
 *         inf_tot_i = 0
 *         pinf_tot_i = 0             # <<<<<<<<<<<<<<
 *         susceptibles_i = 0
 * 
 */
    __pyx_v_pinf_tot_i = 0;

    /* "metawards/extractors/_output_core.pyx":590
 *         inf_tot_i = 0
 *         pinf_tot_i = 0
 *         susceptibles_i = 0             # <<<<<<<<<<<<<<
 * 
 *         # get pointers to the work and play infections for this disease
 */
    __pyx_v_susceptibles_i = 0;

    /* "metawards/extractors/_output_core.pyx":594
 *         # get pointers to the work and play infections for this disease
 *         # stage
 *         infections_i = get_int_array_ptr(infections[i])             # <<<<<<<<<<<<<<
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_infections, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_infections_i = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "metawards/extractors/_output_core.pyx":595
 *         # stage
 *         infections_i = get_int_array_ptr(infections[i])
 *         play_infections_i = get_int_array_ptr(play_infections[i])             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_play_infections, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_play_infections_i = __pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "metawards/extractors/_output_core.pyx":597
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             #loop over all links and accumulate infections associated
 *             # with this link
 */
    {
        #ifdef WITH_THREAD
        PyThreadState *_save;
        Py_UNBLOCK_THREADS
        __Pyx_FastGIL_Remember();
        #endif
        /*try:*/ {

          /* "metawards/extractors/_output_core.pyx":600
 *             #loop over all links and accumulate infections associated
 *             # with this link
 *             for j in range(1, nlinks_plus_one):             # <<<<<<<<<<<<<<
 *                 ifrom = links_ifrom[j]
 * 
 */
          __pyx_t_9 = __pyx_v_nlinks_plus_one;
          __pyx_t_10 = __pyx_t_9;
          for (__pyx_t_11 = 1; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
            __pyx_v_j = __pyx_t_11;

            /* "metawards/extractors/_output_core.pyx":601
 *             # with this link
 *             for j in range(1, nlinks_plus_one):
 *                 ifrom = links_ifrom[j]             # <<<<<<<<<<<<<<
 * 
 *                 if i == 0:
 */
            __pyx_v_ifrom = (__pyx_v_links_ifrom[__pyx_v_j]);

            /* "metawards/extractors/_output_core.pyx":603
 *                 ifrom = links_ifrom[j]
 * 
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += links[j].suscept
 *                     susceptibles_i += <int>(links_suscept[j])
 */
            __pyx_t_12 = ((__pyx_v_i == 0) != 0);
            if (__pyx_t_12) {

              /* "metawards/extractors/_output_core.pyx":605
 *                 if i == 0:
 *                     # susceptibles += links[j].suscept
 *                     susceptibles_i += <int>(links_suscept[j])             # <<<<<<<<<<<<<<
 *                     S_in_wards[ifrom] += <int>(links_suscept[j])
 * 
 */
              __pyx_v_susceptibles_i = (__pyx_v_susceptibles_i + ((int)(__pyx_v_links_suscept[__pyx_v_j])));

              /* "metawards/extractors/_output_core.pyx":606
 *                     # susceptibles += links[j].suscept
 *                     susceptibles_i += <int>(links_suscept[j])
 *                     S_in_wards[ifrom] += <int>(links_suscept[j])             # <<<<<<<<<<<<<<
 * 
 *                     if infections_i[j] != 0:
 */
              __pyx_t_13 = __pyx_v_ifrom;
              (__pyx_v_S_in_wards[__pyx_t_13]) = ((__pyx_v_S_in_wards[__pyx_t_13]) + ((int)(__pyx_v_links_suscept[__pyx_v_j])));

              /* "metawards/extractors/_output_core.pyx":608
 *                     S_in_wards[ifrom] += <int>(links_suscept[j])
 * 
 *                     if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         total_new_inf_ward[ifrom] += infections_i[j]
 */
              __pyx_t_12 = (((__pyx_v_infections_i[__pyx_v_j]) != 0) != 0);
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":610
 *                     if infections_i[j] != 0:
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         total_new_inf_ward[ifrom] += infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                 if infections_i[j] != 0:
 */
                __pyx_t_13 = __pyx_v_ifrom;
                (__pyx_v_total_new_inf_ward[__pyx_t_13]) = ((__pyx_v_total_new_inf_ward[__pyx_t_13]) + (__pyx_v_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":608
 *                     S_in_wards[ifrom] += <int>(links_suscept[j])
 * 
 *                     if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[ifrom] += infections[i][j]
 *                         total_new_inf_ward[ifrom] += infections_i[j]
 */
              }

              /* "metawards/extractors/_output_core.pyx":603
 *                 ifrom = links_ifrom[j]
 * 
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += links[j].suscept
 *                     susceptibles_i += <int>(links_suscept[j])
 */
            }

            /* "metawards/extractors/_output_core.pyx":612
 *                         total_new_inf_ward[ifrom] += infections_i[j]
 * 
 *                 if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                     # inf_tot[i] += infections[i][j]
 *                     inf_tot_i += infections_i[j]
 */
            __pyx_t_12 = (((__pyx_v_infections_i[__pyx_v_j]) != 0) != 0);
            if (__pyx_t_12) {

              /* "metawards/extractors/_output_core.pyx":614
 *                 if infections_i[j] != 0:
 *                     # inf_tot[i] += infections[i][j]
 *                     inf_tot_i += infections_i[j]             # <<<<<<<<<<<<<<
 *                     #total_inf_ward[ifrom] += infections[i][j]
 *                     total_inf_ward[ifrom] += infections_i[j]
 */
              __pyx_v_inf_tot_i = (__pyx_v_inf_tot_i + (__pyx_v_infections_i[__pyx_v_j]));

              /* "metawards/extractors/_output_core.pyx":616
 *                     inf_tot_i += infections_i[j]
 *                     #total_inf_ward[ifrom] += infections[i][j]
 *                     total_inf_ward[ifrom] += infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 */
              __pyx_t_13 = __pyx_v_ifrom;
              (__pyx_v_total_inf_ward[__pyx_t_13]) = ((__pyx_v_total_inf_ward[__pyx_t_13]) + (__pyx_v_infections_i[__pyx_v_j]));

              /* "metawards/extractors/_output_core.pyx":618
 *                     total_inf_ward[ifrom] += infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[ifrom] += infections_i[j]
 *                     elif i == 1:
 */
              __pyx_t_14 = ((__pyx_v_i == 0) != 0);
              if (!__pyx_t_14) {
              } else {
                __pyx_t_12 = __pyx_t_14;
                goto __pyx_L16_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_i == __pyx_v_N_INF_CLASSES_MINUS_ONE) != 0);
              __pyx_t_12 = __pyx_t_14;
              __pyx_L16_bool_binop_done:;
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":619
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[ifrom] += infections_i[j]             # <<<<<<<<<<<<<<
 *                     elif i == 1:
 *                         E_in_wards[ifrom] += infections_i[j]
 */
                __pyx_t_13 = __pyx_v_ifrom;
                (__pyx_v_R_in_wards[__pyx_t_13]) = ((__pyx_v_R_in_wards[__pyx_t_13]) + (__pyx_v_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":618
 *                     total_inf_ward[ifrom] += infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[ifrom] += infections_i[j]
 *                     elif i == 1:
 */
                goto __pyx_L15;
              }

              /* "metawards/extractors/_output_core.pyx":620
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[ifrom] += infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[ifrom] += infections_i[j]
 *                     else:
 */
              __pyx_t_12 = ((__pyx_v_i == 1) != 0);
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":621
 *                         R_in_wards[ifrom] += infections_i[j]
 *                     elif i == 1:
 *                         E_in_wards[ifrom] += infections_i[j]             # <<<<<<<<<<<<<<
 *                     else:
 *                         I_in_wards[ifrom] += infections_i[j]
 */
                __pyx_t_13 = __pyx_v_ifrom;
                (__pyx_v_E_in_wards[__pyx_t_13]) = ((__pyx_v_E_in_wards[__pyx_t_13]) + (__pyx_v_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":620
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[ifrom] += infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[ifrom] += infections_i[j]
 *                     else:
 */
                goto __pyx_L15;
              }

              /* "metawards/extractors/_output_core.pyx":623
 *                         E_in_wards[ifrom] += infections_i[j]
 *                     else:
 *                         I_in_wards[ifrom] += infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *             # end of loop over links
 */
              /*else*/ {
                __pyx_t_13 = __pyx_v_ifrom;
                (__pyx_v_I_in_wards[__pyx_t_13]) = ((__pyx_v_I_in_wards[__pyx_t_13]) + (__pyx_v_infections_i[__pyx_v_j]));
              }
              __pyx_L15:;

              /* "metawards/extractors/_output_core.pyx":612
 *                         total_new_inf_ward[ifrom] += infections_i[j]
 * 
 *                 if infections_i[j] != 0:             # <<<<<<<<<<<<<<
 *                     # inf_tot[i] += infections[i][j]
 *                     inf_tot_i += infections_i[j]
 */
            }
          }

          /* "metawards/extractors/_output_core.pyx":629
 *             # loop over all wards (nodes) and accumulate infections
 *             # from each ward
 *             for j in range(1, nnodes_plus_one):             # <<<<<<<<<<<<<<
 *                 if i == 0:
 *                     # susceptibles += wards[j].suscept
 */
          __pyx_t_9 = __pyx_v_nnodes_plus_one;
          __pyx_t_10 = __pyx_t_9;
          for (__pyx_t_11 = 1; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
            __pyx_v_j = __pyx_t_11;

            /* "metawards/extractors/_output_core.pyx":630
 *             # from each ward
 *             for j in range(1, nnodes_plus_one):
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += wards[j].suscept
 *                     susceptibles_i += <int>(play_suscept[j])
 */
            __pyx_t_12 = ((__pyx_v_i == 0) != 0);
            if (__pyx_t_12) {

              /* "metawards/extractors/_output_core.pyx":632
 *                 if i == 0:
 *                     # susceptibles += wards[j].suscept
 *                     susceptibles_i += <int>(play_suscept[j])             # <<<<<<<<<<<<<<
 * 
 *                     S_in_wards[j] += <int>(play_suscept[j])
 */
              __pyx_v_susceptibles_i = (__pyx_v_susceptibles_i + ((int)(__pyx_v_play_suscept[__pyx_v_j])));

              /* "metawards/extractors/_output_core.pyx":634
 *                     susceptibles_i += <int>(play_suscept[j])
 * 
 *                     S_in_wards[j] += <int>(play_suscept[j])             # <<<<<<<<<<<<<<
 * 
 *                     if play_infections_i[j] > 0:
 */
              __pyx_t_13 = __pyx_v_j;
              (__pyx_v_S_in_wards[__pyx_t_13]) = ((__pyx_v_S_in_wards[__pyx_t_13]) + ((int)(__pyx_v_play_suscept[__pyx_v_j])));

              /* "metawards/extractors/_output_core.pyx":636
 *                     S_in_wards[j] += <int>(play_suscept[j])
 * 
 *                     if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]
 */
              __pyx_t_12 = (((__pyx_v_play_infections_i[__pyx_v_j]) > 0) != 0);
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":638
 *                     if play_infections_i[j] > 0:
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                     if total_new_inf_ward[j] != 0:
 */
                __pyx_t_13 = __pyx_v_j;
                (__pyx_v_total_new_inf_ward[__pyx_t_13]) = ((__pyx_v_total_new_inf_ward[__pyx_t_13]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":636
 *                     S_in_wards[j] += <int>(play_suscept[j])
 * 
 *                     if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                         # total_new_inf_ward[j] += play_infections[i][j]
 *                         total_new_inf_ward[j] += play_infections_i[j]
 */
              }

              /* "metawards/extractors/_output_core.pyx":640
 *                         total_new_inf_ward[j] += play_infections_i[j]
 * 
 *                     if total_new_inf_ward[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new += total_new_info_ward[j]
 *                         total_new_i += total_new_inf_ward[j]
 */
              __pyx_t_12 = (((__pyx_v_total_new_inf_ward[__pyx_v_j]) != 0) != 0);
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":642
 *                     if total_new_inf_ward[j] != 0:
 *                         # total_new += total_new_info_ward[j]
 *                         total_new_i += total_new_inf_ward[j]             # <<<<<<<<<<<<<<
 * 
 *                 if play_infections_i[j] > 0:
 */
                __pyx_v_total_new_i = (__pyx_v_total_new_i + (__pyx_v_total_new_inf_ward[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":640
 *                         total_new_inf_ward[j] += play_infections_i[j]
 * 
 *                     if total_new_inf_ward[j] != 0:             # <<<<<<<<<<<<<<
 *                         # total_new += total_new_info_ward[j]
 *                         total_new_i += total_new_inf_ward[j]
 */
              }

              /* "metawards/extractors/_output_core.pyx":630
 *             # from each ward
 *             for j in range(1, nnodes_plus_one):
 *                 if i == 0:             # <<<<<<<<<<<<<<
 *                     # susceptibles += wards[j].suscept
 *                     susceptibles_i += <int>(play_suscept[j])
 */
            }

            /* "metawards/extractors/_output_core.pyx":644
 *                         total_new_i += total_new_inf_ward[j]
 * 
 *                 if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     pinf_tot_i += play_infections_i[j]
 */
            __pyx_t_12 = (((__pyx_v_play_infections_i[__pyx_v_j]) > 0) != 0);
            if (__pyx_t_12) {

              /* "metawards/extractors/_output_core.pyx":646
 *                 if play_infections_i[j] > 0:
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     pinf_tot_i += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     # total_inf_ward[j] += play_infections[i][j]
 *                     total_inf_ward[j] += play_infections_i[j]
 */
              __pyx_v_pinf_tot_i = (__pyx_v_pinf_tot_i + (__pyx_v_play_infections_i[__pyx_v_j]));

              /* "metawards/extractors/_output_core.pyx":648
 *                     pinf_tot_i += play_infections_i[j]
 *                     # total_inf_ward[j] += play_infections[i][j]
 *                     total_inf_ward[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 */
              __pyx_t_13 = __pyx_v_j;
              (__pyx_v_total_inf_ward[__pyx_t_13]) = ((__pyx_v_total_inf_ward[__pyx_t_13]) + (__pyx_v_play_infections_i[__pyx_v_j]));

              /* "metawards/extractors/_output_core.pyx":650
 *                     total_inf_ward[j] += play_infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 */
              __pyx_t_14 = ((__pyx_v_i == 0) != 0);
              if (!__pyx_t_14) {
              } else {
                __pyx_t_12 = __pyx_t_14;
                goto __pyx_L25_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_i == __pyx_v_N_INF_CLASSES_MINUS_ONE) != 0);
              __pyx_t_12 = __pyx_t_14;
              __pyx_L25_bool_binop_done:;
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":651
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     elif i == 1:
 *                         E_in_wards[j] += play_infections_i[j]
 */
                __pyx_t_13 = __pyx_v_j;
                (__pyx_v_R_in_wards[__pyx_t_13]) = ((__pyx_v_R_in_wards[__pyx_t_13]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":650
 *                     total_inf_ward[j] += play_infections_i[j]
 * 
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:             # <<<<<<<<<<<<<<
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 */
                goto __pyx_L24;
              }

              /* "metawards/extractors/_output_core.pyx":652
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 */
              __pyx_t_12 = ((__pyx_v_i == 1) != 0);
              if (__pyx_t_12) {

                /* "metawards/extractors/_output_core.pyx":653
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:
 *                         E_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 *                     else:
 *                         I_in_wards[j] += play_infections_i[j]
 */
                __pyx_t_13 = __pyx_v_j;
                (__pyx_v_E_in_wards[__pyx_t_13]) = ((__pyx_v_E_in_wards[__pyx_t_13]) + (__pyx_v_play_infections_i[__pyx_v_j]));

                /* "metawards/extractors/_output_core.pyx":652
 *                     if i == 0 or i == N_INF_CLASSES_MINUS_ONE:
 *                         R_in_wards[j] += play_infections_i[j]
 *                     elif i == 1:             # <<<<<<<<<<<<<<
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 */
                goto __pyx_L24;
              }

              /* "metawards/extractors/_output_core.pyx":655
 *                         E_in_wards[j] += play_infections_i[j]
 *                     else:
 *                         I_in_wards[j] += play_infections_i[j]             # <<<<<<<<<<<<<<
 * 
 *                 if (i < N_INF_CLASSES_MINUS_ONE) and total_inf_ward[j] > 0:
 */
              /*else*/ {
                __pyx_t_13 = __pyx_v_j;
                (__pyx_v_I_in_wards[__pyx_t_13]) = ((__pyx_v_I_in_wards[__pyx_t_13]) + (__pyx_v_play_infections_i[__pyx_v_j]));
              }
              __pyx_L24:;

              /* "metawards/extractors/_output_core.pyx":644
 *                         total_new_i += total_new_inf_ward[j]
 * 
 *                 if play_infections_i[j] > 0:             # <<<<<<<<<<<<<<
 *                     # pinf_tot[i] += play_infections[i][j]
 *                     pinf_tot_i += play_infections_i[j]
 */
            }

            /* "metawards/extractors/_output_core.pyx":657
 *                         I_in_wards[j] += play_infections_i[j]
 * 
 *                 if (i < N_INF_CLASSES_MINUS_ONE) and total_inf_ward[j] > 0:             # <<<<<<<<<<<<<<
 *                     # n_inf_wards[i] += 1
 *                     n_inf_wards_i += 1
 */
            __pyx_t_14 = ((__pyx_v_i < __pyx_v_N_INF_CLASSES_MINUS_ONE) != 0);
            if (__pyx_t_14) {
            } else {
              __pyx_t_12 = __pyx_t_14;
              goto __pyx_L28_bool_binop_done;
            }
            __pyx_t_14 = (((__pyx_v_total_inf_ward[__pyx_v_j]) > 0) != 0);
            __pyx_t_12 = __pyx_t_14;
            __pyx_L28_bool_binop_done:;
            if (__pyx_t_12) {

              /* "metawards/extractors/_output_core.pyx":659
 *                 if (i < N_INF_CLASSES_MINUS_ONE) and total_inf_ward[j] > 0:
 *                     # n_inf_wards[i] += 1
 *                     n_inf_wards_i += 1             # <<<<<<<<<<<<<<
 *             # end of loop over nodes
 * 
 */
              __pyx_v_n_inf_wards_i = (__pyx_v_n_inf_wards_i + 1);

              /* "metawards/extractors/_output_core.pyx":657
 *                         I_in_wards[j] += play_infections_i[j]
 * 
 *                 if (i < N_INF_CLASSES_MINUS_ONE) and total_inf_ward[j] > 0:             # <<<<<<<<<<<<<<
 *                     # n_inf_wards[i] += 1
 *                     n_inf_wards_i += 1
 */
            }
          }

          /* "metawards/extractors/_output_core.pyx":662
 *             # end of loop over nodes
 * 
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 */
          __pyx_t_12 = ((__pyx_v_i == 2) != 0);
          if (__pyx_t_12) {

            /* "metawards/extractors/_output_core.pyx":665
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 *                 for j in range(1, nnodes_plus_one):             # <<<<<<<<<<<<<<
 *                     incidence[j] = total_inf_ward[j]
 * 
 */
            __pyx_t_9 = __pyx_v_nnodes_plus_one;
            __pyx_t_10 = __pyx_t_9;
            for (__pyx_t_11 = 1; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
              __pyx_v_j = __pyx_t_11;

              /* "metawards/extractors/_output_core.pyx":666
 *                 # is the incidence
 *                 for j in range(1, nnodes_plus_one):
 *                     incidence[j] = total_inf_ward[j]             # <<<<<<<<<<<<<<
 * 
 *         # end of nogil
 */
              (__pyx_v_incidence[__pyx_v_j]) = (__pyx_v_total_inf_ward[__pyx_v_j]);
            }

            /* "metawards/extractors/_output_core.pyx":662
 *             # end of loop over nodes
 * 
 *             if i == 2:             # <<<<<<<<<<<<<<
 *                 # save the sum of infections up to i <= 2. This
 *                 # is the incidence
 */
          }
        }

        /* "metawards/extractors/_output_core.pyx":597
 *         play_infections_i = get_int_array_ptr(play_infections[i])
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             #loop over all links and accumulate infections associated
 *             # with this link
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_FastGIL_Forget();
            Py_BLOCK_THREADS
            #endif
            goto __pyx_L9;
          }
          __pyx_L9:;
        }
    }

    /* "metawards/extractors/_output_core.pyx":672
 *         # can now reduce across all threads (don't need to lock as each
 *         # thread maintained its own running total)
 *         inf_tot[i] = inf_tot_i             # <<<<<<<<<<<<<<
 *         pinf_tot[i] = pinf_tot_i
 *         n_inf_wards[i] = n_inf_wards_i
 */
    (__pyx_v_inf_tot[__pyx_v_i]) = __pyx_v_inf_tot_i;

    /* "metawards/extractors/_output_core.pyx":673
 *         # thread maintained its own running total)
 *         inf_tot[i] = inf_tot_i
 *         pinf_tot[i] = pinf_tot_i             # <<<<<<<<<<<<<<
 *         n_inf_wards[i] = n_inf_wards_i
 * 
 */
    (__pyx_v_pinf_tot[__pyx_v_i]) = __pyx_v_pinf_tot_i;

    /* "metawards/extractors/_output_core.pyx":674
 *         inf_tot[i] = inf_tot_i
 *         pinf_tot[i] = pinf_tot_i
 *         n_inf_wards[i] = n_inf_wards_i             # <<<<<<<<<<<<<<
 * 
 *         # total_new += total_new[i]
 */
    (__pyx_v_n_inf_wards[__pyx_v_i]) = __pyx_v_n_inf_wards_i;

    /* "metawards/extractors/_output_core.pyx":677
 * 
 *         # total_new += total_new[i]
 *         total_new += total_new_i             # <<<<<<<<<<<<<<
 *         # susceptibles += susceptibles[i]
 *         susceptibles += susceptibles_i
 */
    __pyx_v_total_new = (__pyx_v_total_new + __pyx_v_total_new_i);

    /* "metawards/extractors/_output_core.pyx":679
 *         total_new += total_new_i
 *         # susceptibles += susceptibles[i]
 *         susceptibles += susceptibles_i             # <<<<<<<<<<<<<<
 *     # end of loop over i
 * 
 */
    __pyx_v_susceptibles = (__pyx_v_susceptibles + __pyx_v_susceptibles_i);
  }

  /* "metawards/extractors/_output_core.pyx":683
 * 
 *     # latent is the sum of work and play infections for stage 1
 *     latent = inf_tot[1] + pinf_tot[1]             # <<<<<<<<<<<<<<
 * 
 *     # total infections are sum of work and play for all intermediate
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_inf_tot[1]) + (__pyx_v_pinf_tot[1]))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_latent = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "metawards/extractors/_output_core.pyx":687
 *     # total infections are sum of work and play for all intermediate
 *     # stages
 *     total = 0             # <<<<<<<<<<<<<<
 *     for i in range(2, N_INF_CLASSES-1):
 *         total += inf_tot[i] + pinf_tot[i]
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_total = __pyx_int_0;

  /* "metawards/extractors/_output_core.pyx":688
 *     # stages
 *     total = 0
 *     for i in range(2, N_INF_CLASSES-1):             # <<<<<<<<<<<<<<
 *         total += inf_tot[i] + pinf_tot[i]
 * 
 */
  __pyx_t_15 = (__pyx_v_N_INF_CLASSES - 1);
  __pyx_t_16 = __pyx_t_15;
  for (__pyx_t_5 = 2; __pyx_t_5 < __pyx_t_16; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "metawards/extractors/_output_core.pyx":689
 *     total = 0
 *     for i in range(2, N_INF_CLASSES-1):
 *         total += inf_tot[i] + pinf_tot[i]             # <<<<<<<<<<<<<<
 * 
 *     # recovered is the sum of work and play for the first and last stage, e.g.
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_inf_tot[__pyx_v_i]) + (__pyx_v_pinf_tot[__pyx_v_i]))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_total, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_total, __pyx_t_4);
    __pyx_t_4 = 0;
  }

  /* "metawards/extractors/_output_core.pyx":695
 *     # (Note we don't have reverse indexingas these are plain C pointers)
 *     recovereds = inf_tot[0] + inf_tot[N_INF_CLASSES-1] + \
 *                  pinf_tot[0] + pinf_tot[N_INF_CLASSES-1]             # <<<<<<<<<<<<<<
 * 
 *     cdef int S = 0
 */
  __pyx_v_recovereds = ((((__pyx_v_inf_tot[0]) + (__pyx_v_inf_tot[(__pyx_v_N_INF_CLASSES - 1)])) + (__pyx_v_pinf_tot[0])) + (__pyx_v_pinf_tot[(__pyx_v_N_INF_CLASSES - 1)]));

  /* "metawards/extractors/_output_core.pyx":697
 *                  pinf_tot[0] + pinf_tot[N_INF_CLASSES-1]
 * 
 *     cdef int S = 0             # <<<<<<<<<<<<<<
 *     cdef int E = 0
 *     cdef int I = 0
 */
  __pyx_v_S = 0;

  /* "metawards/extractors/_output_core.pyx":698
 * 
 *     cdef int S = 0
 *     cdef int E = 0             # <<<<<<<<<<<<<<
 *     cdef int I = 0
 *     cdef int R = 0
 */
  __pyx_v_E = 0;

  /* "metawards/extractors/_output_core.pyx":699
 *     cdef int S = 0
 *     cdef int E = 0
 *     cdef int I = 0             # <<<<<<<<<<<<<<
 *     cdef int R = 0
 * 
 */
  __pyx_v_I = 0;

  /* "metawards/extractors/_output_core.pyx":700
 *     cdef int E = 0
 *     cdef int I = 0
 *     cdef int R = 0             # <<<<<<<<<<<<<<
 * 
 *     for j in range(1, nnodes_plus_one):
 */
  __pyx_v_R = 0;

  /* "metawards/extractors/_output_core.pyx":702
 *     cdef int R = 0
 * 
 *     for j in range(1, nnodes_plus_one):             # <<<<<<<<<<<<<<
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]
 */
  __pyx_t_5 = __pyx_v_nnodes_plus_one;
  __pyx_t_7 = __pyx_t_5;
  for (__pyx_t_8 = 1; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_j = __pyx_t_8;

    /* "metawards/extractors/_output_core.pyx":703
 * 
 *     for j in range(1, nnodes_plus_one):
 *         S += S_in_wards[j]             # <<<<<<<<<<<<<<
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]
 */
    __pyx_v_S = (__pyx_v_S + (__pyx_v_S_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":704
 *     for j in range(1, nnodes_plus_one):
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]             # <<<<<<<<<<<<<<
 *         I += I_in_wards[j]
 *         R += R_in_wards[j]
 */
    __pyx_v_E = (__pyx_v_E + (__pyx_v_E_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":705
 *         S += S_in_wards[j]
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]             # <<<<<<<<<<<<<<
 *         R += R_in_wards[j]
 * 
 */
    __pyx_v_I = (__pyx_v_I + (__pyx_v_I_in_wards[__pyx_v_j]));

    /* "metawards/extractors/_output_core.pyx":706
 *         E += E_in_wards[j]
 *         I += I_in_wards[j]
 *         R += R_in_wards[j]             # <<<<<<<<<<<<<<
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 */
    __pyx_v_R = (__pyx_v_R + (__pyx_v_R_in_wards[__pyx_v_j]));
  }

  /* "metawards/extractors/_output_core.pyx":708
 *         R += R_in_wards[j]
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:             # <<<<<<<<<<<<<<
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 */
  __pyx_t_14 = ((__pyx_v_S != __pyx_v_susceptibles) != 0);
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L38_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_E); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_v_latent, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L38_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_I); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_v_total, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_14) {
  } else {
    __pyx_t_12 = __pyx_t_14;
    goto __pyx_L38_bool_binop_done;
  }
  __pyx_t_14 = ((__pyx_v_R != __pyx_v_recovereds) != 0);
  __pyx_t_12 = __pyx_t_14;
  __pyx_L38_bool_binop_done:;
  if (__pyx_t_12) {

    /* "metawards/extractors/_output_core.pyx":710
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \             # <<<<<<<<<<<<<<
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"
 */
    __pyx_t_4 = PyTuple_New(16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_17 = 127;
    __Pyx_INCREF(__pyx_kp_u_Disagreement_in_accumulated_tota);
    __pyx_t_3 += 62;
    __Pyx_GIVEREF(__pyx_kp_u_Disagreement_in_accumulated_tota);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Disagreement_in_accumulated_tota);

    /* "metawards/extractors/_output_core.pyx":711
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \             # <<<<<<<<<<<<<<
 *             f"{R} vs {recovereds}"
 * 
 */
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_S, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_vs);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_susceptibles, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_E, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_kp_u_vs);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_latent, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_kp_u_);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_I, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 9, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 10, __pyx_kp_u_vs);
    __pyx_t_6 = __Pyx_PyObject_FormatSimple(__pyx_v_total, __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 11, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u_);
    PyTuple_SET_ITEM(__pyx_t_4, 12, __pyx_kp_u_);

    /* "metawards/extractors/_output_core.pyx":712
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"             # <<<<<<<<<<<<<<
 * 
 *         from ..utils._console import Console
 */
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_R, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 13, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_vs);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_vs);
    PyTuple_SET_ITEM(__pyx_t_4, 14, __pyx_kp_u_vs);
    __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_recovereds, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 15, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "metawards/extractors/_output_core.pyx":710
 *     if S != susceptibles or E != latent or I != total or R != recovereds:
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \             # <<<<<<<<<<<<<<
 *             f"{S} vs {susceptibles}, {E} vs {latent}, {I} vs {total}, " \
 *             f"{R} vs {recovereds}"
 */
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_4, 16, __pyx_t_3, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_error = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "metawards/extractors/_output_core.pyx":714
 *             f"{R} vs {recovereds}"
 * 
 *         from ..utils._console import Console             # <<<<<<<<<<<<<<
 *         Console.error(error)
 *         raise AssertionError(error)
 */
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_n_s_Console);
    __Pyx_GIVEREF(__pyx_n_s_Console);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_Console);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_utils__console, __pyx_t_6, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_Console); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_6);
    __pyx_v_Console = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":715
 * 
 *         from ..utils._console import Console
 *         Console.error(error)             # <<<<<<<<<<<<<<
 *         raise AssertionError(error)
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_v_error) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_error);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":716
 *         from ..utils._console import Console
 *         Console.error(error)
 *         raise AssertionError(error)             # <<<<<<<<<<<<<<
 * 
 *     if population is not None:
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_v_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 716, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":708
 *         R += R_in_wards[j]
 * 
 *     if S != susceptibles or E != latent or I != total or R != recovereds:             # <<<<<<<<<<<<<<
 *         error = \
 *             f"Disagreement in accumulated totals - indicates a program bug! " \
 */
  }

  /* "metawards/extractors/_output_core.pyx":718
 *         raise AssertionError(error)
 * 
 *     if population is not None:             # <<<<<<<<<<<<<<
 *         population.susceptibles = susceptibles
 *         population.total = total
 */
  __pyx_t_12 = (__pyx_v_population != Py_None);
  __pyx_t_14 = (__pyx_t_12 != 0);
  if (__pyx_t_14) {

    /* "metawards/extractors/_output_core.pyx":719
 * 
 *     if population is not None:
 *         population.susceptibles = susceptibles             # <<<<<<<<<<<<<<
 *         population.total = total
 *         population.recovereds = recovereds
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_susceptibles); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_susceptibles, __pyx_t_4) < 0) __PYX_ERR(0, 719, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":720
 *     if population is not None:
 *         population.susceptibles = susceptibles
 *         population.total = total             # <<<<<<<<<<<<<<
 *         population.recovereds = recovereds
 *         population.latent = latent
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_total, __pyx_v_total) < 0) __PYX_ERR(0, 720, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":721
 *         population.susceptibles = susceptibles
 *         population.total = total
 *         population.recovereds = recovereds             # <<<<<<<<<<<<<<
 *         population.latent = latent
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_recovereds); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_recovereds, __pyx_t_4) < 0) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":722
 *         population.total = total
 *         population.recovereds = recovereds
 *         population.latent = latent             # <<<<<<<<<<<<<<
 * 
 *         # save the number of wards that have at least one new
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_latent, __pyx_v_latent) < 0) __PYX_ERR(0, 722, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":726
 *         # save the number of wards that have at least one new
 *         #infection (index 0 is new infections)
 *         population.n_inf_wards = n_inf_wards[0]             # <<<<<<<<<<<<<<
 * 
 *     return total + latent
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_n_inf_wards[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_population, __pyx_n_s_n_inf_wards, __pyx_t_4) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":718
 *         raise AssertionError(error)
 * 
 *     if population is not None:             # <<<<<<<<<<<<<<
 *         population.susceptibles = susceptibles
 *         population.total = total
 */
  }

  /* "metawards/extractors/_output_core.pyx":728
 *         population.n_inf_wards = n_inf_wards[0]
 * 
 *     return total + latent             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyNumber_Add(__pyx_v_total, __pyx_v_latent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "metawards/extractors/_output_core.pyx":487
 * 
 * 
 * def output_core_serial(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                        workspace: Workspace,
 *                        infections: Infections,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core_serial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_links);
  __Pyx_XDECREF(__pyx_v_wards);
  __Pyx_XDECREF(__pyx_v_play_infections);
  __Pyx_XDECREF(__pyx_v_latent);
  __Pyx_XDECREF(__pyx_v_total);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_Console);
  __Pyx_XDECREF(__pyx_v_infections);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":731
 * 
 * 
 * def _safe_run(func, **kwargs):             # <<<<<<<<<<<<<<
 *     try:
 *         func(**kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_7_safe_run(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9metawards_10extractors_12_output_core_6_safe_run[] = "_safe_run(func, **kwargs)";
static PyMethodDef __pyx_mdef_9metawards_10extractors_12_output_core_7_safe_run = {"_safe_run", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9metawards_10extractors_12_output_core_7_safe_run, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9metawards_10extractors_12_output_core_6_safe_run};
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_7_safe_run(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_safe_run (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "_safe_run") < 0)) __PYX_ERR(0, 731, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_safe_run", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 731, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("metawards.extractors._output_core._safe_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9metawards_10extractors_12_output_core_6_safe_run(__pyx_self, __pyx_v_func, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_6_safe_run(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_func, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_Console = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_safe_run", 0);

  /* "metawards/extractors/_output_core.pyx":732
 * 
 * def _safe_run(func, **kwargs):
 *     try:             # <<<<<<<<<<<<<<
 *         func(**kwargs)
 *     except AssertionError as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "metawards/extractors/_output_core.pyx":733
 * def _safe_run(func, **kwargs):
 *     try:
 *         func(**kwargs)             # <<<<<<<<<<<<<<
 *     except AssertionError as e:
 *         if func is output_core_omp:
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_func, __pyx_empty_tuple, __pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 733, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "metawards/extractors/_output_core.pyx":732
 * 
 * def _safe_run(func, **kwargs):
 *     try:             # <<<<<<<<<<<<<<
 *         func(**kwargs)
 *     except AssertionError as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "metawards/extractors/_output_core.pyx":734
 *     try:
 *         func(**kwargs)
 *     except AssertionError as e:             # <<<<<<<<<<<<<<
 *         if func is output_core_omp:
 *             # this may be a race condition - rerun in serial
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("metawards.extractors._output_core._safe_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 734, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_e = __pyx_t_6;
      /*try:*/ {

        /* "metawards/extractors/_output_core.pyx":735
 *         func(**kwargs)
 *     except AssertionError as e:
 *         if func is output_core_omp:             # <<<<<<<<<<<<<<
 *             # this may be a race condition - rerun in serial
 *             from metawards.utils import Console
 */
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_output_core_omp); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 735, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = (__pyx_v_func == __pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = (__pyx_t_9 != 0);
        if (likely(__pyx_t_10)) {

          /* "metawards/extractors/_output_core.pyx":737
 *         if func is output_core_omp:
 *             # this may be a race condition - rerun in serial
 *             from metawards.utils import Console             # <<<<<<<<<<<<<<
 *             Console.print("Repeating this part of the calculation in serial "
 *                           "to correct this problem")
 */
          __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 737, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_n_s_Console);
          __Pyx_GIVEREF(__pyx_n_s_Console);
          PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_Console);
          __pyx_t_11 = __Pyx_Import(__pyx_n_s_metawards_utils, __pyx_t_8, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 737, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_11, __pyx_n_s_Console); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 737, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_8);
          __pyx_v_Console = __pyx_t_8;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "metawards/extractors/_output_core.pyx":738
 *             # this may be a race condition - rerun in serial
 *             from metawards.utils import Console
 *             Console.print("Repeating this part of the calculation in serial "             # <<<<<<<<<<<<<<
 *                           "to correct this problem")
 *             kwargs["nthreads"] = 1
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_print); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 738, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_11 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_12, __pyx_kp_u_Repeating_this_part_of_the_calcu) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u_Repeating_this_part_of_the_calcu);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 738, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "metawards/extractors/_output_core.pyx":740
 *             Console.print("Repeating this part of the calculation in serial "
 *                           "to correct this problem")
 *             kwargs["nthreads"] = 1             # <<<<<<<<<<<<<<
 *             output_core_serial(**kwargs)
 *             Console.print("Calculation successful, the bug has been mitigated")
 */
          if (unlikely(PyDict_SetItem(__pyx_v_kwargs, __pyx_n_u_nthreads, __pyx_int_1) < 0)) __PYX_ERR(0, 740, __pyx_L14_error)

          /* "metawards/extractors/_output_core.pyx":741
 *                           "to correct this problem")
 *             kwargs["nthreads"] = 1
 *             output_core_serial(**kwargs)             # <<<<<<<<<<<<<<
 *             Console.print("Calculation successful, the bug has been mitigated")
 *         else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_output_core_serial); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 741, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_empty_tuple, __pyx_v_kwargs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 741, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "metawards/extractors/_output_core.pyx":742
 *             kwargs["nthreads"] = 1
 *             output_core_serial(**kwargs)
 *             Console.print("Calculation successful, the bug has been mitigated")             # <<<<<<<<<<<<<<
 *         else:
 *             # the error was in serial, so big problem
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_print); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 742, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_u_Calculation_successful_the_bug_h) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_u_Calculation_successful_the_bug_h);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 742, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "metawards/extractors/_output_core.pyx":735
 *         func(**kwargs)
 *     except AssertionError as e:
 *         if func is output_core_omp:             # <<<<<<<<<<<<<<
 *             # this may be a race condition - rerun in serial
 *             from metawards.utils import Console
 */
          goto __pyx_L16;
        }

        /* "metawards/extractors/_output_core.pyx":745
 *         else:
 *             # the error was in serial, so big problem
 *             raise e             # <<<<<<<<<<<<<<
 * 
 * 
 */
        /*else*/ {
          __Pyx_Raise(__pyx_v_e, 0, 0, 0);
          __PYX_ERR(0, 745, __pyx_L14_error)
        }
        __pyx_L16:;
      }

      /* "metawards/extractors/_output_core.pyx":734
 *     try:
 *         func(**kwargs)
 *     except AssertionError as e:             # <<<<<<<<<<<<<<
 *         if func is output_core_omp:
 *             # this may be a race condition - rerun in serial
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e);
          __pyx_v_e = NULL;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_5 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e);
            __pyx_v_e = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "metawards/extractors/_output_core.pyx":732
 * 
 * def _safe_run(func, **kwargs):
 *     try:             # <<<<<<<<<<<<<<
 *         func(**kwargs)
 *     except AssertionError as e:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "metawards/extractors/_output_core.pyx":731
 * 
 * 
 * def _safe_run(func, **kwargs):             # <<<<<<<<<<<<<<
 *     try:
 *         func(**kwargs)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("metawards.extractors._output_core._safe_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_Console);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "metawards/extractors/_output_core.pyx":748
 * 
 * 
 * def output_core(network: _Union[Network, Networks],             # <<<<<<<<<<<<<<
 *                 population: Population,
 *                 workspace: Workspace,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_9output_core(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9metawards_10extractors_12_output_core_8output_core[] = "output_core(network: _Union[Network, Networks], population: Population, workspace: Workspace, infections: Infections, nthreads: int, profiler: Profiler, **kwargs)\nThis is the core output function that must be called\n       every iteration as it is responsible for accumulating\n       the core data each day, which is used to report a summary\n       to the main output file. This is the serial version\n       of this function.\n\n       Parameters\n       ----------\n       network: Network\n         The network over which the outbreak is being modelled\n       population: Population\n         The population experiencing the outbreak\n       workspace: Workspace\n         A workspace that can be used to extract data\n       infections: Infections\n         All of the infections that have been recorded\n       nthreads: int\n         The number of threads to use to help extract the data\n       kwargs\n         Extra argumentst that are ignored by this function\n    ";
static PyMethodDef __pyx_mdef_9metawards_10extractors_12_output_core_9output_core = {"output_core", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9metawards_10extractors_12_output_core_9output_core, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9metawards_10extractors_12_output_core_8output_core};
static PyObject *__pyx_pw_9metawards_10extractors_12_output_core_9output_core(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_network = 0;
  PyObject *__pyx_v_population = 0;
  PyObject *__pyx_v_workspace = 0;
  PyObject *__pyx_v_infections = 0;
  PyObject *__pyx_v_nthreads = 0;
  PyObject *__pyx_v_profiler = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output_core (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_network,&__pyx_n_s_population,&__pyx_n_s_workspace,&__pyx_n_s_infections,&__pyx_n_s_nthreads,&__pyx_n_s_profiler,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_network)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_population)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, 1); __PYX_ERR(0, 748, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, 2); __PYX_ERR(0, 748, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_infections)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, 3); __PYX_ERR(0, 748, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nthreads)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, 4); __PYX_ERR(0, 748, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_profiler)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, 5); __PYX_ERR(0, 748, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "output_core") < 0)) __PYX_ERR(0, 748, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_network = values[0];
    __pyx_v_population = values[1];
    __pyx_v_workspace = values[2];
    __pyx_v_infections = values[3];
    __pyx_v_nthreads = values[4];
    __pyx_v_profiler = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("output_core", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 748, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9metawards_10extractors_12_output_core_8output_core(__pyx_self, __pyx_v_network, __pyx_v_population, __pyx_v_workspace, __pyx_v_infections, __pyx_v_nthreads, __pyx_v_profiler, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9metawards_10extractors_12_output_core_8output_core(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_network, PyObject *__pyx_v_population, PyObject *__pyx_v_workspace, PyObject *__pyx_v_infections, PyObject *__pyx_v_nthreads, PyObject *__pyx_v_profiler, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_output_func = NULL;
  PyObject *__pyx_v_Console = NULL;
  PyObject *__pyx_v_NullProfiler = NULL;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_subnet = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("output_core", 0);
  __Pyx_INCREF(__pyx_v_profiler);

  /* "metawards/extractors/_output_core.pyx":779
 *     # serial version is much faster than parallel - only worth
 *     # parallelising when more than 4 cores
 *     if nthreads <= 4:             # <<<<<<<<<<<<<<
 *         output_func = output_core_serial
 *     else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_nthreads, __pyx_int_4, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "metawards/extractors/_output_core.pyx":780
 *     # parallelising when more than 4 cores
 *     if nthreads <= 4:
 *         output_func = output_core_serial             # <<<<<<<<<<<<<<
 *     else:
 *         kwargs["nthreads"] = nthreads
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_output_core_serial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_output_func = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":779
 *     # serial version is much faster than parallel - only worth
 *     # parallelising when more than 4 cores
 *     if nthreads <= 4:             # <<<<<<<<<<<<<<
 *         output_func = output_core_serial
 *     else:
 */
    goto __pyx_L3;
  }

  /* "metawards/extractors/_output_core.pyx":782
 *         output_func = output_core_serial
 *     else:
 *         kwargs["nthreads"] = nthreads             # <<<<<<<<<<<<<<
 *         output_func = output_core_omp
 * 
 */
  /*else*/ {
    if (unlikely(PyDict_SetItem(__pyx_v_kwargs, __pyx_n_u_nthreads, __pyx_v_nthreads) < 0)) __PYX_ERR(0, 782, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":783
 *     else:
 *         kwargs["nthreads"] = nthreads
 *         output_func = output_core_omp             # <<<<<<<<<<<<<<
 * 
 *     from ..utils._console import Console
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_output_core_omp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_output_func = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "metawards/extractors/_output_core.pyx":785
 *         output_func = output_core_omp
 * 
 *     from ..utils._console import Console             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(network, Network):
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Console);
  __Pyx_GIVEREF(__pyx_n_s_Console);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Console);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_utils__console, __pyx_t_1, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Console); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Console = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "metawards/extractors/_output_core.pyx":787
 *     from ..utils._console import Console
 * 
 *     if isinstance(network, Network):             # <<<<<<<<<<<<<<
 *         _safe_run(func=output_func,
 *                   network=network, population=population,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Network); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_network, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (__pyx_t_4) {

    /* "metawards/extractors/_output_core.pyx":788
 * 
 *     if isinstance(network, Network):
 *         _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                   network=network, population=population,
 *                   workspace=workspace, infections=infections,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_safe_run); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_func, __pyx_v_output_func) < 0) __PYX_ERR(0, 788, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":789
 *     if isinstance(network, Network):
 *         _safe_run(func=output_func,
 *                   network=network, population=population,             # <<<<<<<<<<<<<<
 *                   workspace=workspace, infections=infections,
 *                   profiler=profiler, **kwargs)
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_network, __pyx_v_network) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_population, __pyx_v_population) < 0) __PYX_ERR(0, 788, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":790
 *         _safe_run(func=output_func,
 *                   network=network, population=population,
 *                   workspace=workspace, infections=infections,             # <<<<<<<<<<<<<<
 *                   profiler=profiler, **kwargs)
 * 
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_workspace, __pyx_v_workspace) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_infections, __pyx_v_infections) < 0) __PYX_ERR(0, 788, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":791
 *                   network=network, population=population,
 *                   workspace=workspace, infections=infections,
 *                   profiler=profiler, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *         Console.print_population(population)
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_profiler, __pyx_v_profiler) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;
    if (__Pyx_MergeKeywords(__pyx_t_1, __pyx_v_kwargs) < 0) __PYX_ERR(0, 791, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":788
 * 
 *     if isinstance(network, Network):
 *         _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                   network=network, population=population,
 *                   workspace=workspace, infections=infections,
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "metawards/extractors/_output_core.pyx":793
 *                   profiler=profiler, **kwargs)
 * 
 *         Console.print_population(population)             # <<<<<<<<<<<<<<
 * 
 *     elif isinstance(network, Networks):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_print_population); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_population) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_population);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "metawards/extractors/_output_core.pyx":787
 *     from ..utils._console import Console
 * 
 *     if isinstance(network, Network):             # <<<<<<<<<<<<<<
 *         _safe_run(func=output_func,
 *                   network=network, population=population,
 */
    goto __pyx_L4;
  }

  /* "metawards/extractors/_output_core.pyx":795
 *         Console.print_population(population)
 * 
 *     elif isinstance(network, Networks):             # <<<<<<<<<<<<<<
 *         if profiler is None:
 *             from ..utils._profiler import NullProfiler
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Networks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_network, __pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = (__pyx_t_4 != 0);
  if (__pyx_t_2) {

    /* "metawards/extractors/_output_core.pyx":796
 * 
 *     elif isinstance(network, Networks):
 *         if profiler is None:             # <<<<<<<<<<<<<<
 *             from ..utils._profiler import NullProfiler
 *             profiler = NullProfiler()
 */
    __pyx_t_2 = (__pyx_v_profiler == Py_None);
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (__pyx_t_4) {

      /* "metawards/extractors/_output_core.pyx":797
 *     elif isinstance(network, Networks):
 *         if profiler is None:
 *             from ..utils._profiler import NullProfiler             # <<<<<<<<<<<<<<
 *             profiler = NullProfiler()
 * 
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_s_NullProfiler);
      __Pyx_GIVEREF(__pyx_n_s_NullProfiler);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_NullProfiler);
      __pyx_t_1 = __Pyx_Import(__pyx_n_s_utils__profiler, __pyx_t_5, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_NullProfiler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_NullProfiler = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "metawards/extractors/_output_core.pyx":798
 *         if profiler is None:
 *             from ..utils._profiler import NullProfiler
 *             profiler = NullProfiler()             # <<<<<<<<<<<<<<
 * 
 *         p = profiler.start("multi-network")
 */
      __Pyx_INCREF(__pyx_v_NullProfiler);
      __pyx_t_5 = __pyx_v_NullProfiler; __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_profiler, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "metawards/extractors/_output_core.pyx":796
 * 
 *     elif isinstance(network, Networks):
 *         if profiler is None:             # <<<<<<<<<<<<<<
 *             from ..utils._profiler import NullProfiler
 *             profiler = NullProfiler()
 */
    }

    /* "metawards/extractors/_output_core.pyx":800
 *             profiler = NullProfiler()
 * 
 *         p = profiler.start("multi-network")             # <<<<<<<<<<<<<<
 * 
 *         if population.subpops is None:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_profiler, __pyx_n_s_start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_kp_u_multi_network) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_multi_network);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_p = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":802
 *         p = profiler.start("multi-network")
 * 
 *         if population.subpops is None:             # <<<<<<<<<<<<<<
 *             population.specialise(network)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_population, __pyx_n_s_subpops); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__pyx_t_1 == Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_4 != 0);
    if (__pyx_t_2) {

      /* "metawards/extractors/_output_core.pyx":803
 * 
 *         if population.subpops is None:
 *             population.specialise(network)             # <<<<<<<<<<<<<<
 * 
 *         for i, subnet in enumerate(network.subnets):
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_population, __pyx_n_s_specialise); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_network) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_network);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "metawards/extractors/_output_core.pyx":802
 *         p = profiler.start("multi-network")
 * 
 *         if population.subpops is None:             # <<<<<<<<<<<<<<
 *             population.specialise(network)
 * 
 */
    }

    /* "metawards/extractors/_output_core.pyx":805
 *             population.specialise(network)
 * 
 *         for i, subnet in enumerate(network.subnets):             # <<<<<<<<<<<<<<
 *             p = p.start(f"output-{i}")
 *             _safe_run(func=output_func,
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_1 = __pyx_int_0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_subnets); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_3 = __pyx_t_5; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 805, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 805, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 805, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 805, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 805, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 805, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 805, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_subnet, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
      __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 805, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "metawards/extractors/_output_core.pyx":806
 * 
 *         for i, subnet in enumerate(network.subnets):
 *             p = p.start(f"output-{i}")             # <<<<<<<<<<<<<<
 *             _safe_run(func=output_func,
 *                       network=subnet,
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_i, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_kp_u_output, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 806, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "metawards/extractors/_output_core.pyx":807
 *         for i, subnet in enumerate(network.subnets):
 *             p = p.start(f"output-{i}")
 *             _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                       network=subnet,
 *                       population=population.subpops[i],
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_safe_run); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_func, __pyx_v_output_func) < 0) __PYX_ERR(0, 807, __pyx_L1_error)

      /* "metawards/extractors/_output_core.pyx":808
 *             p = p.start(f"output-{i}")
 *             _safe_run(func=output_func,
 *                       network=subnet,             # <<<<<<<<<<<<<<
 *                       population=population.subpops[i],
 *                       workspace=workspace.subspaces[i],
 */
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_network, __pyx_v_subnet) < 0) __PYX_ERR(0, 807, __pyx_L1_error)

      /* "metawards/extractors/_output_core.pyx":809
 *             _safe_run(func=output_func,
 *                       network=subnet,
 *                       population=population.subpops[i],             # <<<<<<<<<<<<<<
 *                       workspace=workspace.subspaces[i],
 *                       infections=infections.subinfs[i],
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_population, __pyx_n_s_subpops); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_i); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_population, __pyx_t_11) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "metawards/extractors/_output_core.pyx":810
 *                       network=subnet,
 *                       population=population.subpops[i],
 *                       workspace=workspace.subspaces[i],             # <<<<<<<<<<<<<<
 *                       infections=infections.subinfs[i],
 *                       profiler=p,
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_workspace, __pyx_n_s_subspaces); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_workspace, __pyx_t_9) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "metawards/extractors/_output_core.pyx":811
 *                       population=population.subpops[i],
 *                       workspace=workspace.subspaces[i],
 *                       infections=infections.subinfs[i],             # <<<<<<<<<<<<<<
 *                       profiler=p,
 *                       **kwargs)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_subinfs); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_i); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_infections, __pyx_t_11) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "metawards/extractors/_output_core.pyx":812
 *                       workspace=workspace.subspaces[i],
 *                       infections=infections.subinfs[i],
 *                       profiler=p,             # <<<<<<<<<<<<<<
 *                       **kwargs)
 *             p = p.stop()
 */
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_profiler, __pyx_v_p) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
      __pyx_t_8 = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "metawards/extractors/_output_core.pyx":813
 *                       infections=infections.subinfs[i],
 *                       profiler=p,
 *                       **kwargs)             # <<<<<<<<<<<<<<
 *             p = p.stop()
 * 
 */
      if (__Pyx_MergeKeywords(__pyx_t_8, __pyx_v_kwargs) < 0) __PYX_ERR(0, 813, __pyx_L1_error)

      /* "metawards/extractors/_output_core.pyx":807
 *         for i, subnet in enumerate(network.subnets):
 *             p = p.start(f"output-{i}")
 *             _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                       network=subnet,
 *                       population=population.subpops[i],
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "metawards/extractors/_output_core.pyx":814
 *                       profiler=p,
 *                       **kwargs)
 *             p = p.stop()             # <<<<<<<<<<<<<<
 * 
 *         # aggregate the infection information from across
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_stop); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "metawards/extractors/_output_core.pyx":805
 *             population.specialise(network)
 * 
 *         for i, subnet in enumerate(network.subnets):             # <<<<<<<<<<<<<<
 *             p = p.start(f"output-{i}")
 *             _safe_run(func=output_func,
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":818
 *         # aggregate the infection information from across
 *         # the different demographics
 *         p = p.start("aggregate")             # <<<<<<<<<<<<<<
 *         infections.aggregate(profiler=p, nthreads=nthreads)
 *         network.aggregate(profiler=p, nthreads=nthreads)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_n_u_aggregate) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_aggregate);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 818, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":819
 *         # the different demographics
 *         p = p.start("aggregate")
 *         infections.aggregate(profiler=p, nthreads=nthreads)             # <<<<<<<<<<<<<<
 *         network.aggregate(profiler=p, nthreads=nthreads)
 *         p = p.stop()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_infections, __pyx_n_s_aggregate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_profiler, __pyx_v_p) < 0) __PYX_ERR(0, 819, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_nthreads, __pyx_v_nthreads) < 0) __PYX_ERR(0, 819, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "metawards/extractors/_output_core.pyx":820
 *         p = p.start("aggregate")
 *         infections.aggregate(profiler=p, nthreads=nthreads)
 *         network.aggregate(profiler=p, nthreads=nthreads)             # <<<<<<<<<<<<<<
 *         p = p.stop()
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_aggregate); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_profiler, __pyx_v_p) < 0) __PYX_ERR(0, 820, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_nthreads, __pyx_v_nthreads) < 0) __PYX_ERR(0, 820, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":821
 *         infections.aggregate(profiler=p, nthreads=nthreads)
 *         network.aggregate(profiler=p, nthreads=nthreads)
 *         p = p.stop()             # <<<<<<<<<<<<<<
 * 
 *         p = p.start("overall_output")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_stop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":823
 *         p = p.stop()
 * 
 *         p = p.start("overall_output")             # <<<<<<<<<<<<<<
 *         _safe_run(func=output_func,
 *                   network=network.overall,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_n_u_overall_output) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_u_overall_output);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":824
 * 
 *         p = p.start("overall_output")
 *         _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                   network=network.overall,
 *                   population=population,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_safe_run); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_func, __pyx_v_output_func) < 0) __PYX_ERR(0, 824, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":825
 *         p = p.start("overall_output")
 *         _safe_run(func=output_func,
 *                   network=network.overall,             # <<<<<<<<<<<<<<
 *                   population=population,
 *                   workspace=workspace,
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_overall); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 825, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_network, __pyx_t_8) < 0) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "metawards/extractors/_output_core.pyx":826
 *         _safe_run(func=output_func,
 *                   network=network.overall,
 *                   population=population,             # <<<<<<<<<<<<<<
 *                   workspace=workspace,
 *                   infections=infections,
 */
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_population, __pyx_v_population) < 0) __PYX_ERR(0, 824, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":827
 *                   network=network.overall,
 *                   population=population,
 *                   workspace=workspace,             # <<<<<<<<<<<<<<
 *                   infections=infections,
 *                   profiler=profiler,
 */
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_workspace, __pyx_v_workspace) < 0) __PYX_ERR(0, 824, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":828
 *                   population=population,
 *                   workspace=workspace,
 *                   infections=infections,             # <<<<<<<<<<<<<<
 *                   profiler=profiler,
 *                   **kwargs)
 */
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_infections, __pyx_v_infections) < 0) __PYX_ERR(0, 824, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":829
 *                   workspace=workspace,
 *                   infections=infections,
 *                   profiler=profiler,             # <<<<<<<<<<<<<<
 *                   **kwargs)
 *         p = p.stop()
 */
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_profiler, __pyx_v_profiler) < 0) __PYX_ERR(0, 824, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "metawards/extractors/_output_core.pyx":830
 *                   infections=infections,
 *                   profiler=profiler,
 *                   **kwargs)             # <<<<<<<<<<<<<<
 *         p = p.stop()
 * 
 */
    if (__Pyx_MergeKeywords(__pyx_t_3, __pyx_v_kwargs) < 0) __PYX_ERR(0, 830, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":824
 * 
 *         p = p.start("overall_output")
 *         _safe_run(func=output_func,             # <<<<<<<<<<<<<<
 *                   network=network.overall,
 *                   population=population,
 */
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "metawards/extractors/_output_core.pyx":831
 *                   profiler=profiler,
 *                   **kwargs)
 *         p = p.stop()             # <<<<<<<<<<<<<<
 * 
 *         Console.print_population(population=population,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_stop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_10 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "metawards/extractors/_output_core.pyx":833
 *         p = p.stop()
 * 
 *         Console.print_population(population=population,             # <<<<<<<<<<<<<<
 *                                  demographics=network.demographics)
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_Console, __pyx_n_s_print_population); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_population, __pyx_v_population) < 0) __PYX_ERR(0, 833, __pyx_L1_error)

    /* "metawards/extractors/_output_core.pyx":834
 * 
 *         Console.print_population(population=population,
 *                                  demographics=network.demographics)             # <<<<<<<<<<<<<<
 * 
 *         # double-check that the sums all add up correctly
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_network, __pyx_n_s_demographics); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_demographics, __pyx_t_1) < 0) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":833
 *         p = p.stop()
 * 
 *         Console.print_population(population=population,             # <<<<<<<<<<<<<<
 *                                  demographics=network.demographics)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":837
 * 
 *         # double-check that the sums all add up correctly
 *         population.assert_sane()             # <<<<<<<<<<<<<<
 * 
 *         p = p.stop()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_population, __pyx_n_s_assert_sane); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":839
 *         population.assert_sane()
 * 
 *         p = p.stop()             # <<<<<<<<<<<<<<
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_stop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_p, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "metawards/extractors/_output_core.pyx":795
 *         Console.print_population(population)
 * 
 *     elif isinstance(network, Networks):             # <<<<<<<<<<<<<<
 *         if profiler is None:
 *             from ..utils._profiler import NullProfiler
 */
  }
  __pyx_L4:;

  /* "metawards/extractors/_output_core.pyx":748
 * 
 * 
 * def output_core(network: _Union[Network, Networks],             # <<<<<<<<<<<<<<
 *                 population: Population,
 *                 workspace: Workspace,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("metawards.extractors._output_core.output_core", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_output_func);
  __Pyx_XDECREF(__pyx_v_Console);
  __Pyx_XDECREF(__pyx_v_NullProfiler);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_subnet);
  __Pyx_XDECREF(__pyx_v_profiler);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__output_core(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__output_core},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_output_core",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_kp_u_Calculation_successful_the_bug_h, __pyx_k_Calculation_successful_the_bug_h, sizeof(__pyx_k_Calculation_successful_the_bug_h), 0, 1, 0, 0},
  {&__pyx_n_s_Console, __pyx_k_Console, sizeof(__pyx_k_Console), 0, 0, 1, 1},
  {&__pyx_kp_u_Disagreement_in_accumulated_tota, __pyx_k_Disagreement_in_accumulated_tota, sizeof(__pyx_k_Disagreement_in_accumulated_tota), 0, 1, 0, 0},
  {&__pyx_n_s_E, __pyx_k_E, sizeof(__pyx_k_E), 0, 0, 1, 1},
  {&__pyx_n_s_E_buffer, __pyx_k_E_buffer, sizeof(__pyx_k_E_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_E_in_wards, __pyx_k_E_in_wards, sizeof(__pyx_k_E_in_wards), 0, 0, 1, 1},
  {&__pyx_n_s_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 0, 1, 1},
  {&__pyx_n_s_I_buffer, __pyx_k_I_buffer, sizeof(__pyx_k_I_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_I_in_wards, __pyx_k_I_in_wards, sizeof(__pyx_k_I_in_wards), 0, 0, 1, 1},
  {&__pyx_n_s_Infections, __pyx_k_Infections, sizeof(__pyx_k_Infections), 0, 0, 1, 1},
  {&__pyx_n_s_N_INF_CLASSES, __pyx_k_N_INF_CLASSES, sizeof(__pyx_k_N_INF_CLASSES), 0, 0, 1, 1},
  {&__pyx_n_s_N_INF_CLASSES_MINUS_ONE, __pyx_k_N_INF_CLASSES_MINUS_ONE, sizeof(__pyx_k_N_INF_CLASSES_MINUS_ONE), 0, 0, 1, 1},
  {&__pyx_n_s_Network, __pyx_k_Network, sizeof(__pyx_k_Network), 0, 0, 1, 1},
  {&__pyx_n_s_Networks, __pyx_k_Networks, sizeof(__pyx_k_Networks), 0, 0, 1, 1},
  {&__pyx_n_s_NullProfiler, __pyx_k_NullProfiler, sizeof(__pyx_k_NullProfiler), 0, 0, 1, 1},
  {&__pyx_n_s_Population, __pyx_k_Population, sizeof(__pyx_k_Population), 0, 0, 1, 1},
  {&__pyx_n_s_Profiler, __pyx_k_Profiler, sizeof(__pyx_k_Profiler), 0, 0, 1, 1},
  {&__pyx_n_s_R, __pyx_k_R, sizeof(__pyx_k_R), 0, 0, 1, 1},
  {&__pyx_n_s_R_buffer, __pyx_k_R_buffer, sizeof(__pyx_k_R_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_R_in_wards, __pyx_k_R_in_wards, sizeof(__pyx_k_R_in_wards), 0, 0, 1, 1},
  {&__pyx_kp_u_Repeating_this_part_of_the_calcu, __pyx_k_Repeating_this_part_of_the_calcu, sizeof(__pyx_k_Repeating_this_part_of_the_calcu), 0, 1, 0, 0},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_n_s_S_buffer, __pyx_k_S_buffer, sizeof(__pyx_k_S_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_S_in_wards, __pyx_k_S_in_wards, sizeof(__pyx_k_S_in_wards), 0, 0, 1, 1},
  {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
  {&__pyx_n_s_Union_2, __pyx_k_Union_2, sizeof(__pyx_k_Union_2), 0, 0, 1, 1},
  {&__pyx_n_s_Workspace, __pyx_k_Workspace, sizeof(__pyx_k_Workspace), 0, 0, 1, 1},
  {&__pyx_n_s_aggregate, __pyx_k_aggregate, sizeof(__pyx_k_aggregate), 0, 0, 1, 1},
  {&__pyx_n_u_aggregate, __pyx_k_aggregate, sizeof(__pyx_k_aggregate), 0, 1, 0, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_assert_sane, __pyx_k_assert_sane, sizeof(__pyx_k_assert_sane), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_demographics, __pyx_k_demographics, sizeof(__pyx_k_demographics), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_files, __pyx_k_files, sizeof(__pyx_k_files), 0, 0, 1, 1},
  {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_ifrom, __pyx_k_ifrom, sizeof(__pyx_k_ifrom), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_incidence, __pyx_k_incidence, sizeof(__pyx_k_incidence), 0, 0, 1, 1},
  {&__pyx_n_s_inf_tot, __pyx_k_inf_tot, sizeof(__pyx_k_inf_tot), 0, 0, 1, 1},
  {&__pyx_n_s_inf_tot_i, __pyx_k_inf_tot_i, sizeof(__pyx_k_inf_tot_i), 0, 0, 1, 1},
  {&__pyx_n_s_infections, __pyx_k_infections, sizeof(__pyx_k_infections), 0, 0, 1, 1},
  {&__pyx_n_s_infections_2, __pyx_k_infections_2, sizeof(__pyx_k_infections_2), 0, 0, 1, 1},
  {&__pyx_n_s_infections_i, __pyx_k_infections_i, sizeof(__pyx_k_infections_i), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_latent, __pyx_k_latent, sizeof(__pyx_k_latent), 0, 0, 1, 1},
  {&__pyx_n_s_links, __pyx_k_links, sizeof(__pyx_k_links), 0, 0, 1, 1},
  {&__pyx_n_s_links_ifrom, __pyx_k_links_ifrom, sizeof(__pyx_k_links_ifrom), 0, 0, 1, 1},
  {&__pyx_n_s_links_suscept, __pyx_k_links_suscept, sizeof(__pyx_k_links_suscept), 0, 0, 1, 1},
  {&__pyx_n_s_lock, __pyx_k_lock, sizeof(__pyx_k_lock), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metawards_extractors__output_cor, __pyx_k_metawards_extractors__output_cor, sizeof(__pyx_k_metawards_extractors__output_cor), 0, 0, 1, 1},
  {&__pyx_n_s_metawards_utils, __pyx_k_metawards_utils, sizeof(__pyx_k_metawards_utils), 0, 0, 1, 1},
  {&__pyx_kp_u_multi_network, __pyx_k_multi_network, sizeof(__pyx_k_multi_network), 0, 1, 0, 0},
  {&__pyx_n_s_n_inf_wards, __pyx_k_n_inf_wards, sizeof(__pyx_k_n_inf_wards), 0, 0, 1, 1},
  {&__pyx_n_s_n_inf_wards_i, __pyx_k_n_inf_wards_i, sizeof(__pyx_k_n_inf_wards_i), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_network, __pyx_k_network, sizeof(__pyx_k_network), 0, 0, 1, 1},
  {&__pyx_n_s_network_2, __pyx_k_network_2, sizeof(__pyx_k_network_2), 0, 0, 1, 1},
  {&__pyx_n_s_networks, __pyx_k_networks, sizeof(__pyx_k_networks), 0, 0, 1, 1},
  {&__pyx_n_s_nlinks, __pyx_k_nlinks, sizeof(__pyx_k_nlinks), 0, 0, 1, 1},
  {&__pyx_n_s_nlinks_plus_one, __pyx_k_nlinks_plus_one, sizeof(__pyx_k_nlinks_plus_one), 0, 0, 1, 1},
  {&__pyx_n_s_nnodes, __pyx_k_nnodes, sizeof(__pyx_k_nnodes), 0, 0, 1, 1},
  {&__pyx_n_s_nnodes_plus_one, __pyx_k_nnodes_plus_one, sizeof(__pyx_k_nnodes_plus_one), 0, 0, 1, 1},
  {&__pyx_n_s_nodes, __pyx_k_nodes, sizeof(__pyx_k_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_nthreads, __pyx_k_nthreads, sizeof(__pyx_k_nthreads), 0, 0, 1, 1},
  {&__pyx_n_u_nthreads, __pyx_k_nthreads, sizeof(__pyx_k_nthreads), 0, 1, 0, 1},
  {&__pyx_n_s_num_threads, __pyx_k_num_threads, sizeof(__pyx_k_num_threads), 0, 0, 1, 1},
  {&__pyx_kp_u_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 1, 0, 0},
  {&__pyx_n_s_output_core, __pyx_k_output_core, sizeof(__pyx_k_output_core), 0, 0, 1, 1},
  {&__pyx_n_u_output_core, __pyx_k_output_core, sizeof(__pyx_k_output_core), 0, 1, 0, 1},
  {&__pyx_n_s_output_core_omp, __pyx_k_output_core_omp, sizeof(__pyx_k_output_core_omp), 0, 0, 1, 1},
  {&__pyx_n_u_output_core_omp, __pyx_k_output_core_omp, sizeof(__pyx_k_output_core_omp), 0, 1, 0, 1},
  {&__pyx_n_s_output_core_serial, __pyx_k_output_core_serial, sizeof(__pyx_k_output_core_serial), 0, 0, 1, 1},
  {&__pyx_n_u_output_core_serial, __pyx_k_output_core_serial, sizeof(__pyx_k_output_core_serial), 0, 1, 0, 1},
  {&__pyx_n_s_output_func, __pyx_k_output_func, sizeof(__pyx_k_output_func), 0, 0, 1, 1},
  {&__pyx_n_s_overall, __pyx_k_overall, sizeof(__pyx_k_overall), 0, 0, 1, 1},
  {&__pyx_n_u_overall_output, __pyx_k_overall_output, sizeof(__pyx_k_overall_output), 0, 1, 0, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_pinf_tot, __pyx_k_pinf_tot, sizeof(__pyx_k_pinf_tot), 0, 0, 1, 1},
  {&__pyx_n_s_pinf_tot_i, __pyx_k_pinf_tot_i, sizeof(__pyx_k_pinf_tot_i), 0, 0, 1, 1},
  {&__pyx_n_s_play, __pyx_k_play, sizeof(__pyx_k_play), 0, 0, 1, 1},
  {&__pyx_n_s_play_infections, __pyx_k_play_infections, sizeof(__pyx_k_play_infections), 0, 0, 1, 1},
  {&__pyx_n_s_play_infections_i, __pyx_k_play_infections_i, sizeof(__pyx_k_play_infections_i), 0, 0, 1, 1},
  {&__pyx_n_s_play_suscept, __pyx_k_play_suscept, sizeof(__pyx_k_play_suscept), 0, 0, 1, 1},
  {&__pyx_n_s_population, __pyx_k_population, sizeof(__pyx_k_population), 0, 0, 1, 1},
  {&__pyx_n_s_population_2, __pyx_k_population_2, sizeof(__pyx_k_population_2), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_print_population, __pyx_k_print_population, sizeof(__pyx_k_print_population), 0, 0, 1, 1},
  {&__pyx_n_s_profiler, __pyx_k_profiler, sizeof(__pyx_k_profiler), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_recovereds, __pyx_k_recovereds, sizeof(__pyx_k_recovereds), 0, 0, 1, 1},
  {&__pyx_n_s_redvar, __pyx_k_redvar, sizeof(__pyx_k_redvar), 0, 0, 1, 1},
  {&__pyx_n_s_safe_run, __pyx_k_safe_run, sizeof(__pyx_k_safe_run), 0, 0, 1, 1},
  {&__pyx_n_s_setup_core, __pyx_k_setup_core, sizeof(__pyx_k_setup_core), 0, 0, 1, 1},
  {&__pyx_n_u_setup_core, __pyx_k_setup_core, sizeof(__pyx_k_setup_core), 0, 1, 0, 1},
  {&__pyx_n_s_specialise, __pyx_k_specialise, sizeof(__pyx_k_specialise), 0, 0, 1, 1},
  {&__pyx_kp_s_src_metawards_extractors__output, __pyx_k_src_metawards_extractors__output, sizeof(__pyx_k_src_metawards_extractors__output), 0, 0, 1, 0},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_subinfs, __pyx_k_subinfs, sizeof(__pyx_k_subinfs), 0, 0, 1, 1},
  {&__pyx_n_s_subnet, __pyx_k_subnet, sizeof(__pyx_k_subnet), 0, 0, 1, 1},
  {&__pyx_n_s_subnets, __pyx_k_subnets, sizeof(__pyx_k_subnets), 0, 0, 1, 1},
  {&__pyx_n_s_subpops, __pyx_k_subpops, sizeof(__pyx_k_subpops), 0, 0, 1, 1},
  {&__pyx_n_s_subspaces, __pyx_k_subspaces, sizeof(__pyx_k_subspaces), 0, 0, 1, 1},
  {&__pyx_n_s_suscept, __pyx_k_suscept, sizeof(__pyx_k_suscept), 0, 0, 1, 1},
  {&__pyx_n_s_susceptibles, __pyx_k_susceptibles, sizeof(__pyx_k_susceptibles), 0, 0, 1, 1},
  {&__pyx_n_s_susceptibles_i, __pyx_k_susceptibles_i, sizeof(__pyx_k_susceptibles_i), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_thread_id, __pyx_k_thread_id, sizeof(__pyx_k_thread_id), 0, 0, 1, 1},
  {&__pyx_n_s_total, __pyx_k_total, sizeof(__pyx_k_total), 0, 0, 1, 1},
  {&__pyx_n_s_total_inf_ward, __pyx_k_total_inf_ward, sizeof(__pyx_k_total_inf_ward), 0, 0, 1, 1},
  {&__pyx_n_s_total_inf_ward_buffer, __pyx_k_total_inf_ward_buffer, sizeof(__pyx_k_total_inf_ward_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_total_new, __pyx_k_total_new, sizeof(__pyx_k_total_new), 0, 0, 1, 1},
  {&__pyx_n_s_total_new_i, __pyx_k_total_new_i, sizeof(__pyx_k_total_new_i), 0, 0, 1, 1},
  {&__pyx_n_s_total_new_inf_ward, __pyx_k_total_new_inf_ward, sizeof(__pyx_k_total_new_inf_ward), 0, 0, 1, 1},
  {&__pyx_n_s_total_new_inf_ward_buffer, __pyx_k_total_new_inf_ward_buffer, sizeof(__pyx_k_total_new_inf_ward_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_n_s_utils__console, __pyx_k_utils__console, sizeof(__pyx_k_utils__console), 0, 0, 1, 1},
  {&__pyx_n_s_utils__profiler, __pyx_k_utils__profiler, sizeof(__pyx_k_utils__profiler), 0, 0, 1, 1},
  {&__pyx_kp_u_vs, __pyx_k_vs, sizeof(__pyx_k_vs), 0, 1, 0, 0},
  {&__pyx_n_s_wards, __pyx_k_wards, sizeof(__pyx_k_wards), 0, 0, 1, 1},
  {&__pyx_n_s_work, __pyx_k_work, sizeof(__pyx_k_work), 0, 0, 1, 1},
  {&__pyx_n_s_workspace, __pyx_k_workspace, sizeof(__pyx_k_workspace), 0, 0, 1, 1},
  {&__pyx_n_s_workspace_2, __pyx_k_workspace_2, sizeof(__pyx_k_workspace_2), 0, 0, 1, 1},
  {&__pyx_n_s_zero_all, __pyx_k_zero_all, sizeof(__pyx_k_zero_all), 0, 0, 1, 1},
  {&__pyx_n_s_zero_subspaces, __pyx_k_zero_subspaces, sizeof(__pyx_k_zero_subspaces), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 39, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 805, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "metawards/extractors/_output_core.pyx":167
 * 
 * 
 * def setup_core(nthreads: int = 1, **kwargs):             # <<<<<<<<<<<<<<
 *     """This is the setup function that corresponds with
 *        :meth:`~metawards.extractors.output_core`.
 */
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_n_s_nthreads, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_metawards_extractors__output, __pyx_n_s_setup_core, 167, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 167, __pyx_L1_error)

  /* "metawards/extractors/_output_core.pyx":198
 * 
 * 
 * def output_core_omp(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                     workspace: Workspace,
 *                     infections: Infections,
 */
  __pyx_tuple__4 = PyTuple_Pack(52, __pyx_n_s_network, __pyx_n_s_population, __pyx_n_s_workspace, __pyx_n_s_infections, __pyx_n_s_nthreads, __pyx_n_s_kwargs, __pyx_n_s_links, __pyx_n_s_wards, __pyx_n_s_play_infections, __pyx_n_s_N_INF_CLASSES, __pyx_n_s_inf_tot, __pyx_n_s_pinf_tot, __pyx_n_s_total_inf_ward, __pyx_n_s_total_new_inf_ward, __pyx_n_s_n_inf_wards, __pyx_n_s_incidence, __pyx_n_s_S_in_wards, __pyx_n_s_E_in_wards, __pyx_n_s_I_in_wards, __pyx_n_s_R_in_wards, __pyx_n_s_links_ifrom, __pyx_n_s_links_suscept, __pyx_n_s_play_suscept, __pyx_n_s_nlinks_plus_one, __pyx_n_s_nnodes_plus_one, __pyx_n_s_infections_i, __pyx_n_s_play_infections_i, __pyx_n_s_total_new, __pyx_n_s_susceptibles, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_num_threads, __pyx_n_s_thread_id, __pyx_n_s_ifrom, __pyx_n_s_N_INF_CLASSES_MINUS_ONE, __pyx_n_s_lock, __pyx_n_s_total_new_inf_ward_buffer, __pyx_n_s_total_inf_ward_buffer, __pyx_n_s_S_buffer, __pyx_n_s_E_buffer, __pyx_n_s_I_buffer, __pyx_n_s_R_buffer, __pyx_n_s_redvar, __pyx_n_s_latent, __pyx_n_s_total, __pyx_n_s_recovereds, __pyx_n_s_S, __pyx_n_s_E, __pyx_n_s_I, __pyx_n_s_R, __pyx_n_s_error, __pyx_n_s_Console); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(5, 0, 52, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_metawards_extractors__output, __pyx_n_s_output_core_omp, 198, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 198, __pyx_L1_error)

  /* "metawards/extractors/_output_core.pyx":487
 * 
 * 
 * def output_core_serial(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                        workspace: Workspace,
 *                        infections: Infections,
 */
  __pyx_tuple__6 = PyTuple_Pack(46, __pyx_n_s_network, __pyx_n_s_population, __pyx_n_s_workspace, __pyx_n_s_infections, __pyx_n_s_kwargs, __pyx_n_s_links, __pyx_n_s_wards, __pyx_n_s_play_infections, __pyx_n_s_N_INF_CLASSES, __pyx_n_s_inf_tot, __pyx_n_s_pinf_tot, __pyx_n_s_total_inf_ward, __pyx_n_s_total_new_inf_ward, __pyx_n_s_n_inf_wards, __pyx_n_s_incidence, __pyx_n_s_S_in_wards, __pyx_n_s_E_in_wards, __pyx_n_s_I_in_wards, __pyx_n_s_R_in_wards, __pyx_n_s_links_ifrom, __pyx_n_s_links_suscept, __pyx_n_s_play_suscept, __pyx_n_s_nlinks_plus_one, __pyx_n_s_nnodes_plus_one, __pyx_n_s_infections_i, __pyx_n_s_play_infections_i, __pyx_n_s_total_new, __pyx_n_s_susceptibles, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_ifrom, __pyx_n_s_n_inf_wards_i, __pyx_n_s_total_new_i, __pyx_n_s_inf_tot_i, __pyx_n_s_pinf_tot_i, __pyx_n_s_susceptibles_i, __pyx_n_s_N_INF_CLASSES_MINUS_ONE, __pyx_n_s_latent, __pyx_n_s_total, __pyx_n_s_recovereds, __pyx_n_s_S, __pyx_n_s_E, __pyx_n_s_I, __pyx_n_s_R, __pyx_n_s_error, __pyx_n_s_Console); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(4, 0, 46, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_metawards_extractors__output, __pyx_n_s_output_core_serial, 487, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 487, __pyx_L1_error)

  /* "metawards/extractors/_output_core.pyx":731
 * 
 * 
 * def _safe_run(func, **kwargs):             # <<<<<<<<<<<<<<
 *     try:
 *         func(**kwargs)
 */
  __pyx_tuple__8 = PyTuple_Pack(4, __pyx_n_s_func, __pyx_n_s_kwargs, __pyx_n_s_e, __pyx_n_s_Console); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_metawards_extractors__output, __pyx_n_s_safe_run, 731, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(0, 731, __pyx_L1_error)

  /* "metawards/extractors/_output_core.pyx":748
 * 
 * 
 * def output_core(network: _Union[Network, Networks],             # <<<<<<<<<<<<<<
 *                 population: Population,
 *                 workspace: Workspace,
 */
  __pyx_tuple__10 = PyTuple_Pack(13, __pyx_n_s_network, __pyx_n_s_population, __pyx_n_s_workspace, __pyx_n_s_infections, __pyx_n_s_nthreads, __pyx_n_s_profiler, __pyx_n_s_kwargs, __pyx_n_s_output_func, __pyx_n_s_Console, __pyx_n_s_NullProfiler, __pyx_n_s_p, __pyx_n_s_i, __pyx_n_s_subnet); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(6, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_metawards_extractors__output, __pyx_n_s_output_core, 748, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #ifdef WITH_THREAD
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("metawards.utils._get_array_ptr"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_double_array_ptr", (void (**)(void))&__pyx_f_9metawards_5utils_14_get_array_ptr_get_double_array_ptr, "double *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_int_array_ptr", (void (**)(void))&__pyx_f_9metawards_5utils_14_get_array_ptr_get_int_array_ptr, "int *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_output_core(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_output_core(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__output_core(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__output_core(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__output_core(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_output_core' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__output_core(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_output_core", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_metawards__extractors___output_core) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "metawards.extractors._output_core")) {
      if (unlikely(PyDict_SetItemString(modules, "metawards.extractors._output_core", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "metawards/extractors/_output_core.pyx":10
 * cimport openmp
 * 
 * from typing import Union as _Union             # <<<<<<<<<<<<<<
 * 
 * from .._network import Network
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Union);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Union); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union_2, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":12
 * from typing import Union as _Union
 * 
 * from .._network import Network             # <<<<<<<<<<<<<<
 * from .._networks import Networks
 * from .._population import Population
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Network);
  __Pyx_GIVEREF(__pyx_n_s_Network);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Network);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_network_2, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Network); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Network, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":13
 * 
 * from .._network import Network
 * from .._networks import Networks             # <<<<<<<<<<<<<<
 * from .._population import Population
 * from .._infections import Infections
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Networks);
  __Pyx_GIVEREF(__pyx_n_s_Networks);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Networks);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_networks, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Networks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Networks, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":14
 * from .._network import Network
 * from .._networks import Networks
 * from .._population import Population             # <<<<<<<<<<<<<<
 * from .._infections import Infections
 * from ..utils._profiler import Profiler
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Population);
  __Pyx_GIVEREF(__pyx_n_s_Population);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Population);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_population_2, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Population); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Population, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":15
 * from .._networks import Networks
 * from .._population import Population
 * from .._infections import Infections             # <<<<<<<<<<<<<<
 * from ..utils._profiler import Profiler
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Infections);
  __Pyx_GIVEREF(__pyx_n_s_Infections);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Infections);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_infections_2, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Infections); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Infections, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":16
 * from .._population import Population
 * from .._infections import Infections
 * from ..utils._profiler import Profiler             # <<<<<<<<<<<<<<
 * 
 * from .._workspace import Workspace
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Profiler);
  __Pyx_GIVEREF(__pyx_n_s_Profiler);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Profiler);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_utils__profiler, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Profiler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Profiler, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metawards/extractors/_output_core.pyx":18
 * from ..utils._profiler import Profiler
 * 
 * from .._workspace import Workspace             # <<<<<<<<<<<<<<
 * 
 * from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Workspace);
  __Pyx_GIVEREF(__pyx_n_s_Workspace);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Workspace);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_workspace_2, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Workspace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Workspace, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":22
 * from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr
 * 
 * __all__ = ["setup_core", "output_core", "output_core_omp",             # <<<<<<<<<<<<<<
 *            "output_core_serial"]
 * 
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_setup_core);
  __Pyx_GIVEREF(__pyx_n_u_setup_core);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_setup_core);
  __Pyx_INCREF(__pyx_n_u_output_core);
  __Pyx_GIVEREF(__pyx_n_u_output_core);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_output_core);
  __Pyx_INCREF(__pyx_n_u_output_core_omp);
  __Pyx_GIVEREF(__pyx_n_u_output_core_omp);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_u_output_core_omp);
  __Pyx_INCREF(__pyx_n_u_output_core_serial);
  __Pyx_GIVEREF(__pyx_n_u_output_core_serial);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_u_output_core_serial);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_2) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":154
 * # from one iteration to the next - this is safe as metawards will
 * #only run one model run at a time per process
 * cdef int _buffer_nthreads = 0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _total_new_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _total_inf_ward_buffers = <_inf_buffer*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__buffer_nthreads = 0;

  /* "metawards/extractors/_output_core.pyx":155
 * #only run one model run at a time per process
 * cdef int _buffer_nthreads = 0
 * cdef _inf_buffer * _total_new_inf_ward_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _total_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _S_buffers = <_inf_buffer*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__total_new_inf_ward_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":156
 * cdef int _buffer_nthreads = 0
 * cdef _inf_buffer * _total_new_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _total_inf_ward_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _S_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _E_buffers = <_inf_buffer*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__total_inf_ward_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":157
 * cdef _inf_buffer * _total_new_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _total_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _S_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _E_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _I_buffers = <_inf_buffer*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__S_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":158
 * cdef _inf_buffer * _total_inf_ward_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _S_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _E_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _I_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _R_buffers = <_inf_buffer*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__E_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":159
 * cdef _inf_buffer * _S_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _E_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _I_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * cdef _inf_buffer * _R_buffers = <_inf_buffer*>0
 * 
 */
  __pyx_v_9metawards_10extractors_12_output_core__I_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":160
 * cdef _inf_buffer * _E_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _I_buffers = <_inf_buffer*>0
 * cdef _inf_buffer * _R_buffers = <_inf_buffer*>0             # <<<<<<<<<<<<<<
 * 
 * cdef _red_variables * _redvars = <_red_variables*>0
 */
  __pyx_v_9metawards_10extractors_12_output_core__R_buffers = ((struct __pyx_t_9metawards_10extractors_12_output_core__inf_buffer *)0);

  /* "metawards/extractors/_output_core.pyx":162
 * cdef _inf_buffer * _R_buffers = <_inf_buffer*>0
 * 
 * cdef _red_variables * _redvars = <_red_variables*>0             # <<<<<<<<<<<<<<
 * 
 * _files = None
 */
  __pyx_v_9metawards_10extractors_12_output_core__redvars = ((struct __pyx_t_9metawards_10extractors_12_output_core__red_variables *)0);

  /* "metawards/extractors/_output_core.pyx":164
 * cdef _red_variables * _redvars = <_red_variables*>0
 * 
 * _files = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_files, Py_None) < 0) __PYX_ERR(0, 164, __pyx_L1_error)

  /* "metawards/extractors/_output_core.pyx":167
 * 
 * 
 * def setup_core(nthreads: int = 1, **kwargs):             # <<<<<<<<<<<<<<
 *     """This is the setup function that corresponds with
 *        :meth:`~metawards.extractors.output_core`.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9metawards_10extractors_12_output_core_1setup_core, NULL, __pyx_n_s_metawards_extractors__output_cor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setup_core, __pyx_t_2) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":198
 * 
 * 
 * def output_core_omp(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                     workspace: Workspace,
 *                     infections: Infections,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9metawards_10extractors_12_output_core_3output_core_omp, NULL, __pyx_n_s_metawards_extractors__output_cor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_output_core_omp, __pyx_t_2) < 0) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":487
 * 
 * 
 * def output_core_serial(network: Network, population: Population,             # <<<<<<<<<<<<<<
 *                        workspace: Workspace,
 *                        infections: Infections,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9metawards_10extractors_12_output_core_5output_core_serial, NULL, __pyx_n_s_metawards_extractors__output_cor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_output_core_serial, __pyx_t_2) < 0) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":731
 * 
 * 
 * def _safe_run(func, **kwargs):             # <<<<<<<<<<<<<<
 *     try:
 *         func(**kwargs)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9metawards_10extractors_12_output_core_7_safe_run, NULL, __pyx_n_s_metawards_extractors__output_cor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_safe_run, __pyx_t_2) < 0) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":748
 * 
 * 
 * def output_core(network: _Union[Network, Networks],             # <<<<<<<<<<<<<<
 *                 population: Population,
 *                 workspace: Workspace,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_9metawards_10extractors_12_output_core_9output_core, NULL, __pyx_n_s_metawards_extractors__output_cor); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_output_core, __pyx_t_2) < 0) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metawards/extractors/_output_core.pyx":1
 * #!/bin/env/python3             # <<<<<<<<<<<<<<
 * #cython: linetrace=False
 * # MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init metawards.extractors._output_core", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init metawards.extractors._output_core");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* MergeKeywords */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping) {
    PyObject *iter, *key = NULL, *value = NULL;
    int source_is_dict, result;
    Py_ssize_t orig_length, ppos = 0;
    iter = __Pyx_dict_iterator(source_mapping, 0, __pyx_n_s_items, &orig_length, &source_is_dict);
    if (unlikely(!iter)) {
        PyObject *args;
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        args = PyTuple_Pack(1, source_mapping);
        if (likely(args)) {
            PyObject *fallback = PyObject_Call((PyObject*)&PyDict_Type, args, NULL);
            Py_DECREF(args);
            if (likely(fallback)) {
                iter = __Pyx_dict_iterator(fallback, 1, __pyx_n_s_items, &orig_length, &source_is_dict);
                Py_DECREF(fallback);
            }
        }
        if (unlikely(!iter)) goto bad;
    }
    while (1) {
        result = __Pyx_dict_iter_next(iter, orig_length, &ppos, &key, &value, NULL, source_is_dict);
        if (unlikely(result < 0)) goto bad;
        if (!result) break;
        if (unlikely(PyDict_Contains(kwdict, key))) {
            __Pyx_RaiseDoubleKeywordsError("function", key);
            result = -1;
        } else {
            result = PyDict_SetItem(kwdict, key, value);
        }
        Py_DECREF(key);
        Py_DECREF(value);
        if (unlikely(result < 0)) goto bad;
    }
    Py_XDECREF(iter);
    return 0;
bad:
    Py_XDECREF(iter);
    return -1;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
